/* KameJsonParser.cc */
/* Created by Enomoto Sanshiro on 4 April 2012. */
/* Last updated by Enomoto Sanshiro on 4 April 2012 */


#include <string>
#include <iostream>
#include <cctype>
#include "KameJsonParser.hh"
#include "KameTree.hh"

using namespace std;
using namespace kame;


TKameJsonTokenizer::TKameJsonTokenizer(std::istream& Input)
: _Input(Input)
{
}

TKameJsonTokenizer::~TKameJsonTokenizer()
{
}

bool TKameJsonTokenizer::GetNext(std::string& Token)
{
    if (! _TokenQueue.empty()) {
        Token = _TokenQueue.back();
        _TokenQueue.pop_back();
        return true;
    }

    return Tokenize(Token);
}

TKameJsonTokenizer& TKameJsonTokenizer::PushBack(const std::string& Token)
{
    _TokenQueue.push_back(Token);
    return *this;
}

bool TKameJsonTokenizer::Tokenize(std::string& Token)
{
    bool IsInQuote = false;
    bool IsEscaped = false;
    bool IsEmpty = true;
    Token.clear();

    char Ch;
    while ((Ch = GetNextChar()) != '\0') {
        if (IsEscaped) {
            Token += Ch;
            IsEscaped = false;
            continue;
        }
        if (Ch == '\\') {
            IsEscaped = true;
            continue;
        }
        if (Ch == '"') {
            IsEmpty = false;
            IsInQuote = ! IsInQuote;
            continue;
        }
        if (IsInQuote) {
            Token += Ch;
            continue;
        }
        if (isspace(Ch)) {
            if (IsEmpty && Token.empty()) {
                continue;
            }
            else {
                PushBackChar(Ch);
                break;
            }
        }
        if (
            (Ch == '[') || (Ch == ']') || 
            (Ch == '{') || (Ch == '}') || 
            (Ch == ',') || (Ch == ':')
        ){
            if (IsEmpty && Token.empty()) {
                Token = Ch;
            }
            else {
                PushBackChar(Ch);
            }
            break;
        }

        Token += Ch;
    }

    return ! (IsEmpty && Token.empty());
}

char TKameJsonTokenizer::GetNextChar(void)
{
    if (! _CharQueue.empty()) {
        char Ch = _CharQueue.back();
        _CharQueue.pop_back();
        return Ch;
    }

    char Ch;
    if ((! _Input) || (! _Input.get(Ch))) {
        return '\0';
    }
    return Ch;
}

void TKameJsonTokenizer::PushBackChar(char Ch)
{
    _CharQueue.push_back(Ch);
}



TKameJsonParser::TKameJsonParser(void)
{
}

TKameJsonParser::~TKameJsonParser()
{
}

void TKameJsonParser::Parse(std::istream& Input, TKameTree& Tree)
{
    string Token;
    TKameJsonTokenizer Tokenizer(Input);
    ParseElement(Tokenizer, Tree);

    if (Tokenizer.GetNext(Token)) {
        throw TKameException("JSON syntax error: extra token before EOF");
    }
}

void TKameJsonParser::ParseElement(TKameJsonTokenizer& Tokenizer, TKameTree& Tree)
{
    string Token;
    if (! Tokenizer.GetNext(Token)) {
        return;
    }

    if (Token == "[") {
        ParseIndexedArray(Tokenizer.PushBack(Token), Tree);
    }
    else if (Token == "{") {
        ParseAssosiativeArray(Tokenizer.PushBack(Token), Tree);
    }
    else {
        Tree.Value() = Token;
    }
}

void TKameJsonParser::ParseIndexedArray(TKameJsonTokenizer& Tokenizer, TKameTree& Tree)
{
    string Token;
    if (! Tokenizer.GetNext(Token) || (Token != "[")) {
        throw TKameException("JSON syntax error: '[' is expected");
    }

    Tree = TKameTree::TEmptyArray();

    int Index = 0;
    while (Tokenizer.GetNext(Token)) {
        if (Token == "]") {
            break;
        }

        ParseElement(Tokenizer.PushBack(Token), Tree[Index]);
        Index++;

        Tokenizer.GetNext(Token);
        if (Token == ",") {
            continue;
        }
        else if (Token == "]") {
            break;
        }
        else {
            throw TKameException("JSON syntax error: ']' is expected");
        }
    }
}

void TKameJsonParser::ParseAssosiativeArray(TKameJsonTokenizer& Tokenizer, TKameTree& Tree)
{
    string Token;
    if (! Tokenizer.GetNext(Token) || (Token != "{")) {
        throw TKameException("JSON syntax error: '{' is expected");
    }

    while (Tokenizer.GetNext(Token)) {
        if (Token == "}") {
            break;
        }

        string Key = Token;
        Tokenizer.GetNext(Token);
        if (Token != ":") {
            throw TKameException("JSON syntax error: ':' is expected");
        }
        ParseElement(Tokenizer, Tree[Key]);

        Tokenizer.GetNext(Token);
        if (Token == ",") {
            continue;
        }
        else if (Token == "}") {
            break;
        }
        else {
            throw TKameException("JSON syntax error: '}' is expected");
        }
    }
}
