/* KamePlainTextRepository.cc */
/* Created by Enomoto Sanshiro on 1 July 2009. */
/* Last updated by Enomoto Sanshiro on 7 July 2011. */


#include <string>
#include <iostream>
#include <sstream>
#include <fstream>
#include <iomanip>
#include <map>
#include <cstdio>
#include <cctype>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "KameSystemIO.hh"
#include "KameNtuple.hh"
#include "KameRepository.hh"
#include "KameTableFormat.hh"
#include "KamePlainTextRepository.hh"


using namespace std;
using namespace kame;


TKamePlainTextRepository::TKamePlainTextRepository(void)
: TKameTableFormat()
{
    // repository is a file //
    _RepositoryName = "";
    _IsRepositoryCreated = false;

    _IsCompressionEnabled = false;
}

TKamePlainTextRepository::TKamePlainTextRepository(const std::string& RepositoryName)
: TKameTableFormat()
{
    // repository is a directory //
    _RepositoryName = RepositoryName;
    _IsRepositoryCreated = false;

    _IsCompressionEnabled = false;
}

TKamePlainTextRepository::~TKamePlainTextRepository()
{
}

void TKamePlainTextRepository::EnableCompression(void)
{
    _IsCompressionEnabled = true;
}

string TKamePlainTextRepository::FileNameOf(const string& ObjectName, int Revision)
{
    ostringstream FileNameStream;

    if (! _RepositoryName.empty()) {
	FileNameStream << _RepositoryName << "/";
    }

    FileNameStream << ObjectName;
    if (Revision >= 0) {
	FileNameStream << ";" << Revision;
    }

    if (! _RepositoryName.empty()) {
	FileNameStream << ".knt";
    }

    return FileNameStream.str();
}

ostream* TKamePlainTextRepository::OpenOutputFile(const string& ObjectName, int Revision)
{
    if ((! _IsRepositoryCreated) && (! _RepositoryName.empty())) {
	if (access(_RepositoryName.c_str(), F_OK) != 0) {
	    mkdir(_RepositoryName.c_str(), 0755);
	}
	else {
	    struct stat FileStat;
	    if (stat(_RepositoryName.c_str(), &FileStat) < 0) {
		throw TKameException(
		    "TKamePlainTextRepository::OpenOutputFile()",
		    "systemcall stat(2) failed: " + _RepositoryName
		);
	    }
	    if (! S_ISDIR(FileStat.st_mode)) {
		throw TKameException(
		    "TKamePlainTextRepository::OpenOutputFile()",
		    "file already exists: " + _RepositoryName
		);
	    }
	}
    }

    string FileName = FileNameOf(ObjectName);
    if (Revision > 0) {
	string PreviousFileName = FileNameOf(ObjectName, Revision - 1);
	if (_IsCompressionEnabled) {
	    rename((FileName + ".gz").c_str(), (PreviousFileName + ".gz").c_str());
	}
	else {
	    rename(FileName.c_str(), PreviousFileName.c_str());
	}
    }

    ostream* File = 0;
    if (_IsCompressionEnabled) {
	string Command = "gzip - > " + FileName + ".gz";
	FILE* Pipe = popen(Command.c_str(), "w");
	if ((Pipe == NULL) || (Pipe == (void*) -1)) {
	    throw TKameException(
		"TKamePlainTextRepository::OpenOutputFile()",
		"unable to start 'gzip' for file: " + FileName
	    );
	}
	File = new TKameOutputCFileStream(Pipe);
	_CommandPipeTable[File] = Pipe;
    }
    else {
	File = new ofstream(FileName.c_str());
    }

    if (! *File) {
	delete File;
	throw TKameException(
	    "TKamePlainTextRepository::OpenOutputFile()",
	    "unable to create file: " + FileName
	);
    }

    return File;
}

istream* TKamePlainTextRepository::OpenInputFile(const string& ObjectName, int Revision)
{
    string FileName = FileNameOf(ObjectName, Revision);

    istream* File = 0;
    istream* MyFile = 0;
    if (access(FileName.c_str(), R_OK) == 0) {
	File = new ifstream(FileName.c_str());
	MyFile = File;
    }
    else if (access((FileName+".gz").c_str(), R_OK) == 0) {
	string Command = "gunzip -c " + FileName + ".gz";
	FILE* Pipe = popen(Command.c_str(), "r");
	if ((Pipe == NULL) || (Pipe == (void*) -1)) {
	    throw TKameException(
		"TKamePlainTextRepository::OpenInputFile()",
		"unable to start 'gunzip' for file: " + FileName
	    );
	}
	File = new TKameInputCFileStream(Pipe);
	MyFile = File;
	_CommandPipeTable[File] = Pipe;
    }
    else if (FileName.empty()) {
	File = &cin;
	MyFile = 0;
    }

    if ((! File) || (! *File)) {
	delete MyFile;
	throw TKameException(
	    "TKamePlainTextRepository::OpenInputFile()",
	    "unable to open file: " + FileName
	);
    }
    _MyFileTable[File] = MyFile;

    return File;
}

void TKamePlainTextRepository::CloseOutputFile(ostream* File)
{
    map<ios*, FILE*>::iterator Iterator = _CommandPipeTable.find(File);
    if (Iterator != _CommandPipeTable.end()) {
	pclose(Iterator->second);
	_CommandPipeTable.erase(Iterator);
    }

    delete File;
}

void TKamePlainTextRepository::CloseInputFile(istream* File)
{
    map<ios*, FILE*>::iterator Iterator = _CommandPipeTable.find(File);
    if (Iterator != _CommandPipeTable.end()) {
	pclose(Iterator->second);
	_CommandPipeTable.erase(Iterator);
    }

    delete _MyFileTable[File];
}

void TKamePlainTextRepository::SaveNtuple(const TKameNtuple& Ntuple, const std::string& Name)
{
    int Revision = _RevisionTable[Name];
    _RevisionTable[Name] += 1;

    ostream* File = OpenOutputFile(Name, Revision);
    TKameTableFormat::Write(Ntuple, *File);
    CloseOutputFile(File);
}

void TKamePlainTextRepository::LoadNtuple(TKameNtuple& Ntuple, const std::string& Name, int Revision)
{
    istream* File = OpenInputFile(Name, Revision);
    TKameTableFormat::Read(Ntuple, *File);
    CloseInputFile(File);
}
