/* KameRepository.cc */
/* Created by Enomoto Sanshiro on 9 July 2009. */
/* Last updated by Enomoto Sanshiro on 9 July 2009. */


#include <string>
#include "KameNtuple.hh"
#include "KameGraph.hh"
#include "KameHistogram.hh"
#include "KameHistogram2d.hh"
#include "KameTrend.hh"
#include "KameTabular.hh"
#include "KameHistory.hh"
#include "KameRepository.hh"


using namespace std;
using namespace kame;



void TKameRepository::SaveGraph(const TKameGraph& Graph, const std::string& Name)
{
    TKameNtuple Ntuple;

    Ntuple.Tree() = Graph.PropertyTree();
    Ntuple["Type"] = "Graph";

    Ntuple.SetColumnProperties(0, "X", "float");
    Ntuple.SetColumnProperties(1, "Y", "float");
    if (Graph.HasErrors()) {
	Ntuple.SetColumnProperties(2, "XError", "float");
	Ntuple.SetColumnProperties(3, "YError", "float");
    }

    unsigned NumberOfPoints = Graph.NumberOfPoints();
    const double* XValueList = Graph.XValueList();
    const double* YValueList = Graph.YValueList();
    const double* XErrorList = Graph.XErrorList();
    const double* YErrorList = Graph.YErrorList();

    unsigned SegmentIndex = 0;
    unsigned NextSegmentBoundary = Graph.SegmentSizeOf(SegmentIndex);

    for (unsigned Index = 0; Index < NumberOfPoints; Index++) {
	if (Index == NextSegmentBoundary) {
	    NextSegmentBoundary += Graph.SegmentSizeOf(++SegmentIndex);
	    Ntuple.BreakSegment();
	}

	Ntuple[Index][0] = XValueList[Index];
	Ntuple[Index][1] = YValueList[Index];
	if (Graph.HasErrors()) {
	    Ntuple[Index][2] = XErrorList[Index];
	    Ntuple[Index][3] = YErrorList[Index];
	}
    }

    this->SaveNtuple(Ntuple, Name);
}

void TKameRepository::LoadGraph(TKameGraph& Graph, const std::string& Name, int Revision)
{
    TKameNtuple Ntuple;
    this->LoadNtuple(Ntuple, Name, Revision);

    if (Ntuple["Type"].As<string>() != "Graph") {
	throw TKameException(
	    "TKameRepository::LoadGraph()",
	    "Graph object is expected: " + Name +
	    " (" + Ntuple["Type"].As<string>() + ")"
	);
    }
    Graph.PropertyTree() = Ntuple.Tree();

    int NumberOfRows = Ntuple.NumberOfRows();
    bool HasErrors = (Ntuple.NumberOfColumns() >= 3);
    unsigned CurrentSegmentIndex = 0;
    unsigned CurrentSegmentSize = 0;

    for (int i = 0; i < NumberOfRows; i++) {
	if (Ntuple.SegmentIndexOf(i) != CurrentSegmentIndex) {
	    Graph.BreakSegment();
	    CurrentSegmentIndex = Ntuple.SegmentIndexOf(i);
	    CurrentSegmentSize = 0;
	}

	if (HasErrors) {
	    Graph.Fill(
		Ntuple[i][0], Ntuple[i][1], Ntuple[i][2], Ntuple[i][3]
	    );
	}
	else {
	    Graph.Fill(Ntuple[i][0], Ntuple[i][1]);
	}
    }
    
    if (CurrentSegmentSize > 0) {
	Graph.BreakSegment();
    }
}

void TKameRepository::SaveHistogram(const TKameHistogram& Histogram, const std::string& Name)
{
    TKameNtuple Ntuple;

    Ntuple.Tree() = Histogram.PropertyTree();
    Ntuple["Type"] = "Histogram";

    Ntuple["NumberOfBins"] = Histogram.Scale().NumberOfBins();
    Ntuple["Min"] = Histogram.Scale().Min();
    Ntuple["Max"] = Histogram.Scale().Max();
    Ntuple["Overflow"] = Histogram.OverflowCounts();
    Ntuple["Underflow"] = Histogram.UnderflowCounts();
    Ntuple["AxisTitle"] = Histogram.Scale().Title();

    long AccumulationTime = Histogram.AccumulationTime();
    if (AccumulationTime >= 0) {
	Ntuple["AccumulationTime"] = AccumulationTime;
    }

    Ntuple.SetColumnProperties(0, "BinCenter", "float");
    Ntuple.SetColumnProperties(1, "Counts", "float");
    if (Histogram.HasErrors()) {
	Ntuple.SetColumnProperties(2, "Error", "float");
    }

    for (int Index = 0; Index < Histogram.Scale().NumberOfBins(); Index++) {
	Ntuple[Index][0] = Histogram.Scale().BinCenterOf(Index);
	Ntuple[Index][1] = Histogram.CountsOf(Index);
	if (Histogram.HasErrors()) {
	    Ntuple[Index][2] = Histogram.ErrorOf(Index);
	}
    }

    this->SaveNtuple(Ntuple, Name);
}

void TKameRepository::LoadHistogram(TKameHistogram& Histogram, const std::string& Name, int Revision)
{
    TKameNtuple Ntuple;
    this->LoadNtuple(Ntuple, Name, Revision);

    if (Ntuple["Type"].As<string>() != "Histogram") {
	throw TKameException(
	    "TKameRepository::LoadHistogram()",
	    "Histogram object is expected: " + Name +
	    " (" + Ntuple["Type"].As<string>() + ")"
	);
    }
    Histogram.PropertyTree() = Ntuple.Tree();

    int NumberOfBins = Ntuple["NumberOfBins"].As<int>();
    double Min = Ntuple["Min"].As<double>();
    double Max = Ntuple["Max"].As<double>();
    Histogram.SetScale(TKameHistogramScale(NumberOfBins, Min, Max));
    
    if (! Ntuple["AxisTitle"].Value().IsVoid()) {
	Histogram.Scale().SetTitle(Ntuple["AxisTitle"].As<string>());
    }
    if (! Ntuple["AccumulationTime"].Value().IsVoid()) {
	Histogram.SetAccumulationTime(Ntuple["AccumulationTime"].As<long>());
    }

    int NumberOfRows = Ntuple.NumberOfRows();
    bool HasErrors = (Ntuple.NumberOfColumns() >= 3);
    for (int i = 0; i < NumberOfRows; i++) {
	if (HasErrors) {
	    Histogram.Fill(Ntuple[i][0], Ntuple[i][1], Ntuple[i][2]);
	}
	else {
	    Histogram.Fill(Ntuple[i][0], Ntuple[i][1]);
	}
    }
}

void TKameRepository::SaveHistogram2d(const TKameHistogram2d& Histogram, const std::string& Name)
{
    TKameNtuple Ntuple;

    Ntuple.Tree() = Histogram.PropertyTree();
    Ntuple["Type"] = "Histogram2d";

    Ntuple["NumberOfXBins"] = Histogram.XScale().NumberOfBins();
    Ntuple["XMin"] = Histogram.XScale().Min();
    Ntuple["XMax"] = Histogram.XScale().Max();
    Ntuple["NumberOfYBins"] = Histogram.YScale().NumberOfBins();
    Ntuple["YMin"] = Histogram.YScale().Min();
    Ntuple["YMax"] = Histogram.YScale().Max();
    Ntuple["XAxisTitle"] = Histogram.XScale().Title();
    Ntuple["YAxisTitle"] = Histogram.YScale().Title();

    Ntuple.SetColumnProperties(0, "XBinCenter", "float");
    Ntuple.SetColumnProperties(1, "YBinCenter", "float");
    Ntuple.SetColumnProperties(2, "Counts", "float");

    int XIndex, YIndex, RowIndex = 0;
    for (YIndex = 0; YIndex < Histogram.YScale().NumberOfBins(); YIndex++) {
	for (XIndex = 0; XIndex < Histogram.XScale().NumberOfBins(); XIndex++) {
	    Ntuple[RowIndex][0] = Histogram.XScale().BinCenterOf(XIndex);
	    Ntuple[RowIndex][1] = Histogram.YScale().BinCenterOf(YIndex);
	    Ntuple[RowIndex][2] = Histogram.CountsOf(XIndex, YIndex);
	    RowIndex++;
	}
    }

    this->SaveNtuple(Ntuple, Name);
}

void TKameRepository::LoadHistogram2d(TKameHistogram2d& Histogram, const std::string& Name, int Revision)
{
    TKameNtuple Ntuple;
    this->LoadNtuple(Ntuple, Name, Revision);

    if (Ntuple["Type"].As<string>() != "Histogram2d") {
	throw TKameException(
	    "TKameRepository::LoadHistogram2d()",
	    "Histogram2d object is expected: " + Name +
	    " (" + Ntuple["Type"].As<string>() + ")"
	);
    }
    Histogram.PropertyTree() = Ntuple.Tree();

    int NumberOfXBins = Ntuple["NumberOfXBins"].As<int>();
    double XMin = Ntuple["XMin"].As<double>();
    double XMax = Ntuple["XMax"].As<double>();
    int NumberOfYBins = Ntuple["NumberOfYBins"].As<int>();
    double YMin = Ntuple["YMin"].As<double>();
    double YMax = Ntuple["YMax"].As<double>();
    Histogram.SetScale(
	TKameHistogramScale(NumberOfXBins, XMin, XMax),
	TKameHistogramScale(NumberOfYBins, YMin, YMax)
    );
    
    if (! Ntuple["XAxisTitle"].Value().IsVoid()) {
	Histogram.XScale().SetTitle(Ntuple["XAxisTitle"].As<string>());
    }
    if (! Ntuple["YAxisTitle"].Value().IsVoid()) {
	Histogram.YScale().SetTitle(Ntuple["YAxisTitle"].As<string>());
    }

    int NumberOfRows = Ntuple.NumberOfRows();
    for (int i = 0; i < NumberOfRows; i++) {
	Histogram.Fill(Ntuple[i][0], Ntuple[i][1], Ntuple[i][2]);
    }
}

void TKameRepository::SaveTrend(const TKameTrend& Trend, const std::string& Name)
{
    TKameNtuple Ntuple;

    Ntuple.Tree() = Trend.PropertyTree();
    Ntuple["Type"] = "Trend";

    Ntuple["StartTime"] = Trend.StartTime();
    Ntuple["TickInterval"] = Trend.TickInterval();
    Ntuple["WindowLength"] = Trend.WindowLength();

    Ntuple.SetColumnProperties(0, "Time", "int");
    Ntuple.SetColumnProperties(1, "Counts", "float");
    Ntuple.SetColumnProperties(2, "Sum", "float");
    Ntuple.SetColumnProperties(3, "Mean", "float");
    Ntuple.SetColumnProperties(4, "Deviation", "float");
    Ntuple.SetColumnProperties(5, "MinValue", "float");
    Ntuple.SetColumnProperties(6, "MaxValue", "float");

    unsigned NumberOfPoints = Trend.NumberOfPoints();

    for (unsigned Index = 0; Index < NumberOfPoints; Index++) {
	Ntuple[Index][0] = Trend.PassedTimeOf(Index);
	Ntuple[Index][1] = Trend.CountsOf(Index);
	Ntuple[Index][2] = Trend.SumOf(Index);
	Ntuple[Index][3] = Trend.MeanOf(Index);
	Ntuple[Index][4] = Trend.DeviationOf(Index);
	Ntuple[Index][5] = Trend.MinValueOf(Index);
	Ntuple[Index][6] = Trend.MaxValueOf(Index);
    }

    this->SaveNtuple(Ntuple, Name);
}

void TKameRepository::LoadTrend(TKameTrend& Trend, const std::string& Name, int Revision)
{
    TKameNtuple Ntuple;
    this->LoadNtuple(Ntuple, Name, Revision);

    if (Ntuple["Type"].As<string>() != "Trend") {
	throw TKameException(
	    "TKameRepository::LoadTrend()",
	    "Trend object is expected: " + Name +
	    " (" + Ntuple["Type"].As<string>() + ")"
	);
    }
    Trend.PropertyTree() = Ntuple.Tree();

    long StartTime = Ntuple["StartTime"].As<long>();
    long TickInterval = Ntuple["TickInterval"].As<long>();
    long WindowLength = Ntuple["WindowLength"].As<long>();

    Trend.SetScale(TickInterval, WindowLength);
    Trend.Start(StartTime);

    int NumberOfRows = Ntuple.NumberOfRows();
    for (int i = 0; i < NumberOfRows; i++) {
	try {
	    long Time = Ntuple[i][0];
	    double Counts = Ntuple[i][1];
	    double Mean = Ntuple[i][3];
	    double Deviation = Ntuple[i][4];
	    double MinValue = Ntuple[i][4];
	    double MaxValue = Ntuple[i][4];
	    Trend.Fill(
		StartTime + Time, 
		Mean, Counts, Deviation, 
		MinValue, MaxValue
	    );
	}
	catch (TKameException &e) {
	    cout << "[" << Ntuple[i][0] << "]" << endl;
	    throw;
	}
    }
}

void TKameRepository::SaveTabular(const TKameTabular& Tabular, const std::string& Name)
{
    TKameNtuple Ntuple;

    Ntuple.Tree() = Tabular.PropertyTree();
    Ntuple.Table() = Tabular;
    Ntuple["Type"] = "Tabular";

    this->SaveNtuple(Ntuple, Name);
}

void TKameRepository::LoadTabular(TKameTabular& Tabular, const std::string& Name, int Revision)
{
    TKameNtuple Ntuple;
    this->LoadNtuple(Ntuple, Name, Revision);

    if (Ntuple["Type"].As<string>() != "Tabular") {
	throw TKameException(
	    "TKameRepository::LoadTabular()",
	    "Tabular object is expected: " + Name +
	    " (" + Ntuple["Type"].As<string>() + ")"
	);
    }
    Tabular.PropertyTree() = Ntuple.Tree();
    Tabular.Table() = Ntuple;
}

void TKameRepository::SaveHistory(const TKameHistory& History, const std::string& Name)
{
    TKameNtuple Ntuple;

    Ntuple.Tree() = History.PropertyTree();
    Ntuple["Type"] = "History";

    Ntuple["StartTime"] = History.StartTime();
    Ntuple["MaxNumberOfSamples"] = History.MaxNumberOfSamples();

    Ntuple.SetColumnProperties(0, "Time", "int");
    Ntuple.SetColumnProperties(1, "Counts", "float");
    Ntuple.SetColumnProperties(2, "Sum", "float");
    Ntuple.SetColumnProperties(3, "Mean", "float");
    Ntuple.SetColumnProperties(4, "Deviation", "float");

    unsigned NumberOfPoints = History.NumberOfSamples();

    for (unsigned Index = 0; Index < NumberOfPoints; Index++) {
	Ntuple[Index][0] = History.PassedTimeOf(Index);
	Ntuple[Index][1] = History.CountsOf(Index);
	Ntuple[Index][2] = History.SumOf(Index);
	Ntuple[Index][3] = History.MeanOf(Index);
	Ntuple[Index][4] = History.DeviationOf(Index);
    }

    this->SaveNtuple(Ntuple, Name);
}

void TKameRepository::LoadHistory(TKameHistory& History, const std::string& Name, int Revision)
{
    TKameNtuple Ntuple;
    this->LoadNtuple(Ntuple, Name, Revision);

    if (Ntuple["Type"].As<string>() != "History") {
	throw TKameException(
	    "TKameRepository::LoadHistory()",
	    "History object is expected: " + Name +
	    " (" + Ntuple["Type"].As<string>() + ")"
	);
    }
    History.PropertyTree() = Ntuple.Tree();

    long MaxNumberOfSamples = Ntuple["MaxNumberOfSamples"].As<long>();
    long StartTime = Ntuple["StartTime"].As<long>();

    History.SetScale(MaxNumberOfSamples);
    History.Start(StartTime);

    int NumberOfRows = Ntuple.NumberOfRows();
    for (int i = 0; i < NumberOfRows; i++) {
	try {
	    long Time = Ntuple[i][0];
	    long Counts = Ntuple[i][1];
	    double Sum = Ntuple[i][2];
	    double Mean = Ntuple[i][3];
	    double Deviation = Ntuple[i][4];
	    History.InsertSample(
		StartTime + Time, Counts, Sum, Mean, Deviation
	    );
	}
	catch (TKameException &e) {
	    cout << "[" << Ntuple[i][0] << "]" << endl;
	    throw;
	}
    }
}
