/* KameTable.hh */
/* Created by Enomoto Sanshiro on 29 June 2009. */
/* Last updated by Enomoto Sanshiro on 23 June 2011. */


#ifndef __KameTable_hh__
#define __KameTable_hh__


#include <vector>
#include <map>
#include "KameVariant.hh"


namespace kame {


class TKameTable {
  public:
    class TRow {
      public:
	inline TRow(TKameTable& Table, unsigned RowIndex);
	inline TKameVariant& operator[](unsigned ColumnIndex);
	inline TKameVariant& operator[](const std::string& ColumnName);
	inline const TKameVariant& operator[](unsigned ColumnIndex) const;
	inline const TKameVariant& operator[](const std::string& ColumnName) const;
      private:
	TKameTable& _Table;
	unsigned _RowIndex;
    };
  public:
    TKameTable(void);
    explicit TKameTable(unsigned NumberOfColumns);
    TKameTable(const TKameTable& Table);
    virtual ~TKameTable() {}
    virtual TKameTable& operator=(const TKameTable& Table);
    virtual unsigned AddColumn(const std::string& Name = "", int ColumnIndex = -1);
    virtual unsigned AddRow(int RowIndex = -1);
    virtual void Clear(void);
    virtual void Dump(std::ostream& os = std::cout);
  public:
    inline unsigned NumberOfRows(void) const;
    inline TRow operator[](unsigned RowIndex);
    inline const TRow operator[](unsigned RowIndex) const;
  public:
    inline unsigned NumberOfColumns(void) const;
    inline const std::vector<std::string>& ColumnNameList(void) const;
    inline unsigned ColumnIndexOf(const std::string& ColumnName) const;
    inline std::vector<TKameVariant>& GetColumn(unsigned ColumnIndex);
    inline const std::vector<TKameVariant>& GetColumn(unsigned ColumnIndex) const;
    inline std::vector<TKameVariant>& GetColumn(const std::string& ColumnName);
    inline const std::vector<TKameVariant>& GetColumn(const std::string& ColumnName) const;
  private:
    unsigned _NumberOfRows;
    std::vector<std::vector<TKameVariant> > _Table;
    std::vector<std::string> _ColumnNameList;
    std::map<std::string, unsigned> _ColumnNameIndexTable;
};



inline unsigned TKameTable::NumberOfRows(void) const
{
    return _NumberOfRows;
}

inline TKameTable::TRow TKameTable::operator[](unsigned RowIndex)
{
    if (RowIndex >= _NumberOfRows) {
        AddRow(RowIndex);
    }

    return TRow(*this, RowIndex);
}

inline const TKameTable::TRow TKameTable::operator[](unsigned RowIndex) const
{
    if (RowIndex >= _NumberOfRows) {
	throw TKameException("table row index out of range");
    }

    return TRow(const_cast<TKameTable&>(*this), RowIndex);
}

inline unsigned TKameTable::NumberOfColumns(void) const
{
    return _Table.size();
}

inline const std::vector<std::string>& TKameTable::ColumnNameList(void) const
{
    return _ColumnNameList;
}

inline std::vector<TKameVariant>& TKameTable::GetColumn(unsigned ColumnIndex)
{
    if (ColumnIndex >= _Table.size()) {
        AddColumn("", ColumnIndex);
    }

    return _Table[ColumnIndex];
}

inline const std::vector<TKameVariant>& TKameTable::GetColumn(unsigned ColumnIndex) const
{
    if (ColumnIndex >= _Table.size()) {
	throw TKameException("table column index out of range");
    }

    return _Table[ColumnIndex];
}

inline std::vector<TKameVariant>& TKameTable::GetColumn(const std::string& ColumnName)
{
    std::map<std::string, unsigned>::const_iterator Iterator = (
        _ColumnNameIndexTable.find(ColumnName)
    );
    if (Iterator == _ColumnNameIndexTable.end()) {
        return _Table[AddColumn(ColumnName)];
    }

    return _Table[Iterator->second];
}

inline const std::vector<TKameVariant>& TKameTable::GetColumn(const std::string& ColumnName) const
{
    std::map<std::string, unsigned>::const_iterator Iterator = (
        _ColumnNameIndexTable.find(ColumnName)
    );
    if (Iterator == _ColumnNameIndexTable.end()) {
	throw TKameException("unknown table column name: " + ColumnName);
    }

    return _Table[Iterator->second];
}

inline unsigned TKameTable::ColumnIndexOf(const std::string& ColumnName) const
{
    std::map<std::string, unsigned>::const_iterator Iterator = (
        _ColumnNameIndexTable.find(ColumnName)
    );
    if (Iterator == _ColumnNameIndexTable.end()) {
	throw TKameException("unknown table column name: " + ColumnName);
    }

    return Iterator->second;
}


inline TKameTable::TRow::TRow(TKameTable& Table, unsigned RowIndex)
: _Table(Table), _RowIndex(RowIndex)
{
}

inline TKameVariant& TKameTable::TRow::operator[](unsigned ColumnIndex)
{
    return _Table.GetColumn(ColumnIndex)[_RowIndex];
}

inline TKameVariant& TKameTable::TRow::operator[](const std::string& ColumnName)
{
    return _Table.GetColumn(ColumnName)[_RowIndex];
}

inline const TKameVariant& TKameTable::TRow::operator[](unsigned ColumnIndex) const
{
    return _Table.GetColumn(ColumnIndex)[_RowIndex];
}

inline const TKameVariant& TKameTable::TRow::operator[](const std::string& ColumnName) const
{
    return _Table.GetColumn(ColumnName)[_RowIndex];
}


}
#endif
