/* KameTableFormat.cc */
/* Created by Enomoto Sanshiro on 1 July 2009. */
/* Last updated by Enomoto Sanshiro on 8 November 2011. */


#include <string>
#include <iostream>
#include <sstream>
#include <fstream>
#include <iomanip>
#include <map>
#include <cstdio>
#include <cctype>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include "KameNtuple.hh"
#include "KameTableFormat.hh"


using namespace std;
using namespace kame;


TKameTableFormat::TKameTableFormat(void)
{
    _DefaultHeaderProcessor = 0;
    _HeaderProcessor = 0;

    _DefaultDelimiter = '\0';
    _DefaultQuote = '\0';
    _LineTerminator = '\0';  // auto-detect

    _CommentHeader = '#';
    _TreeIndent = '\0';

    _IsCsvHeaderEnabled = false;
}

TKameTableFormat::~TKameTableFormat()
{
    delete _DefaultHeaderProcessor;
}

void TKameTableFormat::EnableCsvHeader(void)
{
    _IsCsvHeaderEnabled = true;
    _DefaultDelimiter = ',';
}

void TKameTableFormat::SetDelimiter(char Delimiter)
{
    _DefaultDelimiter = Delimiter;
}

void TKameTableFormat::SetQuote(char Quote)
{
    _DefaultQuote = Quote;
}

void TKameTableFormat::SetLineTerminator(char LineTerminator)
{
    _LineTerminator = LineTerminator;
}

void TKameTableFormat::SetCommentHeader(char CommentHeader)
{
    _CommentHeader = CommentHeader;
}

void TKameTableFormat::SetTreeIndent(char TreeIndent)
{
    _TreeIndent = TreeIndent;
}

void TKameTableFormat::SetHeaderProcessor(TKameTableFormatHeaderProcessor* HeaderProcessor)
{
    _HeaderProcessor = HeaderProcessor;
}

void TKameTableFormat::Read(TKameNtuple& Ntuple, istream& Input)
{
    string Line;
    TKameTableFormatColumnParser ColumnParser;
    TKameTableFormatLineExtractor LineExtractor(Input, _LineTerminator);

    unsigned ColumnOffset = Ntuple.NumberOfColumns();
    TKameTree& HeaderTree = Ntuple.Tree();

    if (_HeaderProcessor == 0) {
        _DefaultHeaderProcessor = new TKameTableFormatHeaderProcessor();
        _HeaderProcessor = _DefaultHeaderProcessor;
    }
    _HeaderProcessor->SetStorage(&HeaderTree);
    if (_IsCsvHeaderEnabled) {
        _HeaderProcessor->EnableCsvHeader();
    }
    if (_DefaultDelimiter != '\0') {
        _HeaderProcessor->SetDelimiter(_DefaultDelimiter);
    }
    if (_DefaultQuote != '\0') {
        _HeaderProcessor->SetQuote(_DefaultQuote);
    }
    _HeaderProcessor->SetCommentHeader(_CommentHeader);
    _HeaderProcessor->SetTreeIndent(_TreeIndent);

    int RowIndex = 0, LineNumber = 0;
    bool IsParserInitialized = false;
    while (LineExtractor.GetNext(Line)) {
        LineNumber++;

	// blank line //
	if (Line.find_first_not_of(" \t\r\n\x0a\x0d") == string::npos) {
	    if (RowIndex > 0) {
		Ntuple.BreakSegment();
	    }
	    continue;
	}

	// header/comment line //
        if (LineNumber == 1) {
            if (_HeaderProcessor->ProcessFirstLine(Line)) {
                continue;
            }
        }
        if ((_CommentHeader == '\0') || (Line[0] == _CommentHeader)) {
            _HeaderProcessor->ProcessLine(Line);
            continue;
	}

	// table data line //
	if (! IsParserInitialized) {
            vector<string> ColumnTypeList;
            for (unsigned i = 0; i < HeaderTree["FieldType"].Length(); i++) {
                ColumnTypeList.push_back(
                    HeaderTree["FieldType"][i].As<string>()
                );
            }
            ColumnParser.SetColumnTypeList(ColumnTypeList);
            if (! HeaderTree["Delimiter"].IsVoid()) {
                ColumnParser.SetDelimiter(
                    HeaderTree["Delimiter"].As<string>()[0]
                );
            }
            else if (_DefaultDelimiter != '\0') {
                ColumnParser.SetDelimiter(_DefaultDelimiter);
            }
            if (! HeaderTree["Quote"].IsVoid()) {
		ColumnParser.SetQuote(
                    HeaderTree["Quote"].As<string>()[0]
                );
            }
            else if (_DefaultQuote != '\0') {
		ColumnParser.SetQuote(_DefaultQuote);
            }
	    IsParserInitialized = true;
	}
	try {
	    if (ColumnParser.ProcessLine(Ntuple[RowIndex], Line, ColumnOffset)) {
		RowIndex++;
	    }
	}
	catch (TKameException &e) {
	    ostringstream os;
	    os << "line " << LineNumber << ": " << e.what();
	    throw TKameException(os.str());
	}
    }

    for (unsigned i = 0; i < Ntuple.NumberOfColumns(); i++) {
        string ColumnName = HeaderTree["Field"][i].As<string>();
        string ColumnType = HeaderTree["FieldType"][i].As<string>();
        if (ColumnName.empty()) {
            ostringstream os;
            os << "Column" << setfill('0') << setw(2) << i;
            ColumnName = os.str();
        }
        Ntuple.SetColumnProperties(ColumnOffset + i, ColumnName, ColumnType);
    }
}

void TKameTableFormat::Write(const TKameNtuple& Ntuple, ostream& Output)
{
    TKameTree HeaderTree = Ntuple.Tree();

    string Delimiter, Quote;
    if (! HeaderTree["Delimiter"].Value().IsVoid()) {
        Delimiter = HeaderTree["Delimiter"].As<string>();
    }
    else if (_DefaultDelimiter != '\0') {
        Delimiter = _DefaultDelimiter;
    }
    if (! HeaderTree["Quote"].Value().IsVoid()) {
        Quote = HeaderTree["Quote"].As<string>();
    }
    else if (_DefaultQuote != '\0') {
        Quote = _DefaultQuote;
    }

    string HeaderQuote = Quote;
    if (Delimiter.empty() && Quote.empty()) {
	unsigned NumberOfColumns = Ntuple.NumberOfColumns();
	for (unsigned Column = 0; Column < NumberOfColumns; Column++) {
            string ColumnName = Ntuple.ColumnNameList()[Column];
	    if (ColumnName.find_first_of(" \t") != string::npos) {
		HeaderQuote = "\"";
                break;
	    }
	}
    }

    string ColumnList;
    unsigned NumberOfColumns = Ntuple.NumberOfColumns();
    for (unsigned Column = 0; Column < NumberOfColumns; Column++) {
        string ColumnName = Ntuple.ColumnNameList()[Column];
	if (Column > 0) {
	    ColumnList += Delimiter + " ";
	}
	ColumnList += HeaderQuote + ColumnName + HeaderQuote;
    }
    if (! ColumnList.empty()) {
        //... for backwards compatibility ...//
        HeaderTree["Fields"] = ColumnList;
    }

    const vector<string>& ColumnTypeList = Ntuple.ColumnTypeList();
    if (! ColumnTypeList.empty()) {
	string ColumnTypeListString;
	for (unsigned Column = 0; Column < ColumnTypeList.size(); Column++) {
	    if (Column > 0) {
		ColumnTypeListString += Delimiter + " ";
	    }
	    ColumnTypeListString += HeaderQuote + ColumnTypeList[Column] + HeaderQuote;
	}
        //... for backwards compatibility ...//
	HeaderTree["FieldTypes"] = ColumnTypeListString;
    }

    if (! Delimiter.empty()) {
        Output << _CommentHeader << " Delimiter: " << Delimiter << endl;
    }
    if (! Quote.empty()) {
        Output << _CommentHeader << " Quote: " << Quote << endl;
    }
    TKameKtfTreeSerializer(Output).Serialize(HeaderTree);
    Output << endl;

    unsigned CurrentSegmentIndex = 0;
    for (unsigned Row = 0; Row < Ntuple.NumberOfRows(); Row++) {
	if (Ntuple.SegmentIndexOf(Row) != CurrentSegmentIndex) {
	    CurrentSegmentIndex = Ntuple.SegmentIndexOf(Row);
	    Output << endl;
	}
	for (unsigned Column = 0; Column < NumberOfColumns; Column++) {
	    bool IsString = Ntuple[Row][Column].IsString();
	    if ((Column == 0) && Quote.empty() && IsString) {
		string Value = Ntuple[Row][Column];
		if ((Value.size() > 0) && (Value[0] == _CommentHeader)) {
		    Output << '\\';
		}
	    }
	    if (Column != 0) {
		Output << Delimiter << " ";
	    }
		
	    if (IsString) {
		Output << Quote << Ntuple[Row][Column] << Quote;
	    }
	    else {
		Output << Ntuple[Row][Column];
	    }
	}
	Output << endl;
    }
}
