/* KameTree.cc */
/* Created by Enomoto Sanshiro on 11 January 2011. */
/* Last updated by Enomoto Sanshiro on 26 April 2012. */

#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include <map>
#include <cctype>
#include "KameVariant.hh"
#include "KameTree.hh"

using namespace std;
using namespace kame;


TKameTree::TKameTree(void)
{
    _ParentNode = 0;
    _ReferenceCount = 0;
    _IsArrayContainer = false;
}

TKameTree::TKameTree(const TKameTree& Tree)
{
    _ParentNode = 0;
    _ReferenceCount = 0;
    _IsArrayContainer = false;

    Import(&Tree);
}

TKameTree& TKameTree::operator=(const TKameTree& Tree)
{
    if (&Tree != this) {
        Import(&Tree);
    }

    return *this;
}

TKameTree& TKameTree::operator=(const TKameTree::TEmptyArray& EmptyArray)
{
    Clear();
    _IsArrayContainer = true;

    return *this;
}

TKameTree::~TKameTree()
{
    if (_ReferenceCount > 0) {
        cerr << "### ERROR ###: deleting referenced tree: ";
        cerr << "  reference count = " << _ReferenceCount << endl;
    }

    for (unsigned i = 0; i < _ChildNodeList.size(); i++) {
        _ChildNodeList[i]->_ReferenceCount--;
        if (_ChildNodeList[i]->_ReferenceCount <= 0) {
            delete _ChildNodeList[i];
        }
    }
}

TKameTree& TKameTree::AppendNode(const std::string& Key)
{
    TKameTree* ChildNode = new TKameTree();
    Insert(Key, ChildNode);

    return *ChildNode;
}

void TKameTree::Link(TKameTree* Node, const std::string& Key)
{
    TKameTree* OldParentNode = Node->_ParentNode;
    Insert(Key.empty() ? Node->NodeName() : Key, Node);
    Node->_ParentNode = OldParentNode;
}

void TKameTree::Unlink(TKameTree* Node)
{
    for (unsigned i = 0; i < _ChildNodeList.size(); i++) {
        if (_ChildNodeList[i] == Node) {
            Node->_ReferenceCount--;
            if (Node->_ReferenceCount <= 0) {
                delete Node;
            }
            else if (Node->_ParentNode == this) {
                Node->_ParentNode = 0;
            }
            _ChildNodeList.erase(_ChildNodeList.begin() + i);
            _ChildNodeKeyList.erase(_ChildNodeKeyList.begin() + i);
            _ChildNodeKeyIndexTable.clear();
            for (unsigned j = 0; j < _ChildNodeList.size(); j++) {
                _ChildNodeKeyIndexTable[_ChildNodeKeyList[j]] = j;
            }
        }
        else if (_ChildNodeList[i]->_IsArrayContainer) {
            _ChildNodeList[i]->Unlink(Node);
        }
    }
}

TKameTree* TKameTree::ParentNode(void)
{
    if (_ParentNode == 0) {
        return 0;
    }
    if (_ParentNode->_IsArrayContainer) {
        return _ParentNode->ParentNode();
    }

    return _ParentNode;
}

const TKameTree* TKameTree::ParentNode(void) const
{
    if (_ParentNode == 0) {
        return 0;
    }
    if (_ParentNode->_IsArrayContainer) {
        return _ParentNode->ParentNode();
    }

    return _ParentNode;
}

void TKameTree::Insert(const std::string& Key, TKameTree* ChildNode)
{
    // if there are multiple entries with a same key, make the node an array //
    if ((! Key.empty()) && (_ChildNodeKeyIndexTable.count(Key) > 0)) {
        unsigned Index = _ChildNodeKeyIndexTable[Key];
        TKameTree* Node = _ChildNodeList[Index];

        // if the node is not an array, insert an array node //
        if (! Node->_IsArrayContainer) {
            TKameTree* ArrayNode = new TKameTree();
            _ChildNodeList[Index] = ArrayNode;
            ArrayNode->_ParentNode = this;
            ArrayNode->_ReferenceCount = 1;
            ArrayNode->_IsArrayContainer = true;
            ArrayNode->_ChildNodeList.push_back(Node);
            ArrayNode->_ChildNodeKeyList.push_back("");

            Node->_ParentNode = ArrayNode;
            Node = ArrayNode;
        }

        return Node->Insert("", ChildNode);
    }

    unsigned Index = _ChildNodeList.size();
    _ChildNodeList.push_back(ChildNode);
    _ChildNodeKeyList.push_back(Key);
    _ChildNodeKeyIndexTable[Key] = Index;

    ChildNode->_ParentNode = this;
    ChildNode->_ReferenceCount++;
}

TKameTree& TKameTree::Find(const std::string& Path)
{
    if (Path.empty()) {
	return *this;
    }
    if (Path[0] == '/') {
        return Find(Path.substr(1));
    }

    if (Path == ".") {
	return *this;
    }
    if (Path == "..") {
        TKameTree* parent = ParentNode();
        if (! parent) {
	    throw TKameException("invalid tree path: " + Path);
        }
	return *parent;
    }

    if (Path[0] == '[') {
	string::size_type End = Path.find_first_of(']');
	if (End == string::npos) {
	    throw TKameException("invalid tree path: " + Path);
	}

	istringstream is(Path.substr(1, End-1));
	unsigned Index;
	if (!(is >> Index)) {
	    throw TKameException("invalid tree path: " + Path);
	}
	string Excess;
	if (is >> Excess) {
	    throw TKameException("invalid tree path: " + Path);
	}
        return FindArrayElement(Index).Find(Path.substr(End+1));
    }

    if (_IsArrayContainer) {
        // key access to array: assume first array element //
        if (_ChildNodeList.empty()) {
            AppendNode("");
        }
        return _ChildNodeList[0]->Find(Path);
    }

    string::size_type End;
    if (Path[0] == '@') {
        if (Path.size() > 1) {
            End = Path.substr(1).find_first_of("/[@");
            if (End != string::npos) {
                End += 1;  // for substr(1)
            }
        }
        else {
            End = 1;
        }
    }
    else {
        End = Path.find_first_of("/[@");
    }
    if (End != string::npos) {
        return Find(Path.substr(0, End)).Find(Path.substr(End));
    }
        
    std::map<std::string, unsigned>::const_iterator Entry = (
        _ChildNodeKeyIndexTable.find(Path)
    );
    if (Entry == _ChildNodeKeyIndexTable.end()) {
        return AppendNode(Path);
    }
    else {
        return *_ChildNodeList[Entry->second];
    }
}

const TKameTree& TKameTree::Find(const std::string& Path) const
{
    // const version //
    //... TODO: avoid duplication to the non-const version //
    
    static TKameTree voidNode;

    if (Path.empty()) {
	return *this;
    }
    if (Path[0] == '/') {
        return Find(Path.substr(1));
    }

    if (Path == ".") {
	return *this;
    }
    if (Path == "..") {
        const TKameTree* parent = ParentNode();
        if (! parent) {
	    throw TKameException("invalid tree path: " + Path);
        }
	return *parent;
    }

    if (Path[0] == '[') {
	string::size_type End = Path.find_first_of(']');
	if (End == string::npos) {
	    throw TKameException("invalid tree path: " + Path);
	}

	istringstream is(Path.substr(1, End-1));
	unsigned Index;
	if (!(is >> Index)) {
	    throw TKameException("invalid tree path: " + Path);
	}
	string Excess;
	if (is >> Excess) {
	    throw TKameException("invalid tree path: " + Path);
	}
        return FindArrayElement(Index).Find(Path.substr(End+1));
    }

    if (_IsArrayContainer) {
        // key access to array: assume first array element //
        if (_ChildNodeList.empty()) {
            return voidNode;
        }
        return _ChildNodeList[0]->Find(Path);
    }

    string::size_type End;
    if (Path[0] == '@') {
        if (Path.size() > 1) {
            End = Path.substr(1).find_first_of("/[@");
            if (End != string::npos) {
                End += 1;  // for substr(1)
            }
        }
        else {
            End = 1;
        }
    }
    else {
        End = Path.find_first_of("/[@");
    }
    if (End != string::npos) {
        return Find(Path.substr(0, End)).Find(Path.substr(End));
    }
        
    std::map<std::string, unsigned>::const_iterator Entry = (
        _ChildNodeKeyIndexTable.find(Path)
    );
    if (Entry == _ChildNodeKeyIndexTable.end()) {
        return voidNode;
    }
    else {
        return *_ChildNodeList[Entry->second];
    }
}

TKameTree& TKameTree::FindArrayElement(unsigned ArrayIndex)
{
    // if children are in an array, returns Index-th child element
    // if not, if Index == 0, returns this, else creates children array

    if (! _IsArrayContainer) {
        if(ArrayIndex == 0) {
            return *this;
        }

        TKameTree* ElementNode = new TKameTree(*this);
        ElementNode->_ParentNode = this;
        ElementNode->_ReferenceCount = 1;
        ElementNode->_IsArrayContainer = false;

        _ChildNodeList.clear();
        _ChildNodeKeyList.clear();
        _ChildNodeKeyIndexTable.clear();
        _Value = TKameVariant();

        _ChildNodeList.push_back(ElementNode);
        _ChildNodeKeyList.push_back("");
        _IsArrayContainer = true;
    }

    while (ArrayIndex >= _ChildNodeList.size()) {
        AppendNode("");
    }

    return *_ChildNodeList[ArrayIndex];
}

const TKameTree& TKameTree::FindArrayElement(unsigned ArrayIndex) const
{
    // if children are in an array, returns Index-th child element
    // if not, if Index == 0, returns this, else creates children array

    static TKameTree voidNode;

    if (! _IsArrayContainer) {
        if(ArrayIndex == 0) {
            return *this;
        }
        else {
            return voidNode;
        }
    }

    if (ArrayIndex >= _ChildNodeList.size()) {
        return voidNode;
    }

    return *_ChildNodeList[ArrayIndex];
}

std::string TKameTree::NodeName(void) const
{
    if (! _ParentNode) {
	return "";
    }

    if (_ParentNode->_IsArrayContainer) {
        return _ParentNode->NodeName();
    }

    for (unsigned i = 0; i < _ParentNode->_ChildNodeList.size(); i++) {
        if (_ParentNode->_ChildNodeList[i] == this) {
            return _ParentNode->_ChildNodeKeyList[i];
        }
    }

    return "";
}

std::string TKameTree::NodePath(void) const
{
    if (_ParentNode) {
	return _ParentNode->NodePathOf(this);
    }
    else {
	return "/";
    }
}

std::string TKameTree::NodePathOf(const TKameTree* ChildNode) const
{
    string NodeName;
    for (unsigned i = 0; i < _ChildNodeList.size(); i++) {
	if (_ChildNodeList[i] == ChildNode) {
	    string Key = _ChildNodeKeyList[i];
	    if (! Key.empty()) {
                if (Key[0] == '@') {
                    NodeName = Key;
                }
                else {
                    NodeName = "/" + Key;
                }
	    }
	    else {
		ostringstream ArrayIndex;
		ArrayIndex << "[" << i << "]";
		NodeName = ArrayIndex.str();
	    }
	}
    }
    if (NodeName.empty()) {
        return "";
    }

    if (_ParentNode) {
	return _ParentNode->NodePathOf(this) + NodeName;
    }
    else {
        return NodeName;
    }
}



void TKameTree::Clear(void)
{
    for (unsigned i = 0; i < _ChildNodeList.size(); i++) {
        _ChildNodeList[i]->_ReferenceCount--;
        if (_ChildNodeList[i]->_ReferenceCount <= 0) {
            delete _ChildNodeList[i];
        }
    }
    _ChildNodeList.clear();
    _ChildNodeKeyList.clear();
    _ChildNodeKeyIndexTable.clear();
    
    _Value = TKameVariant();
    _IsArrayContainer = false;
}

void TKameTree::Import(const TKameTree* Tree)
{
    Clear();

    for (unsigned i = 0; i < Tree->_ChildNodeList.size(); i++) {
	TKameTree* Child = new TKameTree();
	Child->_ParentNode = this;
	Child->_ReferenceCount = 1;
	Child->Import(Tree->_ChildNodeList[i]);
	_ChildNodeList.push_back(Child);
	_ChildNodeKeyList.push_back(Tree->_ChildNodeKeyList[i]);
        _ChildNodeKeyIndexTable[Tree->_ChildNodeKeyList[i]] = i;
    }

    _Value = Tree->_Value;
    _IsArrayContainer = Tree->_IsArrayContainer;
}

void TKameTree::Dump(std::ostream& os)
{
    if (! _Value.IsVoid()) {
        os << NodePath() << ": ";
        if (_Value.IsString()) {
            string Str = _Value.AsString();
            if (Str.empty()) {
                os << "\"\"";
            }
            else if (isspace(*Str.begin()) || isspace(*Str.end())) {
                os << '"' << Str << '"';
            }
            else {
                os << Str;
            }
        }
        else {
            os << _Value;
        }
        os << endl;
    }

    for (unsigned i = 0; i < _ChildNodeList.size(); i++) {
        _ChildNodeList[i]->Dump(os);
    }
}


// depreciated //
std::vector<std::string> TKameTree::NodePathList(void) const
{
    vector<string> List;
    if ((! _Value.IsVoid()) && ! (_Value.As<std::string>().empty())) {
	List.push_back("");
    }
    for (unsigned i = 0; i < _ChildNodeList.size(); i++) {
	string Key = _ChildNodeKeyList[i];
	string NodeName;
	if (! Key.empty()) {
	    NodeName = Key;
	}
	else {
	    ostringstream ArrayPath;
	    ArrayPath << "[" << i << "]";
	    NodeName = ArrayPath.str();
	}
	vector<string> ChildList = _ChildNodeList[i]->NodePathList();
	for (unsigned j = 0; j < ChildList.size(); j++) {
	    if (ChildList[j].empty()) {
		List.push_back(NodeName);
	    }
	    else {
                if (ChildList[j][0] == '[') {
                    List.push_back(NodeName + ChildList[j]);
                }
                else {
                    List.push_back(NodeName + "/" + ChildList[j]);
                }
	    }
	}
    }

    return List;
}
