/* KameTree.hh */
/* Created by Enomoto Sanshiro on 11 January 2011. */
/* Last updated by Enomoto Sanshiro on 26 April 2012. */


#ifndef __KameTree_hh__
#define __KameTree_hh__

#include <iostream>
#include <string>
#include <vector>
#include <map>
#include "KameVariant.hh"


namespace kame {


class TKameTree {
  public:
    class TEmptyArray { public: virtual ~TEmptyArray(){} };
  public:
    TKameTree(void);
    TKameTree(const TKameTree& Tree);
    virtual TKameTree& operator=(const TKameTree& Tree);
    virtual TKameTree& operator=(const TEmptyArray& EmptyArray);
    virtual ~TKameTree();
  public:
    // node/leaf access //
    virtual std::string NodeName(void) const;
    virtual std::string NodePath(void) const;
    inline TKameVariant& Value(void);
    inline const TKameVariant& Value(void) const;
    //template<typename T> inline operator T() const;
    template<typename T> inline T As(void) const;
    inline TKameTree& operator=(const TKameVariant& NewValue);
    inline bool IsVoid(void) const;
    inline bool IsLeaf(void) const;
    inline bool IsArray(void) const;
    inline bool IsEmpty(void) const;
    inline TKameVariant Or(const TKameVariant& DefaultValue) const;
  public:
    // node iteration //
    inline size_t Length(void) const;
    inline TKameTree& operator[](unsigned Index);
    inline TKameTree& operator[](const std::string& Key);
    inline const TKameTree& operator[](unsigned Index) const;
    inline const TKameTree& operator[](const std::string& Key) const;
    inline const std::vector<std::string>& KeyList(void) const;
    inline const std::vector<TKameTree*>& ChildNodeList(void);
    virtual TKameTree* ParentNode(void);
    virtual const TKameTree* ParentNode(void) const;
  public:
    // node insertion //
    virtual TKameTree& AppendNode(const std::string& Key);
    virtual void Link(TKameTree* Node, const std::string& Key = "");
    virtual void Unlink(TKameTree* Node);
    inline TKameTree& operator()(const std::string& key, const TKameVariant& variant);
    inline TKameTree& operator()(const std::string& key, const TKameTree& tree);
  public:
    // misc //
    virtual void Dump(std::ostream& os = std::cout);
  public:
    // depreciated //
    virtual std::vector<std::string> NodePathList(void) const;
  protected:
    virtual void Insert(const std::string& Key, TKameTree* Node);
    virtual TKameTree& Find(const std::string& Path);
    virtual const TKameTree& Find(const std::string& Path) const;
    virtual TKameTree& FindArrayElement(unsigned Index);
    virtual const TKameTree& FindArrayElement(unsigned Index) const;
    virtual std::string NodePathOf(const TKameTree* ChildNode) const;
    virtual void Clear(void);
    virtual void Import(const TKameTree* Tree);
  private:
    TKameVariant _Value;
    TKameTree* _ParentNode;
    int _ReferenceCount;
    bool _IsArrayContainer;
    std::vector<TKameTree*> _ChildNodeList;
    std::vector<std::string> _ChildNodeKeyList;
    std::map<std::string, unsigned> _ChildNodeKeyIndexTable;
};



template<typename T> struct TKameTreeTraits {
    static T As(const TKameTree& Tree) {
        try {
            return Tree.Value().As<T>();
        }
        catch (TKameException &e) {
            throw e;
        }
    }
};



inline TKameVariant& TKameTree::Value(void)
{
    if (_IsArrayContainer && ! _ChildNodeList.empty()) {
        return _ChildNodeList[0]->Value();
    }

    return _Value;
}

inline const TKameVariant& TKameTree::Value(void) const
{
    if (_IsArrayContainer && ! _ChildNodeList.empty()) {
        return _ChildNodeList[0]->Value();
    }

    return _Value;
}

#if 0
template<typename T> inline TKameTree::operator T() const
{
    return kame::TKameTreeTraits<T>::As(*this);
}
#endif

template<typename T> inline T TKameTree::As(void) const
{
    return kame::TKameTreeTraits<T>::As(*this);
}

inline TKameTree& TKameTree::operator=(const TKameVariant& NewValue)
{
    this->Value() = NewValue;
    return *this;
}

inline bool TKameTree::IsVoid(void) const
{
    return Value().IsVoid();
}

inline bool TKameTree::IsLeaf(void) const
{
    return (! _IsArrayContainer) && _ChildNodeList.empty();
}

inline bool TKameTree::IsArray(void) const
{
    return _IsArrayContainer;
}

inline bool TKameTree::IsEmpty(void) const
{
    if (! Value().IsVoid()) {
        return false;
    }
    for (unsigned i = 0; i < _ChildNodeList.size(); i++) {
        if (! _ChildNodeList[i]->IsEmpty()) {
            return false;
        }
    }
    return true;
}

inline TKameVariant TKameTree::Or(const TKameVariant& DefaultValue) const
{
    return IsVoid() ? DefaultValue : Value();
}

inline size_t TKameTree::Length(void) const
{
    if (IsArray()) {
        return _ChildNodeList.size();
    }
    else if (IsLeaf() && IsVoid()) {
        return 0;
    }
    else {
        return 1;
    }
}

inline TKameTree& TKameTree::operator[](unsigned Index)
{
    return FindArrayElement(Index);
}

inline TKameTree& TKameTree::operator[](const std::string& Key)
{
    return Find(Key);
}

inline const TKameTree& TKameTree::operator[](unsigned Index) const
{
    return FindArrayElement(Index);
}

inline const TKameTree& TKameTree::operator[](const std::string& Key) const
{
    return Find(Key);
}

inline const std::vector<std::string>& TKameTree::KeyList(void) const
{
    return _ChildNodeKeyList;
}

inline const std::vector<TKameTree*>& TKameTree::ChildNodeList(void)
{
    return _ChildNodeList;
}


inline std::ostream& operator<<(std::ostream&os, const TKameTree& Tree) 
{
    return os << Tree.Value();
}

inline TKameTree& TKameTree::operator()(const std::string& Key, const TKameVariant& Value)
{
    this->AppendNode(Key).Value() = Value;
    return *this;
}

inline TKameTree& TKameTree::operator()(const std::string& Key, const TKameTree& Tree)
{
    this->AppendNode(Key).Import(&Tree);
    return *this;
}

inline TKameTree make_tree(const std::string& Key, const TKameVariant& Value)
{
    return TKameTree()(Key, Value);
}

inline TKameTree make_tree(const std::string& Key, const TKameTree& Tree)
{
    return TKameTree()(Key, Tree);
}


}

// suger for convenient includes //
#include "KameTreeWalker.hh"
#include "KameTreeSerializer.hh"

#endif
