/* KameTreeWalker.cc */
/* Created by Enomoto Sanshiro on 9 April 2012. */
/* Last updated by Enomoto Sanshiro on 9 April 2012. */


#include <string>
#include <iostream>
#include <sstream>
#include <cctype>
#include <deque>
#include "KameTree.hh"
#include "KameTreeWalker.hh"
#include "KameTreeSerializer.hh"

using namespace std;
using namespace kame;



TKameTreeSerializer::TKameTreeSerializer(std::ostream& Output)
: _Output(Output)
{
}

TKameTreeSerializer::~TKameTreeSerializer()
{
}

void TKameTreeSerializer::Serialize(TKameTree& Tree, const std::string& Name)
{
    TKameTreeWalker(this).Process(Tree, Name);
}



TKameKtfTreeSerializer::TKameKtfTreeSerializer(std::ostream& Output)
: TKameTreeSerializer(Output)
{
    _HeaderChar = '#';
    _IndentStep = 2;

    _Depth = -1;
}

TKameKtfTreeSerializer::~TKameKtfTreeSerializer(void)
{
}

bool TKameKtfTreeSerializer::StartTree(TKameTree& Node)
{
    _Depth = -1;
    return true;
}

bool TKameKtfTreeSerializer::StartNode(const std::string& Name, TKameTree& Node, TKameTree& AttributeList)
{
    _Depth++;
    if (_Depth <= 0) {
        return true;
    }

    string Indent(_IndentStep * _Depth, ' ');
    Indent[0] = _HeaderChar;
    
    _Output << Indent << Name << ": ";

    istringstream is(Node.As<string>());
    string Line;
    if (getline(is, Line) && (! Line.empty())) {
        if (isspace(*Line.begin()) || isspace(*Line.end())) {
            _Output << '"' << Line << '"';
        }
        else {
            _Output << Line;
        }
        _Output << endl;
    }
    else if (Node.Value().IsString()) {
        _Output << "\"\"" << endl;
    }
    else {
        _Output << endl;
    }

    while(getline(is, Line)) {
        _Output << Indent << ": " << Line << endl;
    }

    string AttributeIndent(_IndentStep * (_Depth+1), ' ');
    AttributeIndent[0] = _HeaderChar;
    for (unsigned i = 0; i < AttributeList.ChildNodeList().size(); i++) {
        _Output << AttributeIndent;
        _Output << "@" << AttributeList.ChildNodeList()[i]->NodeName() << ": ";
        _Output << AttributeList.ChildNodeList()[i]->Value() << endl;
    }

    return true;
}

void TKameKtfTreeSerializer::EndNode(const std::string& Name, TKameTree& Node)
{
    _Depth--;
}

bool TKameKtfTreeSerializer::StartArray(const std::string& Name, TKameTree& Node)
{
    if (Node.Length() == 0) {
        string Indent(_IndentStep * (_Depth+1), ' ');
        Indent[0] = _HeaderChar;
        _Output << Indent << Name << ": []" << endl;
        return false;
    }

    return true;
}



TKameJsonTreeSerializer::TKameJsonTreeSerializer(std::ostream& Output)
: TKameTreeSerializer(Output)
{
}

TKameJsonTreeSerializer::~TKameJsonTreeSerializer(void)
{
}

bool TKameJsonTreeSerializer::StartTree(TKameTree& Node)
{
    this->ProcessNode(Node, "", "");
    _Output << endl;

    return false;
}

void TKameJsonTreeSerializer::ProcessNode(TKameTree& Node, std::string NodeName, const std::string& Indent)
{
    //... BUG: if the Node has both value and children, the value is ignored

    _Output << Indent;
    if (! NodeName.empty()) {
        _Output << "\"" << NodeName << "\": ";
    }

    if (Node.IsLeaf()) {
        _Output << "\"";
        this->WriteJsonString(_Output, Node.As<string>());
        _Output << "\"";
    }
    else if (Node.IsArray()) {
        unsigned Length = Node.Length();
        if (Length == 0) {
            _Output << "[]";
        }
        else {
            _Output << "[" << endl;
            for (unsigned i = 0; i < Length; i++) {
                ProcessNode(Node[i], "", Indent + "  ");
                if (i+1 < Length) {
                    _Output << ',' << endl;
                }
                else {
                    _Output << endl;
                }
            }
            _Output << Indent << "]";
        }
    }
    else {
        unsigned Size = Node.KeyList().size();
        _Output << "{" << endl;
        for (unsigned i = 0; i < Size; i++) {
            string Key = Node.KeyList()[i];
            ProcessNode(Node[Key], Key, Indent + "  ");
            if (i+1 < Size) {
                _Output << ',' << endl;
            }
            else {
                _Output << endl;
            }
        }
        _Output << Indent << "}";
    }
}

void TKameJsonTreeSerializer::WriteJsonString(std::ostream& os, const std::string& Value)
{
    for (string::const_iterator Ch = Value.begin(); Ch != Value.end(); Ch++) {
        switch (*Ch) {
          case '"':
            os << "\\\"";
            break;
          case '\t':
            os << "\\t";
            break;
          case '\v':
            os << "\\v";
            break;
          case '\n':
            os << "\\n";
            break;
          case '\r':
            os << "\\r";
            break;
          case '\f':
            os << "\\f";
            break;
          case '\\':
            os << "\\\\";
            break;
        default:
            os << *Ch;
            break;
        }
    }
}



TKameXmlTreeSerializer::TKameXmlTreeSerializer(std::ostream& Output)
: TKameTreeSerializer(Output)
{
    _NameForEmptyName = "Node";
}

TKameXmlTreeSerializer::~TKameXmlTreeSerializer(void)
{
}

bool TKameXmlTreeSerializer::StartTree(TKameTree& Node)
{
    _Output << "<?xml version=\"1.0\"?>" << endl << endl;
    _IndentStack.push_back("");
    return true;
}

bool TKameXmlTreeSerializer::StartNode(const std::string& Name, TKameTree& Node, TKameTree& AttributeList)
{
    string Indent = _IndentStack.back();
    _IndentStack.push_back(Indent + "  ");

    _Output << Indent;
    _Output << "<" << (Name.empty() ? _NameForEmptyName : Name );
    for (unsigned i = 0; i < AttributeList.ChildNodeList().size(); i++) {
        _Output << " " << AttributeList.ChildNodeList()[i]->NodeName();
        _Output << "=\"";
        WriteXmlString(_Output, AttributeList.ChildNodeList()[i]->As<string>());
        _Output << "\"";
    }
    _Output << ">";

    if (! Node.IsVoid()) {
        WriteXmlString(_Output, Node.As<string>());
    }
    if (! Node.IsLeaf()) {
        _Output << endl;
    }
    
    return true;
}

void TKameXmlTreeSerializer::EndNode(const std::string& Name, TKameTree& Node)
{
    _IndentStack.pop_back();
    if (! Node.IsLeaf()) {
        _Output << _IndentStack.back();
    }
    _Output << "</" << (Name.empty() ? _NameForEmptyName : Name ) << ">";
    _Output << endl;
}

bool TKameXmlTreeSerializer::StartArray(const std::string& Name, TKameTree& Node)
{
    if (Node.Length() == 0) {
        // empty array using inline JSON //
        string Indent = _IndentStack.back() + "  ";
        _Output << Indent;
        _Output << "<" << Name << ">[]</" << Name << ">" << endl;
        return false;
    }

    return true;
}

void TKameXmlTreeSerializer::WriteXmlString(std::ostream& os, const std::string& Value)
{
    for (string::const_iterator Ch = Value.begin(); Ch != Value.end(); Ch++) {
        switch (*Ch) {
          case '<':
            os << "&lt;";
            break;
          case '>':
            os << "&gt;";
            break;
          case '&':
            os << "&amp;";
            break;
          case '"':
            os << "&quot;";
            break;
          case '\'':
            os << "&apos;";
            break;
          default:
            os << *Ch;
            break;
        }
    }
}



TKamePlistTreeSerializer::TKamePlistTreeSerializer(std::ostream& Output)
: TKameTreeSerializer(Output)
{
}

TKamePlistTreeSerializer::~TKamePlistTreeSerializer(void)
{
}

bool TKamePlistTreeSerializer::StartTree(TKameTree& Node)
{
    _Output << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << endl;
    _Output << "<!DOCTYPE plist PUBLIC \"-//Apple//DTD PLIST 1.0//EN\" \"http://www.apple.com/DTDs/PropertyList-1.0.dtd\">" << endl;

    _Output << "<plist version=\"1.0\">" << endl;
    this->ProcessNode(Node, "", "");
    _Output << "</plist>" << endl;

    return false;
}

void TKamePlistTreeSerializer::ProcessNode(TKameTree& Node, std::string NodeName, const std::string& Indent)
{
    //... BUG: if the Node has both value and children, the value is ignored

    if (! NodeName.empty()) {
        _Output << Indent << "<key>" << NodeName << "<key>" << endl;
    }

    if (Node.IsLeaf()) {
        _Output << Indent;
        if (Node.IsVoid()) {
            _Output << "<null/>";
        }
        else if (Node.Value().IsBool()) {
            _Output << (Node.As<bool>() ? "<true/>" : "<false/>");
        }
        else if (Node.Value().IsInteger()) {
            _Output << "<integer>" << Node << "</integer>";
        }
        else if (Node.Value().IsNumeric()) {
            _Output << "<real>" << Node << "</real>";
        }
        else {
            _Output << "<string>";
            this->WritePlistString(_Output, Node.As<string>());
            _Output << "</string>";
        }
        _Output << endl;
    }
    else if (Node.IsArray()) {
        _Output << Indent << "<array>" << endl;
        for (unsigned i = 0; i < Node.Length(); i++) {
            ProcessNode(Node[i], "", Indent + "    ");
        }
        _Output << Indent << "</array>" << endl;
    }
    else {
        unsigned Size = Node.KeyList().size();
        _Output << Indent << "<dict>" << endl;
        for (unsigned i = 0; i < Size; i++) {
            string Key = Node.KeyList()[i];
            ProcessNode(Node[Key], Key, Indent + "    ");
        }
        _Output << Indent << "</dict>" << endl;
    }
}

void TKamePlistTreeSerializer::WritePlistString(std::ostream& os, const std::string& Value)
{
    for (string::const_iterator Ch = Value.begin(); Ch != Value.end(); Ch++) {
        switch (*Ch) {
          case '<':
            os << "&lt;";
            break;
          case '>':
            os << "&gt;";
            break;
          case '&':
            os << "&amp;";
            break;
          case '"':
            os << "&quot;";
            break;
          case '\'':
            os << "&apos;";
            break;
          default:
            os << *Ch;
            break;
        }
    }
}
