/* KameVariant.hh */
/* Created by Enomoto Sanshiro on 29 June 2009. */
/* Last updated by Enomoto Sanshiro on 29 June 2009. */


#ifndef __KameVariant_hh__
#define __KameVariant_hh__

#include <string>
#include <iostream>
#include <sstream>
#include "KameDefs.hh"


// "variant" is not "any" as in boost, because it does not support
// general types as a template, but limits to primitive types.
// In return, variant provides non-explicit constructors and
// type-cast operators which make conversions invisible. 
// (If you do this with a template, it will generate a global
// type convesion path between any types, and can destroy the 
// C++ conversion rules.)
//
// "variant" uses "union" instead of dynamically allocated templated 
// type holder object (like "any" in boost), which will be an advantage 
// for large scale data table etc. 


namespace kame {


class TKameUnknown {
  public:
    // use RTTI to get actual type //
    TKameUnknown(void) {}
    virtual ~TKameUnknown() {}
    virtual TKameUnknown* Clone(void) const = 0;
};


class TKameVariant {
  public:
    inline TKameVariant(void);
    inline TKameVariant(bool Value);
    inline TKameVariant(int Value);
    inline TKameVariant(unsigned int Value);
    inline TKameVariant(long Value);
    inline TKameVariant(unsigned long Value);
    inline TKameVariant(float Value);
    inline TKameVariant(double Value);
    inline TKameVariant(const std::string& Value);
    inline TKameVariant(const char* Value);
    inline TKameVariant(const TKameUnknown& Value);
    inline TKameVariant(const TKameVariant& Value);
    inline ~TKameVariant();
    inline TKameVariant& operator=(const TKameVariant& Value);
    inline void Assign(const TKameVariant& Value);
  public:
    template<typename T> inline T As(void) const;
    inline TKameVariant Or(const TKameVariant& DefaultValue) const;
    inline operator bool() const;
    inline operator int() const;
    inline operator unsigned int() const;
    inline operator long() const;
    inline operator unsigned long() const;
    inline operator float() const;
    inline operator double() const;
    inline operator std::string() const;
    inline operator const TKameUnknown&() const;
  public:
    inline bool IsVoid(void) const;
    inline bool IsBool(void) const;
    inline bool IsInteger(void) const;
    inline bool IsNumeric(void) const;
    inline bool IsString(void) const;
    inline bool IsUnknown(void) const;
    bool AsBool(void) const;
    long AsLong(void) const;
    double AsDouble(void) const;
    std::string AsString(void) const;
    TKameUnknown& AsUnknown(void);
    const TKameUnknown& AsUnknown(void) const;
  private:
    enum TValueType { 
	Type_Void, Type_Bool, Type_Long, Type_Double, Type_String, Type_Unknown
    } _Type;
    union TPrimitive {
        bool _BoolValue;
        long _LongValue;
        double _DoubleValue;
	std::string* _StringValue;
	TKameUnknown* _UnknownValue;
    } _Primitive;
};



template<typename T>
struct TKameVariantTraits {
  // explicit partial instantiations only //
  private:
    static T As(const TKameVariant& Value) { return T(); }
};

template<> struct TKameVariantTraits<void> {
    static void As(const TKameVariant& Value) {}
};

template<> struct TKameVariantTraits<bool> {
    static bool As(const TKameVariant& Value) {
        return Value.AsBool();
    }
};

template<> struct TKameVariantTraits<int> {
    static int As(const TKameVariant& Value) {
        return Value.AsLong();
    }
};

template<> struct TKameVariantTraits<unsigned int> {
    static unsigned int As(const TKameVariant& Value) {
        return Value.AsLong();
    }
};

template<> struct TKameVariantTraits<long> {
    static long As(const TKameVariant& Value) {
        return Value.AsLong();
    }
};

template<> struct TKameVariantTraits<unsigned long> {
    static unsigned long As(const TKameVariant& Value) {
        return Value.AsLong();
    }
};

template<> struct TKameVariantTraits<float> {
    static float As(const TKameVariant& Value) {
        return Value.AsDouble();
    }
};

template<> struct TKameVariantTraits<double> {
    static double As(const TKameVariant& Value) {
        return Value.AsDouble();
    }
};

template<> struct TKameVariantTraits<std::string> {
    static std::string As(const TKameVariant& Value) {
        return Value.AsString();
    }
};

template<> struct TKameVariantTraits<const char*> {
    static std::string As(const TKameVariant& Value) {
        return Value.AsString().c_str();
    }
};

template<> struct TKameVariantTraits<const TKameUnknown&> {
    static const TKameUnknown& As(const TKameVariant& Value) {
        return Value.AsUnknown();
    }
};



TKameVariant::TKameVariant(void)
: _Type(Type_Void)
{
}

TKameVariant::TKameVariant(bool Value)
: _Type(Type_Bool)
{
    _Primitive._BoolValue = Value;
}

TKameVariant::TKameVariant(int Value)
: _Type(Type_Long)
{
    _Primitive._LongValue = Value;
}

TKameVariant::TKameVariant(unsigned int Value)
: _Type(Type_Long)
{
    _Primitive._LongValue = Value;
}

TKameVariant::TKameVariant(long Value)
: _Type(Type_Long)
{
    _Primitive._LongValue = Value;
}

TKameVariant::TKameVariant(unsigned long Value)
: _Type(Type_Long)
{
    _Primitive._LongValue = Value;
}

TKameVariant::TKameVariant(float Value)
: _Type(Type_Double)
{
    _Primitive._DoubleValue = Value;
}

TKameVariant::TKameVariant(double Value)
: _Type(Type_Double)
{
    _Primitive._DoubleValue = Value;
}

TKameVariant::TKameVariant(const std::string& Value)
: _Type(Type_String)
{
    _Primitive._StringValue = new std::string(Value);
}

TKameVariant::TKameVariant(const char* Value)
: _Type(Type_String)
{
    _Primitive._StringValue = new std::string(Value);
}

TKameVariant::TKameVariant(const TKameUnknown& Value)
: _Type(Type_Unknown)
{
    _Primitive._UnknownValue = Value.Clone();
}

TKameVariant::TKameVariant(const TKameVariant& Value)
: _Type(Value._Type), _Primitive(Value._Primitive)
{
    if (_Type == Type_String) {
	_Primitive._StringValue = (
	    new std::string(*Value._Primitive._StringValue)
	);
    }
    else if (_Type == Type_Unknown) {
	_Primitive._UnknownValue = Value._Primitive._UnknownValue->Clone();
    }
}

TKameVariant::~TKameVariant()
{
    if (_Type == Type_String) {
	delete _Primitive._StringValue;
    }
    else if (_Type == Type_Unknown) {
	delete _Primitive._UnknownValue;
    }
}

TKameVariant& TKameVariant::operator=(const TKameVariant& Value)
{
    if (&Value == this) {
	return *this;
    }

    if (_Type == Type_String) {
	delete _Primitive._StringValue;
    }
    else if (_Type == Type_Unknown) {
	delete _Primitive._UnknownValue;
    }

    if (Value._Type == Type_String) {
	_Primitive._StringValue = (
	    new std::string(*Value._Primitive._StringValue)
	);
    }
    else if (Value._Type == Type_Unknown) {
	_Primitive._UnknownValue = Value._Primitive._UnknownValue->Clone();
    }
    else {
	_Primitive = Value._Primitive;
    }
    _Type = Value._Type;

    return *this;
}

void TKameVariant::Assign(const TKameVariant& Value)
{
    if (_Type == Type_Void) {
	throw TKameException("assignment to void variable not allowed");
    }
    else if (_Type == Type_Bool) {
        _Primitive._BoolValue = Value.AsBool();
    }
    else if (_Type == Type_Long) {
        _Primitive._LongValue = Value.AsLong();
    }
    else if (_Type == Type_Double) {
        _Primitive._DoubleValue = Value.AsDouble();
    }
    else if (_Type == Type_String) {
        *(_Primitive._StringValue) = Value.AsString();
    }
    else {
	throw TKameException("assignment to Unknown variable not allowed");
    }
}

template <typename T> 
inline T TKameVariant::As(void) const
{
    return kame::TKameVariantTraits<T>::As(*this);
}

inline TKameVariant TKameVariant::Or(const TKameVariant& DefaultValue) const
{
    if (_Type == Type_Void) {
        return DefaultValue;
    }
    else {
        return *this;
    }
}

TKameVariant::operator bool() const
{
    return As<bool>();
}

TKameVariant::operator int() const
{
    return As<int>();
}

TKameVariant::operator unsigned int() const
{
    return As<unsigned int>();
}

TKameVariant::operator long() const
{
    return As<long>();
}

TKameVariant::operator unsigned long() const
{
    return As<unsigned long>();
}

TKameVariant::operator float() const
{
    return As<float>();
}

TKameVariant::operator double() const
{
    return As<double>();
}

TKameVariant::operator std::string() const
{
    return As<std::string>();
}

TKameVariant::operator const TKameUnknown&() const
{
    return As<const TKameUnknown&>();
}

bool TKameVariant::IsVoid(void) const
{
    return (_Type == Type_Void);
}

bool TKameVariant::IsBool(void) const
{
    return (_Type == Type_Bool);
}

bool TKameVariant::IsInteger(void) const
{
    return (_Type == Type_Long);
}

bool TKameVariant::IsNumeric(void) const
{
    return ((_Type == Type_Long) || (_Type == Type_Double));
}

bool TKameVariant::IsString(void) const
{
    return (_Type == Type_String);
}

bool TKameVariant::IsUnknown(void) const
{
    return (_Type == Type_Unknown);
}
    


inline std::ostream& operator<<(std::ostream& os, const TKameVariant& Value) 
{
    if (Value.IsInteger()) {
	os << (long) Value;
    }
    else if (Value.IsNumeric()) {
	os << (double) Value;
    }
    else {
	os << (const std::string&) Value;
    }

    return os;
}



#if 1
// do not use const TKameVariant& for the following operator overloading
// otherwise the operators will be ambiguous (e.g., enum + int)

template <typename T>
inline TKameVariant operator+(TKameVariant& Left, const T& Right)
{
    if (Left.IsInteger()) {
        return TKameVariant(Left.As<long>() + Right);
    }
    else if (Left.IsNumeric()) {
        return TKameVariant(Left.As<double>() + Right);
    }
    else if (Left.IsString()) {
        return TKameVariant(
            Left.As<std::string>() + TKameVariant(Right).As<std::string>()
        );
    }

    return TKameVariant(Left.As<T>() + Right);
}

template <typename T>
inline TKameVariant operator+(const T& Left, TKameVariant& Right)
{
    if (Right.IsInteger()) {
        return TKameVariant(Left + Right.As<long>());
    }
    else if (Right.IsNumeric()) {
        return TKameVariant(Left + Right.As<double>());
    }
    else if (Right.IsString()) {
        return TKameVariant(
            TKameVariant(Left).As<std::string>() + Right.As<std::string>()
        );
    }

    return TKameVariant(Left + Right.As<T>());
}

inline TKameVariant operator+(TKameVariant& Left, TKameVariant& Right)
{
    if (Right.IsInteger()) {
        return operator+(Left, Right.As<long>());
    }
    else if (Right.IsNumeric()) {
        return operator+(Left, Right.As<double>());
    }
    else {
        return TKameVariant(Left.As<std::string>() + Right.As<std::string>());
    }
}

inline TKameVariant operator+(TKameVariant& Left, const std::string& Right) 
{
    return TKameVariant(Left.As<std::string>() + Right);
}

inline TKameVariant operator+(const std::string& Left, TKameVariant& Right)
{
    return TKameVariant(Left + Right.As<std::string>());
}

inline TKameVariant operator+(TKameVariant& Left, const char Right[]) 
{
    return TKameVariant(Left.As<std::string>() + std::string(Right));
}

inline TKameVariant operator+(const char Left[], TKameVariant& Right)
{
    return TKameVariant(std::string(Left) + Right.As<std::string>());
}

template <typename T>
inline TKameVariant& operator+=(TKameVariant& This, const T& Value)
{
    return This = operator+(This, Value);
}

template <typename T>
inline TKameVariant& operator-=(TKameVariant& This, const T& Value)
{
    if (This.IsInteger()) {
        return This = This.As<long>() - Value;
    }
    else if (This.IsNumeric()) {
        return This = This.As<double>() - Value;
    }

    return This = This.As<T>() - Value;
}

template <typename T>
inline TKameVariant& operator*=(TKameVariant& This, const T& Value)
{
    if (This.IsInteger()) {
        return This = This.As<long>() * Value;
    }
    else if (This.IsNumeric()) {
        return This = This.As<double>() * Value;
    }

    return This = This.As<T>() * Value;
}

template <typename T>
inline TKameVariant& operator/=(TKameVariant& This, const T& Value)
{
    if (This.IsInteger()) {
        return This = This.As<long>() / Value;
    }
    else if (This.IsNumeric()) {
        return This = This.As<double>() / Value;
    }

    return This = This.As<T>() / Value;
}

template <typename T>
inline bool operator==(TKameVariant& This, const T& Value)
{
    return This.As<T>() == Value;
}

inline bool operator==(TKameVariant& This, const char Value[])
{
    return This.As<std::string>() == std::string(Value);
}

template <typename T>
inline bool operator!=(TKameVariant& This, const T& Value)
{
    return This.As<T>() != Value;
}

inline bool operator!=(TKameVariant& This, const char Value[])
{
    return This.As<std::string>() != std::string(Value);
}

template <typename T>
inline bool operator>(TKameVariant& This, const T& Value)
{
    return This.As<T>() > Value;
}

template <typename T>
inline bool operator<(TKameVariant& This, const T& Value)
{
    return This.As<T>() < Value;
}

template <typename T>
inline bool operator>=(TKameVariant& This, const T& Value)
{
    return This.As<T>() >= Value;
}

template <typename T>
inline bool operator<=(TKameVariant& This, const T& Value)
{
    return This.As<T>() <= Value;
}

#endif


typedef TKameVariant var;


}
#endif
