/* KcomProcess.cc */
/* Created by Enomoto Sanshiro on 25 March 2000. */
/* Updated by Enomoto Sanshiro on 19 May 2003. */
/* Last updated by Enomoto Sanshiro on 19 January 2010. */


#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include "MushNetworkSocket.hh"
#include "MushSocketStream.hh"
#include "MushArgumentList.hh"
#include "MushTimer.hh"
#include "MushSignal.hh"
#include "MushMisc.hh"
#include "KorbOrb.hh"
#include "KorbObjectAdapter.hh"
#include "KorbObjectMessenger.hh"
#include "KorbNamingContext.hh"
#include "KcomComponent.hh"
#include "KcomComponentOrb.hh"
#include "KcomEvent.hh"
#include "KcomTerminal.hh"
#include "KcomConsoleTerminal.hh"
#include "KcomTerminalStream.hh"
#include "KcomProcess.hh"

using namespace std;


TKcomProcess::TKcomProcess(TKcomComponent *Component)
{
    _Component = Component;
    _ComponentName = "";
    _ExportsObject = false;

    _Orb = new TKorbOrb();
    _ObjectAdapter = _Orb->ObjectAdapter();

    _ComponentMessenger = 0;

    _InputStream = 0;
    _OutputStream = 0;
    _Terminal = 0;

    _TerminalClient = 0;
    _Socket = 0;
    _LocalTerminal = 0;
}

TKcomProcess::~TKcomProcess()
{
    delete _LocalTerminal;
    delete _Socket;
    delete _TerminalClient;

    delete _InputStream;
    delete _OutputStream;

    delete _ComponentMessenger;
    delete _Orb;
}

void TKcomProcess::Start(TMushArgumentList& ArgumentList)
{
    _ComponentName = ArgumentList[0];
    string Control = ArgumentList[1]; // optional

    if (_ComponentName.empty()) {
	_Component->Descriptor()->Print(cout);
    }
    else {
	try {
	    AttachIO(Control);
	}
	catch (TKcomException &e) {
	    cerr << "ERROR: " << _ComponentName << ": " << e << endl;
	    cerr << "attach null-stream..." << endl;
	    AttachNullIO();
	}

	AttachSystem();

	StartTransaction();

	DetachIO();
	DetachSystem();
    }
}

void TKcomProcess::AttachIO(const string& Control)
{
    if (Control.empty()) {
	AttachConsoleIO();
    }
    else {
	string ControlType;
	istringstream ControlStream(Control);
	getline(ControlStream, ControlType, ':');

	if (ControlType == "display") {
	    string Display;
	    ControlStream >> Display;
	    AttachTerminalIO(Display);
	}
	else if (ControlType == "port") {
	    int PortNumber;
	    if (ControlStream >> PortNumber) {
		AttachSocketIO(PortNumber);
	    }
	    else {
		throw TKcomException(
		    "TKcomProcess::Start()", "Invalid port number"
		);
	    }
	}
	else if (ControlType == "file") {
	    string FileName;
	    ControlStream >> FileName;
	    AttachFileIO(FileName);
	}
	else if (ControlType == "null") {
	    AttachNullIO();
	}
	else {
	    throw TKcomException(
		"TKcomProcess::Start()", "Invalid control argument"
	    );
	}
    }
}

void TKcomProcess::AttachSocketIO(int PortNumber)
{
    _Socket = new TMushServerNetworkSocket(PortNumber);
    try {
	_Socket->EnableLocalAddressReuse();
	_Socket->Open();
    }
    catch (TSystemCallException &e) {
	throw TKcomException(
	    "TKcomProcess::AttachSocketIO()",
	    "system call exception: " + e.Message()
	);
    }

    _InputStream = new TMushInputSocketStream(_Socket);
    _OutputStream = new TMushOutputSocketStream(_Socket);    
    
    _Component->AttachIO(_InputStream, _OutputStream);
}

void TKcomProcess::AttachTerminalIO(const string& Display)
{
    _TerminalClient = new TKcomConsoleTerminalClient(
	_ComponentName, _Orb, Display
    );
    _Terminal = _TerminalClient->Attach();

    _InputStream = new TKcomInputTerminalStream(_Terminal);
    _OutputStream = new TKcomOutputTerminalStream(_Terminal);    
    
    _Component->AttachIO(_InputStream, _OutputStream, _Terminal);
}

void TKcomProcess::AttachFileIO(const string& FileName)
{
    _InputStream = new ifstream("/dev/null");
    _OutputStream = new ofstream(FileName.c_str(), ios::app);
    
    _Component->AttachIO(_InputStream, _OutputStream);
}

void TKcomProcess::AttachConsoleIO(void)
{
    _LocalTerminal = new TKcomConsoleTerminal();
    _Terminal = _LocalTerminal;

    _InputStream = new TKcomInputTerminalStream(_Terminal);
    _OutputStream = new TKcomOutputTerminalStream(_Terminal);    

    _Component->AttachIO(_InputStream, _OutputStream, _Terminal);
}

void TKcomProcess::AttachNullIO(void)
{
    _InputStream = new ifstream("/dev/null");
    _OutputStream = new ofstream("/dev/null");
    
    _Component->AttachIO(_InputStream, _OutputStream);
}

void TKcomProcess::AttachSystem(void)
{
    try {
	_ComponentMessenger = new TKcomComponentMessenger(_Component);
	_ObjectAdapter->AddMessenger(_ComponentMessenger, _ComponentName);
    }
    catch (TKorbException &e) {
	throw TKcomException(
	    "TKcomProcess::Attach()",
	    "korb exception: " + e.Message()
	);
    }

    _Component->AttachSystem(_ComponentName, _Orb);    
    if (_ObjectAdapter->NumberOfMessengers() > 1) {
	_ExportsObject = true;
    }
}

void TKcomProcess::DetachSystem(void)
{
    try {
	_ObjectAdapter->RemoveMessenger(_ComponentName);
	_Component->DetachSystem();
	_Orb->Finalize();
    }
    catch (TKorbException &e) {
	throw TKcomException(
	    "TKcomProcess::DetachSystem()",
	    "korb exception: " + e.Message()
	);
    }
}

void TKcomProcess::DetachIO(void)
{
    try {
	if (_Socket != 0) {
	    _Socket->Close();
	}
	if (_TerminalClient != 0) {
	    _TerminalClient->Detach();
	}
    }
    catch (TSystemCallException &e) {
	throw TKcomException(
	    "TKcomProcess::DetachIO()",
	    "system call exception: " + e.Message()
	);
    }
    catch (TKorbException &e) {
	throw TKcomException(
	    "TKcomProcess::DetachIO()",
	    "korb exception: " + e.Message()
	);
    }
}

void TKcomProcess::StartTransaction(void)
{
    try {
	TMushSignalHandler SignalHandler;
	TMushSignalCounter SignalCounter;
	SignalHandler.RegisterClient(SIGINT, &SignalCounter);
	SignalHandler.RegisterClient(SIGTERM, &SignalCounter);
	SignalHandler.StartHandling();

	_Component->Initialize();

	while (! _Component->IsStopRequested()) {
	    if (SignalCounter.SignalCount() > 0) {
		cerr << _ComponentName << ": signaled: terminating..." << endl;
		break;
	    }

	    _ObjectAdapter->DoNowaitTransaction();
	    _Component->OnTransaction();
	    _Component->DoTransaction();
	}

	SignalHandler.StartDefaultAction(SIGINT);
	SignalHandler.StartDefaultAction(SIGTERM);

	_Component->Finalize();
    }
    catch (TKorbException &e) {
	throw TKcomException(
	    "TKcomProcess::Start()",
	    "korb exception: " + e.Message()
        );
    }
    catch (TKcomException &e) {
	throw;
    }
}
