/* KorbBroker.cc */
/* Created by Enomoto Sanshiro on 13 February 2000. */
/* Updated by Enomoto Sanshiro on 29 July 2001. */
/* Last updated by Enomoto Sanshiro on 4 February 2010. */


#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include "MushProcess.hh"
#include "MushNetworkSocket.hh"
#include "MushSignal.hh"
#include "MushTimer.hh"
#include "MushFileSystem.hh"
#include "MushMisc.hh"
#include "MushDecoratedSocket.hh"
#include "KiscFancyConfigFile.hh"
#include "KorbNetworkHub.hh"
#include "KorbNetworkAdapter.hh"
#include "KorbNetworkPacket.hh"
#include "KorbOrbPacket.hh"
#include "KorbOrb.hh"
#include "KorbBroker.hh"

using namespace std;


//#define DEBUG
#ifdef DEBUG
static ofstream dumpfile("Dump-KorbBroker", ios::app);
#endif


static const int MaxNumberOfPortScans = 64;
static const int LaunchTimeOut_sec = 10;
static const int IdlingSleepTime_usec = 1000;


TKorbBroker::TKorbBroker(const string& PathList)
{
    _DomainId = PrimaryDomainId();
    _PathListString = PathList;

    _NetworkHub = 0;

    _IsStopRequested = false;
    _ParentConnection = 0;
}

TKorbBroker::TKorbBroker(int DomainId, int ParentDomainId, const string& ParentHostName, int ParentPortNumber, const string& WorkingDirectory, const string& PathList)
{
    _DomainId = DomainId;
    _ParentDomainId = ParentDomainId;
    _ParentHostName = ParentHostName;
    _ParentPortNumber = ParentPortNumber;

    _WorkingDirectory = WorkingDirectory;
    _PathListString = PathList;

    _NetworkHub = 0;

    _IsStopRequested = false;
    _ParentConnection = 0;
}

TKorbBroker::~TKorbBroker()
{
    for (unsigned i = 0; i < _ChildProcessList.size(); i++) {
	delete _ChildProcessList[i];
    }
    for (unsigned i = 0; i < _ChildBrokerProcessList.size(); i++) {
	delete _ChildBrokerProcessList[i];
    }
    for (unsigned j = 0; j < _SocketList.size(); j++) {
	delete _SocketList[j];
    }

    delete _NetworkHub;
}

void TKorbBroker::Initialize(void)
{
    if (_NetworkHub != 0) {
	return;
    }

    try {
	_RemoteShell = TMushEnvironmentVariable("KINOKO_RSH").AsString();
	_NextPortNumber = TMushEnvironmentVariable("KINOKO_CONTROL_PORT_BASE").AsLong();
    }
    catch (TSystemCallException &e) {
	throw TKorbException(
	    "TKorbBroker::Initialize() ",
	    "system call exception: " + e.Message()
	);
    }

    if (_WorkingDirectory.empty()) {
	_WorkingDirectory = TMushFileSystem::CurrentDirectory();
    }
    else {
	try {
	    TMushFileSystem::ChangeDirectory(_WorkingDirectory);
	}
	catch (TSystemCallException &e) {
	    _WorkingDirectory = TMushFileSystem::CurrentDirectory();
	}
    }

    string LocalHostName = TMushLocalHost::HostName();
    cout << "KORB_PROJECT_PATH at " << LocalHostName << ": ";
    cout << TKorbOrb::ProjectPath() << endl;
    cout << "Working Directory at " << LocalHostName << ": " ;
    cout << _WorkingDirectory << endl;

    string PathListStringBuffer = _PathListString;
    char Separator = PathListSeparator();
    while (! PathListStringBuffer.empty()) {
	if (PathListStringBuffer[0] == Separator) {
	    PathListStringBuffer.erase(PathListStringBuffer.begin());
	    continue;
	}

	string::size_type Length = PathListStringBuffer.find_first_of(Separator);
	string PathEntry = PathListStringBuffer.substr(0, Length);

	_LaunchPathList.push_back(PathEntry);
	PathListStringBuffer.erase(0, Length);
    }

    string ProjectPath = TKorbOrb::ProjectPath();
    int ProjectId = TKorbOrb::ProjectId();
    _NetworkHub = new TKorbNetworkHub(_DomainId, ProjectId, ProjectPath);
    _NetworkHub->Initialize();

    if (_DomainId != PrimaryDomainId()) {
	ConnectParent();
    }
}

void TKorbBroker::Finalize(void)
{
    for (unsigned i = 0; i < _ChildProcessList.size(); i++) {
	TMushChildProcess* Process = _ChildProcessList[i];
	if (Process->IsRunning()) {
	    string Name = Process->Name();
	    Name += (_HostName.empty()) ? "" : ("@" + _HostName);
	    cerr << "WARNING: process " << Name << " still running. ";
	    cerr << "terminating..." << endl;
	    TMushSignalSender::SendSignal(Process->ProcessId(), SIGTERM);
	    TMushRealTimeTimer(0, 100000).Suspend();
	}
    }

    _NetworkHub->Finalize();
}

void TKorbBroker::ConnectParent(void)
{
    TMushNetworkSocket* Socket = 0;
    try {
	Socket = new TMushClientNetworkSocket(_ParentHostName, _ParentPortNumber);
	Socket->Open();
    }
    catch (TSystemCallException &e) {
	delete Socket;
	ostringstream PortNumberStrStream;
	PortNumberStrStream  << _ParentPortNumber;
	throw TKorbException(
            "TKorbBroker::ConnectParent()",
	    "system call exception: " + e.Message() +
            ": " + _ParentHostName + ":" + PortNumberStrStream.str()
	);
    }
    _ParentConnection = Socket;

    TMushSocket* FramedSocket = new TMushFramedSocket(Socket);
    _NetworkHub->AddConnection(_ParentDomainId, FramedSocket);
    _SocketList.push_back(FramedSocket);
}

void TKorbBroker::Start(void)
{
    TMushSignalHandler SignalHandler;
    TMushSignalCounter SignalCounter;
    SignalHandler.RegisterClient(SIGINT, &SignalCounter);
    SignalHandler.RegisterClient(SIGTERM, &SignalCounter);
    SignalHandler.RegisterClient(SIGHUP, &SignalCounter);
    SignalHandler.StartHandling();

    Initialize();
    _IsStopRequested = false;

    int Activity = 0;
    while ((! _IsStopRequested) || (Activity > 0)) {
	if (SignalCounter.SignalCount() > 0) {
	    cerr << "KorbBroker" << (_HostName.empty() ? "" : "@" + _HostName);
	    cerr << ": signaled: terminating..." << endl;
	    break;
	}
	if (
	    (_ParentConnection != 0) && 
	    (_ParentConnection->Status() != TMushSocket::ssOpened)
	){
	    cerr << "KorbBroker" << (_HostName.empty() ? "" : "@" + _HostName);
	    cerr << ": parent connection closed: terminating..." << endl;
	    break;
	}

	int ThisActivity = _NetworkHub->DoTransaction();
	
	while (_NetworkHub->Receive(_NetworkPacket) > 0) {
	    TKorbOrbPacket OrbPacket(&_NetworkPacket);
	    ProcessOrbPacket(OrbPacket);
	    ThisActivity++;
	}
	
	if (_IsStopRequested) {
	    ThisActivity += _NetworkHub->NumberOfPendingPackets();
	}
	if (Activity < 10) {
	    Activity += 2 * ThisActivity;
	}
	if (Activity > 0) {
	    Activity--;
	}

	if (Activity <= 0) {
	    TMushRealTimeTimer(0, IdlingSleepTime_usec).Suspend();
	}
    }

    SignalHandler.StartDefaultAction(SIGINT);
    SignalHandler.StartDefaultAction(SIGTERM);
    SignalHandler.StartDefaultAction(SIGHUP);

    Finalize();
}

int TKorbBroker::ProcessOrbPacket(TKorbOrbPacket& OrbPacket)
{
    OrbPacket.OnReceive();

#ifdef DEBUG
    OrbPacket.Dump(dumpfile);
#endif

    int Result = 0;
    if (OrbPacket.MethodId() == MethodId_LaunchChildBroker) {
	Result = LaunchChildBroker(OrbPacket);
    }
    else if (OrbPacket.MethodId() == MethodId_LaunchProcess) {
	Result = LaunchProcess(OrbPacket);
    }
    else if (OrbPacket.MethodId() == MethodId_SendSignal) {
	Result = SendSignal(OrbPacket);
    }
    else if (OrbPacket.MethodId() == MethodId_Exit) {
	Result = Exit(OrbPacket);
    }

    if (Result > 0) {
	if (! OrbPacket.IsOneWay()) {
	    TKorbNetworkPacket* NetworkPacket = OrbPacket.NetworkPacket();
	    int ReceiverDomainId = NetworkPacket->SenderDomainId();
	    int ReceiverAdapterId = NetworkPacket->SenderAdapterId();
	    NetworkPacket->SetSender(_DomainId, this->AdapterId());
	    NetworkPacket->SetReceiver(ReceiverDomainId, ReceiverAdapterId);
	    NetworkPacket->SetDataSize(OrbPacket.PacketSize());
	    OrbPacket.IsRequest() = 0;
	    OrbPacket.OnSend();
	    
#ifdef DEBUG
	    OrbPacket.Dump(dumpfile);
#endif
	    
	    _NetworkHub->Send(*NetworkPacket);
	    _NetworkHub->DoTransaction();
	}
    }
    else {
	// Here we might have to send error status...
#if 0
	throw TKorbException(
	    "TKorbBroker::ProcessOrbPacket()", "invalid method id"
	);
#else
	cerr << "ERROR: KorbBroker: invalid method id" << endl;
	OrbPacket.Dump(cerr);
#endif
    }

    return Result;
}

int TKorbBroker::LaunchChildBroker(TKorbOrbPacket& OrbPacket)
{
    string RemoteHostName;
    int ChildDomainId;
    string LoginName;
    string ProcessPathList;
    string RemoteWorkingDirectory;
    OrbPacket.ArgumentSerializer().GetString(RemoteHostName);
    OrbPacket.ArgumentSerializer().GetInt(ChildDomainId);
    OrbPacket.ArgumentSerializer().GetString(LoginName);
    OrbPacket.ArgumentSerializer().GetString(ProcessPathList);
    OrbPacket.ArgumentSerializer().GetString(RemoteWorkingDirectory);

    int PortNumber;
    TMushServerNetworkSocket* Socket = 0;
    for (int i = 0; i < MaxNumberOfPortScans; i++) {
	PortNumber = _NextPortNumber++;
	Socket = new TMushServerNetworkSocket(PortNumber);
	try {
	    Socket->Bind();
	}
	catch (TSystemCallException &e) {
	    delete Socket;
	    Socket = 0;
	    continue;
	}
	break;
    }
    if (Socket == 0) {
	throw TKorbException(
	    "TKorbBroker::LaunchChildBroker()", "unable to bind socket"
	);
    }
    string LocalHostName = Socket->LocalIPAddress();

    if (RemoteWorkingDirectory.empty()) {
	RemoteWorkingDirectory = _WorkingDirectory;
    }

    ostringstream ParentDomainIdStream;
    ostringstream ChildDomainIdStream;
    ostringstream PortNumberStream;

    ParentDomainIdStream << _DomainId;
    ChildDomainIdStream << ChildDomainId;
    PortNumberStream << PortNumber;

    vector<string> ArgumentList;
    ArgumentList.push_back(_RemoteShell);
    if (! LoginName.empty()) {
        ArgumentList.push_back("-l");
        ArgumentList.push_back(LoginName);
    }
    ArgumentList.push_back(RemoteHostName);
    ArgumentList.push_back("korb-broker");
    ArgumentList.push_back(ChildDomainIdStream.str());
    ArgumentList.push_back(ParentDomainIdStream.str());
    ArgumentList.push_back(LocalHostName);
    ArgumentList.push_back(PortNumberStream.str());
    ArgumentList.push_back(RemoteWorkingDirectory);
    ArgumentList.push_back(ProcessPathList);

    TMushChildProcess* Process = new TMushChildProcess(_RemoteShell, ArgumentList);

    int ProcessId = -1;
    string ErrorMessage;
    try {
	Socket->Listen();
	Process->Run();
	ProcessId = Process->ProcessId();

	TMushRealTimeTimer TimeKeeper(LaunchTimeOut_sec, 0);
	TimeKeeper.Start();
	Socket->Accept();
	TimeKeeper.Stop();

	TMushSocket* FramedSocket = new TMushFramedSocket(Socket);
	_SocketList.push_back(FramedSocket);
	_NetworkHub->AddConnection(ChildDomainId, FramedSocket);
	_ChildBrokerProcessList.push_back(Process);
    }
    catch (TSystemCallException &e) {
	delete Process; Process = 0;
	delete Socket; Socket = 0;
	ProcessId = -1;

	if (ErrorMessage.empty()) {
	    ErrorMessage = (
		"TKorbBroker::LaunchBroker(): "
		"unable to start or connect to remote broker process"
		" at host " + RemoteHostName +
		" from PATH " + ProcessPathList
	    );
	}
    }

    OrbPacket.ArgumentSerializer().Rewind();
    OrbPacket.ArgumentSerializer().PutInt(ProcessId);
    OrbPacket.ArgumentSerializer().PutString(ErrorMessage);    

    return 1;
}

int TKorbBroker::LaunchProcess(TKorbOrbPacket& OrbPacket)
{
    string Path;
    int NumberOfArguments;
    string Argument;
    vector<string> ArgumentList;

    OrbPacket.ArgumentSerializer().GetString(Path);
    OrbPacket.ArgumentSerializer().GetInt(NumberOfArguments);
    for (int i = 0; i < NumberOfArguments; i++) {
	OrbPacket.ArgumentSerializer().GetString(Argument);
	ArgumentList.push_back(Argument);
    }

    TMushChildProcess* Process;
    int ProcessId = -1;
    string ErrorMessage;
    try {
	if (! _LaunchPathList.empty()) {
	    Process = new TMushChildProcess(_LaunchPathList, Path, ArgumentList);
	}
	else {
	    Process = new TMushChildProcess(Path, ArgumentList);
	}
	Process->Run();

	ProcessId = Process->ProcessId();
	_ChildProcessList.push_back(Process);
    }
    catch (TSystemCallException &e) {
	delete Process;

	ErrorMessage = (
	    "TKorbBroker::LaunchProcess(): "
	    "system call exception: " + e.Message()
	);
    }

    OrbPacket.ArgumentSerializer().Rewind();
    OrbPacket.ArgumentSerializer().PutInt(ProcessId);
    OrbPacket.ArgumentSerializer().PutString(ErrorMessage);    

    return 1;
}

int TKorbBroker::SendSignal(TKorbOrbPacket& OrbPacket)
{
    int ProcessId;
    int SignalId;

    OrbPacket.ArgumentSerializer().GetInt(ProcessId);
    OrbPacket.ArgumentSerializer().GetInt(SignalId);

    TMushSignalSender::SendSignal(ProcessId, SignalId);

    return 1;
}

int TKorbBroker::Exit(TKorbOrbPacket& OrbPacket)
{
    _IsStopRequested = true;

    return 1;
}
