/* KorbCommander.cc */
/* Created by Enomoto Sanshiro on 2 March 2000. */
/* Last updated by Enomoto Sanshiro on 4 February 2010. */


#include <unistd.h>
#include "MushProcess.hh"
#include "MushTimer.hh"
#include "MushMisc.hh"
#include "KorbNetworkAdapter.hh"
#include "KorbNetworkPacket.hh"
#include "KorbOrb.hh"
#include "KorbOrbPacket.hh"
#include "KorbBroker.hh"
#include "KorbNamingContext.hh"
#include "KorbCommander.hh"

using namespace std;


TKorbCommander::TKorbCommander(TKorbOrb* Orb)
{
    _Orb = Orb;
    _BrokerProcess = 0;
    _NameServerProcess = 0;
}

TKorbCommander::~TKorbCommander()
{
}

int TKorbCommander::Initialize(const string& ProcessPathList)
{
    if (_BrokerProcess != 0) {
	return 0;
    }

    string Path = "korb-broker";
    vector<string> ArgumentList;
    ArgumentList.push_back("korb-broker");
    ArgumentList.push_back(ProcessPathList);

    try {
	_BrokerProcess = new TMushChildProcess(Path, ArgumentList);
	_BrokerProcess->Run();
    }
    catch (TSystemCallException &e) {
	delete _BrokerProcess;
	_BrokerProcess = 0;

	throw TKorbException(
	    "TKorbCommander::LaunchBroker()",
	    "system call exception: " + e.Message()
	);
    }

#if 1
    /* wait for the broker process to get launched */
    // The following _Orb->Initialize() waits for the broker
    // by looking for the message queue which will be created
    // by the broker process, but it does not work if an old 
    // message queue is left on the system.
    TMushRealTimeTimer(1, 0).Suspend();
#endif

#if 0
    static const int Provation_sec = 1;
    TMushRealTimeTimer(Provation_sec, 0).Suspend();
    if (! _BrokerProcess->IsRunning()) {
	throw TKorbException(
	    "TKorbCommander::LaunchBroker()",
	    "broker process seems to be crashed."
	);
    }
#endif

    _Orb->Initialize();
    LaunchNameServer();

    return 1;
}

int TKorbCommander::Finalize(void)
{
    if (_BrokerProcess == 0) {
	return 0;
    }

    bool IsTerminationSuccess;
    try {
	IsTerminationSuccess = (TerminateNameServer() != 0);
    }
    catch (TKorbException &e) {
	IsTerminationSuccess = false;
    }
    if (! IsTerminationSuccess) {
	cerr << "WARNING: korb-nameserver still running. ";
	cerr << "Terminating..." << endl;
	TMushSignalSender::SendSignal(
	    _NameServerProcess->ProcessId(), SIGTERM
	);
    }
    
    try {
	IsTerminationSuccess = (TerminateBroker() != 0);
    }
    catch (TKorbException &e) {
	IsTerminationSuccess = false;
    }
    if (! IsTerminationSuccess) {
	cerr << "WARNING: korb-broker still running. ";
	cerr << "Terminating..." << endl;
	TMushSignalSender::SendSignal(
	    _BrokerProcess->ProcessId(), SIGTERM
	);
    }

    _NameServerProcess->WaitToExit();
    _BrokerProcess->WaitToExit();

    delete _NameServerProcess; _NameServerProcess = 0;
    delete _BrokerProcess; _BrokerProcess = 0;

    return 1;
}

TKorbOrbPacket TKorbCommander::CreateOrbPacket(int MethodId)
{
    return _Orb->CreateClientOrbPacket(TKorbObjectReference(), MethodId);
}

int TKorbCommander::InvokeBrokerMethod(int DomainId, TKorbOrbPacket& OrbPacket, int TimeOut_sec)
{
    static const char* ClassName = "KorbBroker";
    static const char* ObjectName = "broker";
    int AdapterId = TKorbBroker::AdapterId();
    int ObjectId = OrbPacket.ObjectId();

    TKorbObjectReference BrokerReference(
	ClassName, ObjectName, 
	DomainId, AdapterId, ObjectId
    );

    return _Orb->InvokeMethod(BrokerReference, OrbPacket, TimeOut_sec);
}

int TKorbCommander::LaunchRemoteBroker(int ParentDomainId, const string& HostName, const string& LoginName, const string& ProcessPathList, const std::string& RemoteWorkingDirectory)
{
    int MethodId = TKorbBroker::MethodId_LaunchChildBroker;
    TKorbOrbPacket OrbPacket = CreateOrbPacket(MethodId);

    static int _NextDomainId = 2;
    int ChildDomainId = _NextDomainId++;

    OrbPacket.ArgumentSerializer().PutString(HostName);
    OrbPacket.ArgumentSerializer().PutInt(ChildDomainId);
    OrbPacket.ArgumentSerializer().PutString(LoginName);
    OrbPacket.ArgumentSerializer().PutString(ProcessPathList);
    OrbPacket.ArgumentSerializer().PutString(RemoteWorkingDirectory);

    InvokeBrokerMethod(ParentDomainId, OrbPacket);

    int ProcessId;
    OrbPacket.ArgumentSerializer().GetInt(ProcessId);

    if (ProcessId < 0) {
	string Message;
	OrbPacket.ArgumentSerializer().GetString(Message);
	
	throw TKorbException(
	    "TKorbCommander::LaunchRemoteBroker()", Message
	);
    }

    return ChildDomainId;
}

int TKorbCommander::LaunchProcess(int DomainId, const string& Path, const vector<string>& ArgumentList)
{
    int MethodId = TKorbBroker::MethodId_LaunchProcess;
    TKorbOrbPacket OrbPacket = CreateOrbPacket(MethodId);

    int NumberOfArguments = ArgumentList.size();
    OrbPacket.ArgumentSerializer().PutString(Path);
    OrbPacket.ArgumentSerializer().PutInt(NumberOfArguments);
    for (int i = 0; i < NumberOfArguments; i++) {
	OrbPacket.ArgumentSerializer().PutString(ArgumentList[i]);
    }

    InvokeBrokerMethod(DomainId, OrbPacket);

    int ProcessId;
    OrbPacket.ArgumentSerializer().GetInt(ProcessId);

    if (ProcessId < 0) {
	string Message;
	OrbPacket.ArgumentSerializer().GetString(Message);
	
	throw TKorbException(
	    "TKorbCommander::LaunchProcess()",
	    Message
	);
    }

    return ProcessId;
}

void TKorbCommander::LaunchNameServer(void)
{
    if (_NameServerProcess != 0) {
	return;
    }

    string Path = "korb-nameserver";
    vector<string> ArgumentList;
    ArgumentList.push_back("korb-nameserver");

    _NameServerProcess = new TMushChildProcess(Path, ArgumentList);
    try {
	_NameServerProcess->Run();
    }
    catch (TSystemCallException &e) {
	throw TKorbException(
	    "TKorbBroker::LaunchNameServer()",
	    "system call exception: " + e.Message()
	);
    }
}

int TKorbCommander::SendSignal(int DomainId, int ProcessId, int SignalId)
{
    int MethodId = TKorbBroker::MethodId_SendSignal;
    TKorbOrbPacket OrbPacket = CreateOrbPacket(MethodId);

    OrbPacket.ArgumentSerializer().PutInt(ProcessId);
    OrbPacket.ArgumentSerializer().PutInt(SignalId);

    InvokeBrokerMethod(DomainId, OrbPacket);

    return 0;
}

int TKorbCommander::TerminateBroker(int DomainId)
{
    int MethodId = TKorbBroker::MethodId_Exit;
    TKorbOrbPacket OrbPacket = CreateOrbPacket(MethodId);
    int TimeOut_sec = 1;
    
    return InvokeBrokerMethod(DomainId, OrbPacket, TimeOut_sec);
}

int TKorbCommander::TerminateNameServer(void)
{
    if ((_NameServerProcess == 0) || ! _NameServerProcess->IsRunning()) {
	return 1;
    }

    return _Orb->TerminateAdapter(TKorbNamingContext::GetInitialReference());
}
