/* KorbOrb.cc */
/* Created by Enomoto Sanshiro on 26 February 2000. */
/* Updated by Enomoto Sanshiro on 29 April 2002. */
/* Last updated by Enomoto Sanshiro on 19 January 2010. */


#include <fstream>
#include <string>
#include <deque>
#include <climits>
#include "MushProcess.hh"
#include "MushTimer.hh"
#include "MushMisc.hh"
#include "MushFileSystem.hh"
#include "KorbNetworkHub.hh"
#include "KorbNetworkAdapter.hh"
#include "KorbOrbPacket.hh"
#include "KorbObjectReference.hh"
#include "KorbObjectAdapter.hh"
#include "KorbNamingContext.hh"
#include "KorbBroker.hh"
#include "KorbOrb.hh"

using namespace std;


//#define DEBUG
#ifdef DEBUG
static ofstream dumpfile("Dump-KorbOrb", ios::app);
#endif


static const int MessageQueueProjectId = 1;
static const int MaxRequestCount = 100;
static const int MaxRequestIdBase = 100000000;


TKorbOrb::TKorbOrb(void)
{
    _NetworkDomainId = TKorbNetworkHub::DomainId_Unknown;
    _NetworkAdapterId = TMushProcess().ProcessId();

    _ObjectAdapter = 0;
    _NamingContext = 0;
    _NetworkAdapter = 0;

    _RequestIdBase = (_NetworkAdapterId % (MaxRequestIdBase / MaxRequestCount)) * MaxRequestCount;
    _RequestCount = 0;
}

TKorbOrb::TKorbOrb(int NetworkAdapterId)
{
    _NetworkDomainId = TKorbNetworkHub::DomainId_Unknown;
    _NetworkAdapterId = NetworkAdapterId;

    _ObjectAdapter = 0;
    _NamingContext = 0;
    _NetworkAdapter = 0;

    _RequestIdBase = (_NetworkAdapterId % (MaxRequestIdBase / MaxRequestCount)) * MaxRequestCount;
    _RequestCount = 0;
}

TKorbOrb::~TKorbOrb()
{
    delete _NamingContext;
    delete _ObjectAdapter;
    delete _NetworkAdapter;
}

int TKorbOrb::ProjectId(void)
{
    return MessageQueueProjectId;
}

string TKorbOrb::ProjectPath(void)
{
    string ProjectPath;
    try {
	ProjectPath = TMushEnvironmentVariable("KORB_PROJECT_PATH").AsString();
    }
    catch (TSystemCallException &e) {
	try {
	    ProjectPath = TMushEnvironmentVariable("KINOKO_ROOT").AsString();
	    ProjectPath += "/tmp";
	}
	catch (TSystemCallException &e) {
	    throw TKorbException(
		"TKorbOrb::ProjectPath()", 
		"neither KORB_PROJECT_PATH nor KINOKO_ROOT defined"
	    );
	}
    }    

    return ProjectPath;
}

int TKorbOrb::QueueCapacity(void)
{
    return _NetworkAdapter->QueueCapacity();
}

int TKorbOrb::DomainId(void)
{
    if (_NetworkAdapter == 0) {
	Initialize();
    }

    return _NetworkDomainId;
}

int TKorbOrb::AdapterId(void)
{
    return _NetworkAdapterId;
}

TKorbObjectAdapter* TKorbOrb::ObjectAdapter(void)
{
    if (_ObjectAdapter == 0) {
	_ObjectAdapter = new TKorbObjectAdapter(this);
    }
    
    return _ObjectAdapter;
}

TKorbNamingContext* TKorbOrb::NamingContext(void)
{
    if (_NamingContext == 0) {
	_NamingContext = new TKorbNamingContextProxy(this);
	_NamingContext->Bind(
	    TKorbNamingContext::GetInitialReference()
	);
    }
    
    return _NamingContext;
}

void TKorbOrb::Initialize(void)
{
    if (_NetworkAdapter != 0) {
	return;
    }

    _NetworkAdapter = new TKorbNetworkAdapter(
	_NetworkAdapterId, ProjectId(), ProjectPath()
    );

    static const int AttachWait_usec = 10000;
    static const int MaxNumberOfTries = 3000;  // i.e. 30 sec.
    int NumberOfTries = 0;
    while (1) {
	try {
	    _NetworkAdapter->Attach();
	}
	catch (TKorbException& e) {
	    NumberOfTries++;
	    if (NumberOfTries < MaxNumberOfTries) {
		TMushRealTimeTimer(0, AttachWait_usec).Suspend();
		continue;
	    }
	    throw e;
	}
	break;
    }

    _NetworkDomainId = _NetworkAdapter->DomainId();
}

void TKorbOrb::Finalize(void)
{
    static const int IsNonBlocking = false;
    while (Flush(IsNonBlocking)) {
	;
    }
}

TKorbOrbPacket TKorbOrb::CreateServerOrbPacket(void)
{
    TKorbOrbPacket OrbPacket(&_ServerNetworkPacket);

    return OrbPacket;
}

TKorbOrbPacket TKorbOrb::CreateClientOrbPacket(const TKorbObjectReference& ObjectReference, int MethodId)
{
    TKorbOrbPacket OrbPacket(&_ClientNetworkPacket);

    OrbPacket.ObjectId() = ObjectReference.ObjectId();
    OrbPacket.MethodId() = MethodId;
    OrbPacket.IsOneWay() = 0;

    return OrbPacket;
}

int TKorbOrb::InvokeMethod(const TKorbObjectReference& ObjectReference, TKorbOrbPacket& OrbPacket, int TimeOut_sec, bool IsAsynchronous)
{
    int RemoteDomainId = ObjectReference.DomainId();
    int RemoteAdapterId = ObjectReference.AdapterId();

    int RequestId = SendRequest(RemoteDomainId, RemoteAdapterId, OrbPacket);

    int Result;
    if (OrbPacket.IsOneWay() || IsAsynchronous) {
	Result = RequestId;
    }
    else {
	Result = WaitReplyFor(RequestId, OrbPacket, TimeOut_sec);
    }
    
    return Result;
}

void TKorbOrb::SendSignal(const TKorbObjectReference& ObjectReference, int SignalId, bool IsOneWay)
{
    TKorbOrbPacket OrbPacket(&_ClientNetworkPacket);
    OrbPacket.MethodId() = TKorbBroker::MethodId_SendSignal;
    if (IsOneWay) {
	OrbPacket.IsOneWay() = 1;
    }

    int ProcessId = ObjectReference.AdapterId();

    OrbPacket.ArgumentSerializer().PutInt(ProcessId);
    OrbPacket.ArgumentSerializer().PutInt(SignalId);

    int RemoteDomainId = ObjectReference.DomainId();
    int RemoteAdapterId = TKorbBroker::AdapterId();

    int RequestId = SendRequest(RemoteDomainId, RemoteAdapterId, OrbPacket);

    if (! IsOneWay) {
	WaitReplyFor(RequestId, OrbPacket);
    }
}

int TKorbOrb::TerminateAdapter(const TKorbObjectReference& ObjectReference)
{
    //... FIXME: one object adapter may have several object messengers ...//

    TKorbOrbPacket OrbPacket(&_ClientNetworkPacket);
    OrbPacket.ObjectId() = 0;
    OrbPacket.MethodId() = TKorbObjectAdapter::MethodId_Exit;
    OrbPacket.IsOneWay() = 0;
    int TimeOut_sec = 1;

    return InvokeMethod(ObjectReference, OrbPacket, TimeOut_sec);
}

int TKorbOrb::SendRequest(int DomainId, int AdapterId, TKorbOrbPacket& OrbPacket)
{
    if (_NetworkAdapter == 0) {
	Initialize();
    }

    _RequestCount = _RequestCount % (MaxRequestCount - 1) + 1;
    int RequestId = _RequestIdBase + _RequestCount;

    OrbPacket.RequestId() = RequestId;
    OrbPacket.IsRequest() = 1;

#ifdef DEBUG
    dumpfile << _NetworkDomainId << "." << _NetworkAdapterId << ": ";
    dumpfile << "send request: " << OrbPacket.RequestId() << endl;
#endif

    static const bool IsNonBlocking = true;
    Send(DomainId, AdapterId, OrbPacket, IsNonBlocking);

    return RequestId;
}

int TKorbOrb::SendReplyFor(int RequestId, int DomainId, int AdapterId, TKorbOrbPacket& OrbPacket)
{
    OrbPacket.RequestId() = RequestId;
    OrbPacket.IsRequest() = 0;

    static const bool IsNonBlocking = true;
    Send(DomainId, AdapterId, OrbPacket, IsNonBlocking);

    return RequestId;
}

int TKorbOrb::WaitRequest(TKorbOrbPacket& OrbPacket)
{
    if (! _RequestPacketQueue.empty()) {
	*(OrbPacket.NetworkPacket()) = _RequestPacketQueue.front();
	_RequestPacketQueue.pop_front();
	OrbPacket.OnReceive();

#ifdef DEBUG
	dumpfile << _NetworkDomainId << "." << _NetworkAdapterId << ": ";
	dumpfile << "pop request: " << OrbPacket.RequestId() << endl;
	OrbPacket.Dump(dumpfile);
#endif

	return OrbPacket.RequestId();
    }

    while (Receive(OrbPacket) > 0) {
	if (! OrbPacket.IsRequest()) {
	    _ReplyPacketTable[OrbPacket.RequestId()] = *OrbPacket.NetworkPacket();
	    continue;
	}

#ifdef DEBUG
	dumpfile << _NetworkDomainId << "." << _NetworkAdapterId << ": ";
	dumpfile << "receive request: " << OrbPacket.RequestId() << endl;
#endif

	return OrbPacket.RequestId();
    }

    return 0;
}

int TKorbOrb::GetRequest(TKorbOrbPacket& OrbPacket)
{
    if (! _RequestPacketQueue.empty()) {
	return WaitRequest(OrbPacket);
    }

    static const bool IsNonBlocking = true;
    while (Receive(OrbPacket, IsNonBlocking) > 0) {
	if (! OrbPacket.IsRequest()) {
	    _ReplyPacketTable[OrbPacket.RequestId()] = *OrbPacket.NetworkPacket();
	    continue;
	}

#ifdef DEBUG
	dumpfile << _NetworkDomainId << "." << _NetworkAdapterId << ": ";
	dumpfile << "receive request: " << OrbPacket.RequestId() << endl;
#endif

	return OrbPacket.RequestId();
    }

    return 0;
}

int TKorbOrb::WaitReplyFor(int RequestId, TKorbOrbPacket& OrbPacket, int TimeOut_sec)
{
    int StartTime = TMushDateTime::SecSinceEpoch();
    int Result = 0;

    std::map<int, TKorbNetworkPacket>::iterator ReplyPacketIterator;
    ReplyPacketIterator = _ReplyPacketTable.find(RequestId);
    if (ReplyPacketIterator != _ReplyPacketTable.end()) {
	*(OrbPacket.NetworkPacket()) = (*ReplyPacketIterator).second;
	_ReplyPacketTable.erase(ReplyPacketIterator);
	OrbPacket.OnReceive();
	return 1;
    }

    while (1) {
	static const bool IsNonBlocking = true;
	if (Receive(OrbPacket, IsNonBlocking) > 0) {
	    if (OrbPacket.IsRequest()) {
		_RequestPacketQueue.push_back(*OrbPacket.NetworkPacket());
#ifdef DEBUG
		dumpfile << _NetworkDomainId << ".";
		dumpfile << _NetworkAdapterId << ": ";
		dumpfile << "push request: " << OrbPacket.RequestId() << endl;
		OrbPacket.Dump(dumpfile);
#endif
		continue;
	    }
	    else if (OrbPacket.RequestId() != RequestId) {
		_ReplyPacketTable[OrbPacket.RequestId()] = *OrbPacket.NetworkPacket();
		continue;
	    }
	    else {
#ifdef DEBUG
		dumpfile << _NetworkDomainId << ".";
		dumpfile << _NetworkAdapterId << ": ";
		dumpfile << "receive reply: " << OrbPacket.RequestId() << endl;
#endif
		Result = RequestId;
		break;
	    }
	}

	if (
	    (TimeOut_sec > 0) &&
	    (TMushDateTime::SecSince(StartTime) > TimeOut_sec)
	){
	    Result = 0;
	    break;
	}
	
	TMushRealTimeTimer(0, 10000).Suspend();
    }

    return Result;
}

int TKorbOrb::Send(int NetworkDomainId, int NetworkAdapterId, TKorbOrbPacket& OrbPacket, bool IsNonBlocking)
{
    if (_NetworkAdapter == 0) {
	Initialize();
    }
    Flush(IsNonBlocking);

    if (! OrbPacket.IsPacketValid()) {
	throw TKorbException(
	    "TKorbOrb::Send()", "ORB Packet overflow detected"
	);
    }

    TKorbNetworkPacket* NetworkPacket = OrbPacket.NetworkPacket();
    NetworkPacket->SetSender(_NetworkDomainId, _NetworkAdapterId);
    NetworkPacket->SetReceiver(NetworkDomainId, NetworkAdapterId);
    NetworkPacket->SetDataSize(OrbPacket.PacketSize());

    OrbPacket.OnSend();

    int Length = 0;
    if (! _SendingQueue.empty()) {
	// keep the packet order //
	_SendingQueue.push_back(*NetworkPacket);
    }
    else {
	Length = _NetworkAdapter->Send(
	    *NetworkPacket,
	    (IsNonBlocking ? TKorbNetworkAdapter::Flag_NonBlocking : 0)
	);
	if (Length <= 0) {
	    _SendingQueue.push_back(*NetworkPacket);
#if 0
	    static bool IsWarningDisplayed = false;
	    if (! IsWarningDisplayed) {
		cerr << "WARNING: KinokoORB: MessageQueue Full." << endl;
		IsWarningDisplayed = true;
	    }
#endif
	}
    }

    Flush(IsNonBlocking);

    return (Length > 0) ? 1 : 0;
}

int TKorbOrb::Receive(TKorbOrbPacket& OrbPacket, bool IsNonBlocking)
{
    if (_NetworkAdapter == 0) {
	Initialize();
    }
    static const bool IsFlushNonBlocking = true;
    Flush(IsFlushNonBlocking);

    TKorbNetworkPacket* NetworkPacket = OrbPacket.NetworkPacket();
    
    int Length;
    if (! _ReceivingQueue.empty()) {
	*NetworkPacket = _ReceivingQueue.front();
	_ReceivingQueue.pop_front();
	Length = NetworkPacket->DataSize();
    }
    else {
	Length = _NetworkAdapter->Receive(
	    *NetworkPacket,
	    (IsNonBlocking ? TKorbNetworkAdapter::Flag_NonBlocking : 0)
	);
    }
    
    if (Length > 0) {
	OrbPacket.OnReceive();
    }

    return (Length > 0) ? 1 : 0;
}

int TKorbOrb::Flush(bool IsNonBlocking)
{
    // First just try sending. If in success, this will avoid the overhead of
    // ReceivingQueue copying.

    while (! _SendingQueue.empty()) {
	int Length = _NetworkAdapter->Send(
	    _SendingQueue.front(),
	    TKorbNetworkAdapter::Flag_NonBlocking
	);
	if (Length <= 0) {
	    break;
	}
	_SendingQueue.pop_front();
    }

    if (_SendingQueue.empty()) {
	return 0;
    }

    while (1) {
	TKorbNetworkPacket NetworkPacket;
	int Length = _NetworkAdapter->Receive(
	    NetworkPacket, TKorbNetworkAdapter::Flag_NonBlocking
	);
	if (Length <= 0) {
	    break;
	}
	_ReceivingQueue.push_back(NetworkPacket);
    }

    while (! _SendingQueue.empty()) {
	int Length = _NetworkAdapter->Send(
	    _SendingQueue.front(),
	    (IsNonBlocking ? TKorbNetworkAdapter::Flag_NonBlocking : 0)
	);
	if (Length <= 0) {
	    break;
	}
	_SendingQueue.pop_front();
    }

    return _SendingQueue.size();
}
