/* RoomVmeAccess.cc */
/* Created by Enomoto Sanshiro on 6 October 1997. */
/* Last updated by Enomoto Sanshiro on 9 August 2002. */


#include <cerrno>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/mman.h>
#include "RoomVmeAccess.hh"

using namespace std;


static const int VmeTransferDataSize[] = {
    /* this must be consistent with TVmeTarnsferMode */
    /* (declared in RoomVmeAccess.hh) */
    sizeof(Word), sizeof(DoubleWord),
    sizeof(Word), sizeof(DoubleWord),
    sizeof(Word), sizeof(DoubleWord)
};


TRoomVmeAccessProperty::TRoomVmeAccessProperty(void)
{
    VmeAddress = 0;
    TransferMode = VmeTransferMode_Unknown;
    DeviceDescriptor = -1;
    MappedAddress = 0;
    MapSize = 0;
    MapOffset = 0;
    DmaDeviceDescriptor = -1;
    InterruptNumber = 0;
    InterruptVector = 0;
    InterruptSignalId = 0;
    IsInterruptAutodisableEnabled = false;
    IsByteOrderCorrect = true;
    DeviceSpecificData = 0;
}



TRoomVmeController::TRoomVmeController(void)
{
}

TRoomVmeController::~TRoomVmeController()
{
}

TRoomVmeAccessProperty* TRoomVmeController::Open(TRoomVmeTransferMode TransferMode)
{
    if (TransferMode >= VmeTransferMode_Unknown) {
	throw THardwareException(
            "TRoomVmeController::Open()", "bad vme transfer mode"
        );	
    }

    int DeviceDescriptor = open(DeviceFileName(TransferMode), O_RDWR);
    if (DeviceDescriptor < 0) {
	throw THardwareException(
            "TRoomVmeController::Open(): open(2)",
            strerror(errno)
        );
    }

    TRoomVmeAccessProperty* Property = new TRoomVmeAccessProperty;
    Property->TransferMode = TransferMode;
    Property->VmeAddress = 0;
    Property->DeviceDescriptor = DeviceDescriptor;
    Property->DmaDeviceDescriptor = 0;
    Property->MappedAddress = 0;
    Property->MapSize = 0;
    Property->MapOffset = 0;
    Property->InterruptNumber = 0;
    Property->InterruptVector = 0;
    Property->DeviceSpecificData = 0;

    Seek(Property, 0);
    
    return Property;
}

void TRoomVmeController::Close(TRoomVmeAccessProperty* Property)
{
    DmaClose(Property);
    close(Property->DeviceDescriptor);
    delete Property;
}

long TRoomVmeController::Read(TRoomVmeAccessProperty* Property, void* Buffer, int Size)
{
    long ReadSize = read(Property->DeviceDescriptor, Buffer, Size);

    if (ReadSize < 0) {
	throw THardwareException(
            "TRoomVmeController::Read(): read(2)",
            strerror(errno)
        );
    }

    return ReadSize;
}

long TRoomVmeController::Write(TRoomVmeAccessProperty* Property, const void* Buffer, int Size)
{
    long WrittenSize = write(Property->DeviceDescriptor, Buffer, Size);

    if (WrittenSize < 0) {
	throw THardwareException(
            "TRoomVmeController::Write(): write(2)",
            strerror(errno)
        );
    }

    return WrittenSize;
}

off_t TRoomVmeController::Seek(TRoomVmeAccessProperty* Property, off_t OffsetAddress)
{
    off_t SeekPosition = Property->VmeAddress + OffsetAddress;

    off_t NewPosition;
    NewPosition = lseek(Property->DeviceDescriptor, SeekPosition, SEEK_SET);

    if (NewPosition == (off_t) -1) {
	throw THardwareException(
            "TRoomVmeController::Seek(): lseek(2)",
            strerror(errno)
        );	
    }

    return NewPosition;
}

caddr_t TRoomVmeController::Map(TRoomVmeAccessProperty* Property, off_t VmeAddress, size_t MapSize, off_t MapOffset)
{
    caddr_t MappedAddress = 0;

    if (MapSize > 0) {
	caddr_t MapAddress = (caddr_t) 0;  // assigned by kernel
	int MapProtection = PROT_READ | PROT_WRITE;
	int MapFlag = MAP_SHARED;
	MappedAddress = (caddr_t) mmap(
	    MapAddress,
	    MapSize,
	    MapProtection,
	    MapFlag,
	    Property->DeviceDescriptor,
	    VmeAddress + MapOffset
	);
	if ((long long) MappedAddress == (long long) MAP_FAILED) {
	    throw THardwareException(
		"TRoomVmeController::Map(): mmap(2)",
		strerror(errno)
	    );
	}
    }

    Property->VmeAddress = VmeAddress;
    Property->MapSize = MapSize;
    Property->MapOffset = MapOffset;
    Property->MappedAddress = MappedAddress;

    return MappedAddress;
}

void TRoomVmeController::Unmap(TRoomVmeAccessProperty* Property)
{
    if ((long long) Property->MappedAddress > 0) {
	int Result = munmap(
	    (caddr_t) Property->MappedAddress, Property->MapSize
	);
	if (Result < 0) {
	    throw THardwareException(
		"TRoomVmeController::Unmap(): munmap(2)",
		strerror(errno)
	    );	     
	}

	Property->MappedAddress = 0;
    }
}

void TRoomVmeController::PutWord(TRoomVmeAccessProperty* Property, off_t OffsetAddress, Word Value)
{
    this->Seek(Property, OffsetAddress);
    this->Write(Property, &Value, sizeof(Word));
}

Word TRoomVmeController::GetWord(TRoomVmeAccessProperty* Property, off_t OffsetAddress)
{
    Word Value;
    this->Seek(Property, OffsetAddress);
    this->Read(Property, &Value, sizeof(Word));
    return Value;
}

void TRoomVmeController::PutDoubleWord(TRoomVmeAccessProperty* Property, off_t OffsetAddress, DoubleWord Value)
{
    this->Seek(Property, OffsetAddress);
    this->Write(Property, &Value, sizeof(DoubleWord));
}

DoubleWord TRoomVmeController::GetDoubleWord(TRoomVmeAccessProperty* Property, off_t OffsetAddress)
{
    DoubleWord Value;
    this->Seek(Property, OffsetAddress);
    this->Read(Property, &Value, sizeof(DoubleWord));
    return Value;
}

void TRoomVmeController::DmaOpen(TRoomVmeAccessProperty* Property)
{
    TRoomVmeTransferMode TransferMode = Property->TransferMode;
    int DmaDeviceDescriptor = open(DmaDeviceFileName(TransferMode), O_RDWR);
    if (DmaDeviceDescriptor < 0) {
	throw THardwareException(
            "TRoomVmeController::DmaOpen(): open(2)",
            strerror(errno)
        );
    }

    Property->DmaDeviceDescriptor = DmaDeviceDescriptor;
    DmaSeek(Property, 0);
}

void TRoomVmeController::DmaClose(TRoomVmeAccessProperty* Property)
{
    if (Property->DmaDeviceDescriptor > 0) {
	close(Property->DmaDeviceDescriptor);
    }
}

long TRoomVmeController::DmaRead(TRoomVmeAccessProperty* Property, void* Buffer, int Size)
{
    if (Property->DmaDeviceDescriptor <= 0) {
	DmaOpen(Property);
    }  

    long ReadSize = read(Property->DmaDeviceDescriptor, Buffer, Size);

    if (ReadSize < 0) {
	throw THardwareException(
            "TRoomVmeController::DmaRead(): read(2)",
            strerror(errno)
        );
    }

    return ReadSize;
}

long TRoomVmeController::DmaWrite(TRoomVmeAccessProperty* Property, const void* Buffer, int Size)
{
    if (Property->DmaDeviceDescriptor <= 0) {
       DmaOpen(Property);
    }  

    long WrittenSize = write(Property->DmaDeviceDescriptor, Buffer, Size);

    if (WrittenSize < 0) {
	throw THardwareException(
            "TRoomVmeController::DmaWrite(): write(2)",
            strerror(errno)
        );
    }

    return WrittenSize;
}

off_t TRoomVmeController::DmaSeek(TRoomVmeAccessProperty* Property, off_t OffsetAddress)
{
    if (Property->DmaDeviceDescriptor <= 0) {
       DmaOpen(Property);
    }  

    off_t SeekPosition = Property->VmeAddress + OffsetAddress;

    off_t NewPosition;
    NewPosition = lseek(Property->DmaDeviceDescriptor, SeekPosition, SEEK_SET);

    if (NewPosition == (off_t) -1) {
	throw THardwareException(
            "TRoomVmeController::DmaSeek(): lseek(2)",
            strerror(errno)
        );	
    }

    return NewPosition;
}

bool TRoomVmeController::IsSignalOnInterruptAvailable(void)
{
    return false;
}



TRoomNullVmeController::TRoomNullVmeController(void)
{
}

TRoomNullVmeController::~TRoomNullVmeController()
{
}

TRoomVmeController* TRoomNullVmeController::CloneController(void)
{
    return new TRoomNullVmeController;
}

caddr_t TRoomNullVmeController::Map(TRoomVmeAccessProperty* Property, off_t VmeAddress, size_t MapSize, off_t MapOffset)
{
    caddr_t MappedAddress = (caddr_t) (new U8bit[MapSize]);
    memset(MappedAddress, 0, MapSize);

    Property->VmeAddress = VmeAddress;
    Property->MapSize = MapSize;
    Property->MapOffset = MapOffset;
    Property->MappedAddress = MappedAddress;

    return MappedAddress;
}

void TRoomNullVmeController::Unmap(TRoomVmeAccessProperty* Property)
{
    delete[] (U8bit*) Property->MappedAddress;
}

void TRoomNullVmeController::RegisterInterruptNotification(TRoomVmeAccessProperty* Property, int SignalId)
{
    Property->InterruptSignalId = SignalId;
}

void TRoomNullVmeController::UnregisterInterruptNotification(TRoomVmeAccessProperty* Property)
{
}

void TRoomNullVmeController::EnableInterruptNotification(TRoomVmeAccessProperty* Property)
{
}

void TRoomNullVmeController::DisableInterruptNotification(TRoomVmeAccessProperty* Property)
{
}

bool TRoomNullVmeController::WaitForInterruptNotification(TRoomVmeAccessProperty* Property, unsigned TimeOut_sec)
{
    sleep(TimeOut_sec);
    return false;
}

void TRoomNullVmeController::SetInterruptNotificationAutodisable(TRoomVmeAccessProperty* Property)
{
}

const char* TRoomNullVmeController::DeviceFileName(TRoomVmeTransferMode TransferMode)
{
    return "/dev/zero";
}

const char* TRoomNullVmeController::DmaDeviceFileName(TRoomVmeTransferMode TransferMode)
{
    return "/dev/zero";
}



TRoomCheckerVmeController::TRoomCheckerVmeController(void)
{
}

TRoomCheckerVmeController::~TRoomCheckerVmeController()
{
}

TRoomVmeController* TRoomCheckerVmeController::CloneController(void)
{
    return new TRoomCheckerVmeController;
}

caddr_t TRoomCheckerVmeController::Map(TRoomVmeAccessProperty* Property, off_t VmeAddress, size_t MapSize, off_t MapOffset)
{
    throw THardwareException(
	"TRoomVmeController::Map()",
	"VME controller is not connected"
    );

    return 0;
}

void TRoomCheckerVmeController::Unmap(TRoomVmeAccessProperty* Property)
{
    throw THardwareException(
	"TRoomVmeController::Unmap()",
	"VME controller is not connected"
    );
}

void TRoomCheckerVmeController::RegisterInterruptNotification(TRoomVmeAccessProperty* Property, int SignalId)
{
    throw THardwareException(
	"TRoomVmeController::RegisterInterruptNotification()",
	"VME controller is not connected"
    );
}

void TRoomCheckerVmeController::UnregisterInterruptNotification(TRoomVmeAccessProperty* Property)
{
    throw THardwareException(
	"TRoomVmeController::UnregisterInterruptNotification()",
	"VME controller is not connected"
    );
}

void TRoomCheckerVmeController::EnableInterruptNotification(TRoomVmeAccessProperty* Property)
{
    throw THardwareException(
	"TRoomVmeController::EnableInterruptNotification()",
	"VME controller is not connected"
    );
}

void TRoomCheckerVmeController::DisableInterruptNotification(TRoomVmeAccessProperty* Property)
{
    throw THardwareException(
	"TVmeController::DisableInterruptNotification()",
	"VME controller is not connected"
    );
}

bool TRoomCheckerVmeController::WaitForInterruptNotification(TRoomVmeAccessProperty* Property, unsigned TimeOut_sec)
{
    throw THardwareException(
	"TVmeController::WaitForInterruptNotification()",
	"VME controller is not connected"
    );

    return false;
}

void TRoomCheckerVmeController::SetInterruptNotificationAutodisable(TRoomVmeAccessProperty* Property)
{
    throw THardwareException(
	"TVmeController::SetInterruptNotificationAutoDisable()",
	"VME controller is not connected"
    );
}

const char* TRoomCheckerVmeController::DeviceFileName(TRoomVmeTransferMode TransferMode)
{
    throw THardwareException(
	"TRoomVmeController::DeviceFileName()",
	"VME controller is not connected"
    );

    return "/dev/zero";
}

const char* TRoomCheckerVmeController::DmaDeviceFileName(TRoomVmeTransferMode TransferMode)
{
    throw THardwareException(
	"TRoomVmeController::DmaDeviceFileName()",
	"VME controller is not connected"
    );

    return "/dev/zero";
}



TRoomVmeCrate::TRoomVmeCrate(TRoomVmeController *VmeController, int NumberOfSlots) 
{
    _VmeController = VmeController;
    _NumberOfSlots = NumberOfSlots;
}

TRoomVmeCrate::TRoomVmeCrate(int NumberOfSlots) 
{
    _VmeController = 0;
    _NumberOfSlots = NumberOfSlots;
}

TRoomVmeCrate::~TRoomVmeCrate() 
{
}

void TRoomVmeCrate::InstallController(TRoomVmeController *VmeController)
{
    _VmeController = VmeController;
}

void TRoomVmeCrate::Install(TRoomVmeModule *VmeModule, off_t VmeAddress, int InterruptNumber, int InterruptVector)
{
    if (_VmeController == 0) {
        throw THardwareException(
            "TRoomVmeCrate::Install()", "controller is not installed"
        );
    }

    TRoomVmeTransferMode TransferMode = VmeModule->TransferMode();
    size_t MapSize = VmeModule->MapSize();
    off_t MapOffset = VmeModule->MapOffset();

    TRoomVmeAccessProperty* Property = _VmeController->Open(TransferMode);
    caddr_t MappedAddress = _VmeController->Map(Property, VmeAddress, MapSize, MapOffset);
    
    Property->InterruptNumber = InterruptNumber;
    Property->InterruptVector = InterruptVector;
    Property->InterruptSignalId = 0;

    VmeModule->Attach(this, Property, MappedAddress);
    _ModuleList.push_back(VmeModule);
}

void TRoomVmeCrate::Install(TRoomVmeModule *VmeModule, off_t VmeAddress)
{
    int InterruptNumber = 0;
    int Vector = 0;

    Install(VmeModule, VmeAddress, InterruptNumber, Vector);
}

void TRoomVmeCrate::Uninstall(TRoomVmeModule *VmeModule) 
{
    TRoomVmeAccessProperty* Property = VmeModule->VmeAccessProperty();

    VmeModule->Detach();
    _VmeController->Unmap(Property);
    _VmeController->Close(Property);

}

TRoomVmeController* TRoomVmeCrate::VmeController(void) const
{
    return _VmeController;
}



TRoomVmeModule::TRoomVmeModule(TRoomVmeTransferMode TransferMode, size_t MapSize, off_t MapOffset)
: TRoomModule("Unknown-Type", "Unknown-Model")
{
    _TransferMode = TransferMode;
    _MapSize = MapSize;
    _MapOffset = MapOffset;

    _CheckerVmeController = new TRoomCheckerVmeController;
    _VmeController = _CheckerVmeController;
    _VmeCrate = 0;
    _VmeAccessProperty = 0;

    _IsInterruptRegistered = false;
    _IsInterruptEnabled = false;

    _IsByteOrderCorrect = true;
}

TRoomVmeModule::TRoomVmeModule(const string& ModuleType, const string& ModelName, TRoomVmeTransferMode TransferMode, size_t MapSize, off_t MapOffset)
: TRoomModule(ModuleType, ModelName)
{
    _TransferMode = TransferMode;
    _MapSize = MapSize;
    _MapOffset = MapOffset;

    _CheckerVmeController = new TRoomCheckerVmeController;
    _VmeController = _CheckerVmeController;
    _VmeCrate = 0;
    _VmeAccessProperty = 0;

    _IsInterruptRegistered = false;
    _IsInterruptEnabled = false;

    _IsByteOrderCorrect = true;
}

TRoomVmeModule::~TRoomVmeModule()
{
    delete _CheckerVmeController;
}

void TRoomVmeModule::Destroy(void)
{
    if (_VmeCrate != 0) {
	_VmeCrate->Uninstall(this);
    }
}

size_t TRoomVmeModule::MapSize(void) const
{
    return _MapSize;
}

off_t TRoomVmeModule::MapOffset(void) const
{
    return _MapOffset;
}

TRoomVmeTransferMode TRoomVmeModule::TransferMode(void) const
{
    return _TransferMode;
}

volatile caddr_t TRoomVmeModule::MappedAddress(void) const
{
    return _MappedAddress;
}

TRoomVmeAccessProperty* TRoomVmeModule::VmeAccessProperty(void) const
{
    return _VmeAccessProperty;
}

long TRoomVmeModule::PioRead(off_t Address, void* Buffer, int Size)
{
    _VmeController->Seek(_VmeAccessProperty, Address);
    return _VmeController->Read(_VmeAccessProperty, Buffer, Size);
}

long TRoomVmeModule::PioWrite(off_t Address, const void* Buffer, int Size)
{
    _VmeController->Seek(_VmeAccessProperty, Address);
    return _VmeController->Write(_VmeAccessProperty, Buffer, Size);
}

long TRoomVmeModule::DmaRead(off_t Address, void* Buffer, int Size)
{
    _VmeController->DmaSeek(_VmeAccessProperty, Address);
    return _VmeController->DmaRead(_VmeAccessProperty, Buffer, Size);
}

long TRoomVmeModule::DmaWrite(off_t Address, const void* Buffer, int Size)
{
    _VmeController->DmaSeek(_VmeAccessProperty, Address);
    return _VmeController->DmaWrite(_VmeAccessProperty, Buffer, Size);
}

int TRoomVmeModule::BlockRead(int Address, void* Data, int MaxSize)
{
    return DmaRead(Address, Data, MaxSize);
}

int TRoomVmeModule::BlockWrite(int Address, const void* Data, int Size)
{
    return DmaWrite(Address, Data, Size);
}

int TRoomVmeModule::ReadRegister(int Address, int& Data)
{
    return PioRead(Address, &Data, VmeTransferDataSize[_TransferMode]);
}

int TRoomVmeModule::WriteRegister(int Address, int Data)
{
    return PioWrite(Address, &Data, VmeTransferDataSize[_TransferMode]);
}

int TRoomVmeModule::EnableInterrupt(int SignalId)
{
    if (SignalId != _VmeAccessProperty->InterruptSignalId) {
	if (_IsInterruptRegistered) {
	    _VmeController->UnregisterInterruptNotification(_VmeAccessProperty);
	    _IsInterruptRegistered = false;
	}
    }

    if (! _IsInterruptRegistered) {
	_VmeController->RegisterInterruptNotification(
	    _VmeAccessProperty, SignalId
	);
	_IsInterruptRegistered = true;
	_IsInterruptEnabled = false;

	if (_VmeAccessProperty->IsInterruptAutodisableEnabled) {
	    _VmeController->SetInterruptNotificationAutodisable(
		_VmeAccessProperty
	    );
	}
    }

    if (! _IsInterruptEnabled) {
        _VmeController->EnableInterruptNotification(_VmeAccessProperty);
	_IsInterruptEnabled = true;
    }

    return 0;
}

int TRoomVmeModule::DisableInterrupt(void)
{
    if (_IsInterruptEnabled) {
	_VmeController->DisableInterruptNotification(_VmeAccessProperty);
	_IsInterruptEnabled = false;
    }
	
    return 0;
}

int TRoomVmeModule::ClearInterrupt(void)
{
    return 0;
}

bool TRoomVmeModule::WaitData(unsigned TimeOut_sec)
{
    if (! _IsInterruptEnabled) {
	EnableInterrupt();
    }

    int WaitResult = _VmeController->WaitForInterruptNotification( 
	_VmeAccessProperty, TimeOut_sec
    );

    return (WaitResult > 0);
}
  
bool TRoomVmeModule::IsSignalOnServiceRequestAvailable(void)
{
    return _VmeController->IsSignalOnInterruptAvailable();
}

void TRoomVmeModule::EnableSignalOnServiceRequest(int SignalId)
{
    EnableInterrupt(SignalId);
}

void TRoomVmeModule::DisableSignalOnServiceRequest(void)
{
    DisableInterrupt();
}

bool TRoomVmeModule::IsInterruptAutodisableEnabled(void)
{
    return false;
}

void TRoomVmeModule::Attach(TRoomVmeCrate* Crate, TRoomVmeAccessProperty* Property, caddr_t MappedAddress)
{
    _VmeCrate = Crate;
    _VmeController = Crate->VmeController();
    _VmeAccessProperty = Property;
    _MappedAddress = MappedAddress;

    _VmeAccessProperty->IsInterruptAutodisableEnabled = (
	this->IsInterruptAutodisableEnabled()
    );

    _IsByteOrderCorrect = Property->IsByteOrderCorrect;
}

void TRoomVmeModule::Detach(void)
{
    if (_IsInterruptRegistered) {
	DisableInterrupt();
	_VmeController->UnregisterInterruptNotification(_VmeAccessProperty);
	_IsInterruptRegistered = false;
    }

    _VmeController = _CheckerVmeController;

    _VmeCrate = 0;
    _VmeAccessProperty = 0;
    _MappedAddress = 0;
}
