/* RoomVmeAccess.hh */
/* Created by Enomoto Sanshiro on 6 October 1997. */
/* Last updated by Enomoto Sanshiro on 9 August 2002. */


#ifndef __RoomVmeAccess_hh__
#define __RoomVmeAccess_hh__

#include <string>
#include <vector>
#include <sys/types.h>
#include "RoomDefs.hh"
#include "RoomDeviceBase.hh"
#include "RoomModule.hh"



enum TRoomVmeTransferMode {
    VmeTransferMode_A16D16, 
    VmeTransferMode_A16D32, 
    VmeTransferMode_A24D16, 
    VmeTransferMode_A24D32, 
    VmeTransferMode_A32D16, 
    VmeTransferMode_A32D32,
    VmeTransferMode_Unknown
};


class TRoomVmeAccessProperty {
  public:
    off_t VmeAddress;
    TRoomVmeTransferMode TransferMode;
    int DeviceDescriptor;
    caddr_t MappedAddress;
    size_t MapSize;
    off_t MapOffset;
    int DmaDeviceDescriptor;
    int InterruptNumber;
    int InterruptVector;
    int InterruptSignalId;
    bool IsInterruptAutodisableEnabled;
    bool IsByteOrderCorrect;
    void* DeviceSpecificData;
    TRoomVmeAccessProperty(void);
};


class TRoomVmeModule;

class TRoomVmeController: public TRoomController {
  public:
    TRoomVmeController(void);
    virtual ~TRoomVmeController();
    virtual TRoomVmeController* CloneController(void) = 0;
  public:
    virtual TRoomVmeAccessProperty* Open(TRoomVmeTransferMode TransferMode);
    virtual void Close(TRoomVmeAccessProperty* Property);
    // PIO & Mapped I/O //
    virtual long Read(TRoomVmeAccessProperty* Property, void* Buffer, int Size);
    virtual long Write(TRoomVmeAccessProperty* Property, const void* Buffer, int Size);
    virtual off_t Seek(TRoomVmeAccessProperty* Property, off_t OffsetAddress);
    virtual caddr_t Map(TRoomVmeAccessProperty* Property, off_t VmeAddress, size_t MapSize, off_t MapOffset);
    virtual void Unmap(TRoomVmeAccessProperty* Property);
    // PIO when Mapped I/O is not available //
    virtual void PutWord(TRoomVmeAccessProperty* Property, off_t OffsetAddress, Word Value);
    virtual Word GetWord(TRoomVmeAccessProperty* Property, off_t OffsetAddress);
    virtual void PutDoubleWord(TRoomVmeAccessProperty* Property, off_t OffsetAddress, DoubleWord Value);
    virtual DoubleWord GetDoubleWord(TRoomVmeAccessProperty* Property, off_t OffsetAddress);
    // Interrupt Handling //
    virtual void RegisterInterruptNotification(TRoomVmeAccessProperty* Property, int SignalId = 0) = 0;
    virtual void UnregisterInterruptNotification(TRoomVmeAccessProperty* Property) = 0;
    virtual void EnableInterruptNotification(TRoomVmeAccessProperty* Property) = 0;
    virtual void DisableInterruptNotification(TRoomVmeAccessProperty* Property) = 0;
    virtual bool WaitForInterruptNotification(TRoomVmeAccessProperty* Property, unsigned TimeOut_sec) = 0;
    virtual void SetInterruptNotificationAutodisable(TRoomVmeAccessProperty* Property) = 0;
    virtual bool IsSignalOnInterruptAvailable(void);
    // DMA //
    virtual long DmaRead(TRoomVmeAccessProperty* Property, void* Buffer, int Size);
    virtual long DmaWrite(TRoomVmeAccessProperty* Property, const void* Buffer, int Size);
    virtual off_t DmaSeek(TRoomVmeAccessProperty* Property, off_t OffsetAddress);
  protected:
    virtual const char* DeviceFileName(TRoomVmeTransferMode TransferMode) = 0;
    virtual const char* DmaDeviceFileName(TRoomVmeTransferMode TransferMode) = 0;
    virtual void DmaOpen(TRoomVmeAccessProperty* Property);
    virtual void DmaClose(TRoomVmeAccessProperty* Property);
};


class TRoomNullVmeController: public TRoomVmeController {
  public:
    TRoomNullVmeController(void);
    virtual ~TRoomNullVmeController();
    virtual TRoomVmeController* CloneController(void);
  public:
    virtual caddr_t Map(TRoomVmeAccessProperty* Property, off_t VmeAddress, size_t MapSize, off_t MapOffset);
    virtual void Unmap(TRoomVmeAccessProperty* Property);
    virtual void RegisterInterruptNotification(TRoomVmeAccessProperty* Property, int SignalId = 0);
    virtual void UnregisterInterruptNotification(TRoomVmeAccessProperty* Property);
    virtual void EnableInterruptNotification(TRoomVmeAccessProperty* Property);
    virtual void DisableInterruptNotification(TRoomVmeAccessProperty* Property);
    virtual bool WaitForInterruptNotification(TRoomVmeAccessProperty* Property, unsigned TimeOut_sec);
    virtual void SetInterruptNotificationAutodisable(TRoomVmeAccessProperty* Property);
  protected:
    virtual const char* DeviceFileName(TRoomVmeTransferMode TransferMode);
    virtual const char* DmaDeviceFileName(TRoomVmeTransferMode TransferMode);
};


class TRoomCheckerVmeController: public TRoomVmeController {
  public:
    TRoomCheckerVmeController(void);
    virtual ~TRoomCheckerVmeController();
    virtual TRoomVmeController* CloneController(void);
  public:
    virtual caddr_t Map(TRoomVmeAccessProperty* Property, off_t VmeAddress, size_t MapSize, off_t MapOffset);
    virtual void Unmap(TRoomVmeAccessProperty* Property);
    virtual void RegisterInterruptNotification(TRoomVmeAccessProperty* Property, int SignalId = 0);
    virtual void UnregisterInterruptNotification(TRoomVmeAccessProperty* Property);
    virtual void EnableInterruptNotification(TRoomVmeAccessProperty* Property);
    virtual void DisableInterruptNotification(TRoomVmeAccessProperty* Property);
    virtual bool WaitForInterruptNotification(TRoomVmeAccessProperty* Property, unsigned TimeOut_sec);
    virtual void SetInterruptNotificationAutodisable(TRoomVmeAccessProperty* Property);
  protected:
    virtual const char* DeviceFileName(TRoomVmeTransferMode TransferMode);
    virtual const char* DmaDeviceFileName(TRoomVmeTransferMode TransferMode);
};


class TRoomVmeCrate: public TRoomCrate {
  public:
    TRoomVmeCrate(TRoomVmeController *VmeController, int NumberOfSlots = 21);
    TRoomVmeCrate(int NumberOfSlots = 21);
    virtual ~TRoomVmeCrate();
    virtual void InstallController(TRoomVmeController *VmeController);
    virtual void Install(TRoomVmeModule *VmeModule, off_t VmeAddress);
    virtual void Install(TRoomVmeModule *VmeModule, off_t VmeAddress, int InterruptNumber, int InterruptVector);
    virtual void Uninstall(TRoomVmeModule *VmeModule);
    virtual TRoomVmeController* VmeController(void) const;
  protected:
    TRoomVmeController *_VmeController;
    std::vector<TRoomVmeModule*> _ModuleList;
    int _NumberOfSlots;
};


class TRoomVmeModule: public TRoomModule {
  public:
    TRoomVmeModule(TRoomVmeTransferMode TransferMode, size_t MapSize = 0, off_t MapOffset = 0);
    TRoomVmeModule(const std::string& ModuleType, const std::string& ModelName, TRoomVmeTransferMode TransferMode, size_t MapSize = 0, off_t MapOffset = 0);
    virtual ~TRoomVmeModule();
    virtual TRoomVmeModule* Clone(void) = 0;
    virtual void Destroy(void);
  public:
    // VME access functions //
    TRoomVmeTransferMode TransferMode(void) const;
    volatile caddr_t MappedAddress(void) const;
    size_t MapSize(void) const;
    off_t MapOffset(void) const;
    long PioRead(off_t Address, void* Buffer, int Size);
    long PioWrite(off_t Address, const void* Buffer, int Size);
    long DmaRead(off_t Address, void* Buffer, int Size);
    long DmaWrite(off_t Address, const void* Buffer, int Size);
  public:
    inline Word CorrectEndian16(Word Value);
    inline DoubleWord CorrectEndian32(DoubleWord Value);
    // Accessors with no endian correction (direct mapping) //
    inline volatile Word* WordPointer(off_t OffsetAddress = 0);
    inline volatile Word& WordAt(off_t OffsetAddress);
    inline volatile DoubleWord* DoubleWordPointer(off_t OffsetAddress = 0);
    inline volatile DoubleWord& DoubleWordAt(off_t OffsetAddress);
    // Accessors with endian correction //
    inline void PutWord(off_t OffsetAddress, Word Value);
    inline volatile Word GetWord(off_t OffsetAddress);
    inline void PutDoubleWord(off_t OffsetAddress, DoubleWord Value);
    inline volatile DoubleWord GetDoubleWord(off_t OffsetAddress);
  public:
    // Common implementations of the Module interface //
    virtual int BlockRead(int Address, void* Data, int MaxSize);
    virtual int BlockWrite(int Address, const void* Data, int Size);
    virtual int ReadRegister(int Address, int& Data);
    virtual int WriteRegister(int Address, int Data);
    virtual int EnableInterrupt(int SignalId = 0);
    virtual int DisableInterrupt(void);
    virtual int ClearInterrupt(void);
    virtual bool WaitData(unsigned TimeOut_sec);
  public:
    // Common implementations of the ServiceRequester interface //
    virtual bool IsSignalOnServiceRequestAvailable(void);
    virtual void EnableSignalOnServiceRequest(int SignalId);
    virtual void DisableSignalOnServiceRequest(void);
  public:
    // Override the following methods if necessary //
    virtual bool IsInterruptAutodisableEnabled(void);
  protected:
    TRoomVmeAccessProperty* VmeAccessProperty(void) const;
  private:
    friend class TRoomVmeCrate;
    void Attach(TRoomVmeCrate* Crate, TRoomVmeAccessProperty* Property, caddr_t MappedAddress);
    void Detach(void);
  private:
    TRoomVmeCrate* _VmeCrate;
    TRoomVmeController* _VmeController;
    TRoomVmeController* _CheckerVmeController;
    TRoomVmeAccessProperty* _VmeAccessProperty;
    TRoomVmeTransferMode _TransferMode;
    volatile caddr_t _MappedAddress;
    size_t _MapSize;
    off_t _MapOffset;
    bool _IsInterruptRegistered;
    bool _IsInterruptEnabled;
    bool _IsByteOrderCorrect;
};



inline Word TRoomVmeModule::CorrectEndian16(Word Value)
{
    return _IsByteOrderCorrect ? Value : SwapByte16(Value);
}

inline DoubleWord TRoomVmeModule::CorrectEndian32(DoubleWord Value)
{
    return _IsByteOrderCorrect ? Value : SwapByte32(Value);
}

inline volatile Word* TRoomVmeModule::WordPointer(off_t OffsetAddress) 
{
    return (volatile Word*) (_MappedAddress + OffsetAddress);
}

inline volatile Word& TRoomVmeModule::WordAt(off_t OffsetAddress) 
{
    return *((volatile Word*) (_MappedAddress + OffsetAddress));
}

inline volatile DoubleWord* TRoomVmeModule::DoubleWordPointer(off_t OffsetAddress) 
{
    return (volatile DoubleWord *) (_MappedAddress + OffsetAddress);
}

inline volatile DoubleWord& TRoomVmeModule::DoubleWordAt(off_t OffsetAddress) 
{
    return *((volatile DoubleWord *) (_MappedAddress + OffsetAddress));
}

inline void TRoomVmeModule::PutWord(off_t OffsetAddress, Word Value) 
{
    if ((long long) _MappedAddress > 0) {
        *((volatile Word*) (_MappedAddress + OffsetAddress)) = CorrectEndian16(Value);
    }
    else {
        _VmeController->PutWord(_VmeAccessProperty, OffsetAddress, CorrectEndian16(Value));
    }
}

inline volatile Word TRoomVmeModule::GetWord(off_t OffsetAddress) 
{
    if ((long long) _MappedAddress > 0) {
        return CorrectEndian16(*((volatile Word*) (_MappedAddress + OffsetAddress)));
    }
    else {
        return CorrectEndian16(_VmeController->GetWord(_VmeAccessProperty, OffsetAddress));
    }
}

inline void TRoomVmeModule::PutDoubleWord(off_t OffsetAddress, DoubleWord Value) 
{
    if ((long long) _MappedAddress > 0) {
        *((volatile DoubleWord *) (_MappedAddress + OffsetAddress)) = CorrectEndian32(Value);
    }
    else {
        _VmeController->PutDoubleWord(_VmeAccessProperty, OffsetAddress, CorrectEndian32(Value));
    }
}

inline volatile DoubleWord TRoomVmeModule::GetDoubleWord(off_t OffsetAddress) 
{
    if ((long long) _MappedAddress > 0) {
        return CorrectEndian32(*((volatile DoubleWord *) (_MappedAddress + OffsetAddress)));
    }
    else {
        return CorrectEndian32(_VmeController->GetDoubleWord(_VmeAccessProperty, OffsetAddress));
    }
}


#endif
