/* controller-UniverseII.cc */
/* Created by Enomoto Sanshiro on 21 April 2014. */


#include <sys/types.h>
#include "RoomDeviceFactory.hh"
#include "controller-UniverseII.hh"

#include <vmelib.h>

using namespace std;


static TRoomVmeControllerCreator Creator(
    "UniverseII", new TVmeController_UniverseII()
);

static const int AddressWidthTable[] = {
    /* this must be consistent with TRoomVmeTarnsferMode */
    /* (declared in RoomVmeAccess.hh) */
    A16, A16,
    A24, A24,
    A32, A32
};
static const int DataWidthTable[] = {
    /* this must be consistent with TRoomVmeTarnsferMode */
    /* (declared in RoomVmeAccess.hh) */
    D16, D32,
    D16, D32,
    D16, D32
};


struct TDeviceSpecificData {
    off_t SeekPosition;
    caddr_t DmaBuffer;
};


TVmeController_UniverseII::TVmeController_UniverseII(void)
{
    _VmeBridge = 0;
    _OpenCount = 0;
    _DmaOpenCount = 0;
}

TVmeController_UniverseII::~TVmeController_UniverseII()
{
    if (_VmeBridge) {
        delete _VmeBridge;
    }
}

TRoomVmeController* TVmeController_UniverseII::CloneController(void)
{
    return new TVmeController_UniverseII();
}

const char* TVmeController_UniverseII::DeviceFileName(TRoomVmeTransferMode TransferMode)
{
    return 0;
}

const char* TVmeController_UniverseII::DmaDeviceFileName(TRoomVmeTransferMode TransferMode)
{
    return 0;
}

TRoomVmeAccessProperty* TVmeController_UniverseII::Open(TRoomVmeTransferMode TransferMode)
{
    if (TransferMode >= VmeTransferMode_Unknown) {
	throw THardwareException(
            "TVmeController_UniverseII::Open()", "bad vme transfer mode"
        );	
    }

    if (! _VmeBridge) {
        _VmeBridge = new VMEBridge();
        _OpenCount = 0;
    }
    _OpenCount++;

    TRoomVmeAccessProperty* Property = new TRoomVmeAccessProperty;
    Property->TransferMode = TransferMode;
    Property->VmeAddress = 0;
    Property->DeviceDescriptor = -1;
    Property->DmaDeviceDescriptor = -1;
    Property->MappedAddress = 0;
    Property->MapSize = 0;
    Property->MapOffset = 0;
    Property->InterruptNumber = 0;
    Property->InterruptVector = 0;

    //... TODO: check CPU's byte order. Here we just assume Intel...
    Property->IsByteOrderCorrect = false;

    Property->DeviceSpecificData = new TDeviceSpecificData();
    ((TDeviceSpecificData*) Property->DeviceSpecificData)->SeekPosition = 0;
    ((TDeviceSpecificData*) Property->DeviceSpecificData)->DmaBuffer = 0;
    
    return Property;
}

void TVmeController_UniverseII::Close(TRoomVmeAccessProperty* Property)
{
    DmaClose(Property);
    delete (TDeviceSpecificData*) Property->DeviceSpecificData;

    delete Property;

    _OpenCount--;
    if (_OpenCount <= 0) {
        delete _VmeBridge;
        _VmeBridge = 0;
    }
}

long TVmeController_UniverseII::Read(TRoomVmeAccessProperty* Property, void* Buffer, int Size)
{
    throw THardwareException(
        "TVmeController_UniverseII::Read()", "function not implemented"
    );

    return -1;
}

long TVmeController_UniverseII::Write(TRoomVmeAccessProperty* Property, const void* Buffer, int Size)
{
    throw THardwareException(
        "TVmeController_UniverseII::Write()", "function not implemented"
    );

    return -1;
}

off_t TVmeController_UniverseII::Seek(TRoomVmeAccessProperty* Property, off_t OffsetAddress)
{
    ((TDeviceSpecificData*) Property->DeviceSpecificData)->SeekPosition = OffsetAddress;
    return OffsetAddress;
}

caddr_t TVmeController_UniverseII::Map(TRoomVmeAccessProperty* Property, off_t VmeAddress, size_t MapSize, off_t MapOffset)
{
    int AddressWidth = AddressWidthTable[Property->TransferMode];
    int DataWidth = AddressWidthTable[Property->TransferMode];
    if (Property->DeviceDescriptor >= 0) {
        _VmeBridge->releaseImage(Property->DeviceDescriptor);
    }
    Property->DeviceDescriptor = _VmeBridge->getImage(
        VmeAddress+MapOffset, MapSize, AddressWidth, DataWidth, MASTER
    );
    if (Property->DeviceDescriptor < 0) {
        throw THardwareException(
            "TVmeController_UniverseII::Map()",
            "Failure on VMEBridge::getImage()"
	);
    }

    Property->MappedAddress = (caddr_t) _VmeBridge->getPciBaseAddr(
        Property->DeviceDescriptor
    );

    Property->VmeAddress = VmeAddress+MapOffset;
    Property->MapSize = MapSize;
    Property->MapOffset = 0;
    
    return Property->MappedAddress;
}

void TVmeController_UniverseII::Unmap(TRoomVmeAccessProperty* Property)
{
    if (Property->DeviceDescriptor >= 0) {
        _VmeBridge->releaseImage(Property->DeviceDescriptor);
        Property->VmeAddress = 0;
        Property->MapSize = 0;
        Property->MapOffset = 0;
        Property->MappedAddress = 0;
    }
}

void TVmeController_UniverseII::DmaOpen(TRoomVmeAccessProperty* Property)
{
    TDeviceSpecificData* Dev = (
        (TDeviceSpecificData*) Property->DeviceSpecificData
    );

    if (! Dev->DmaBuffer) {
        caddr_t DmaBuffer = (caddr_t) _VmeBridge->requestDMA();
        if (! DmaBuffer) {
            throw THardwareException(
                "TVmeController_UniverseII::DmaOpen()",
                "Failure on VMEBridge::requestDMA()"
            );
        }
        Dev->DmaBuffer = DmaBuffer;
        _VmeBridge->setOption(DMA, BLT_ON);

        _DmaOpenCount = 0;
    }
    _DmaOpenCount++;
}

void TVmeController_UniverseII::DmaClose(TRoomVmeAccessProperty* Property)
{
    _DmaOpenCount--;
    if (_DmaOpenCount <= 0) {
        _VmeBridge->releaseDMA();
        ((TDeviceSpecificData*) Property->DeviceSpecificData)->DmaBuffer = 0;
    }
}

long TVmeController_UniverseII::DmaRead(TRoomVmeAccessProperty* Property, void* Buffer, int Size)
{
    TDeviceSpecificData* Dev = (
        (TDeviceSpecificData*) Property->DeviceSpecificData
    );
    if (! Dev->DmaBuffer) {
        DmaOpen(Property);
    }

    off_t VmeAddress = Dev->SeekPosition;
    int AddressWidth = AddressWidthTable[Property->TransferMode];
    int DataWidth = DataWidthTable[Property->TransferMode];

    off_t DataOffset = _VmeBridge->DMAread(
        VmeAddress, Size, AddressWidth, DataWidth
    );
    if (DataOffset < 0) {
	throw THardwareException(
            "TVmeController_UniverseII::DmaRead()",
            "failure on VMEBridge::DMAread()"
        );
    }
    Dev->SeekPosition += Size;

    if (Property->IsByteOrderCorrect) {
        memcpy(Buffer, Dev->DmaBuffer + DataOffset, Size);
    }
    else {
        if (DataWidth == D16) {
            Word* Source = (Word*) (Dev->DmaBuffer + DataOffset);
            Word* Dest = (Word*) Buffer;
            int Count = Size / 2;
            for (int i = 0; i < Count; i++) {
                Dest[i] = SwapByte16(Source[i]);
            }
        }
        else {
            DoubleWord* Source = (DoubleWord*) (Dev->DmaBuffer + DataOffset);
            DoubleWord* Dest = (DoubleWord*) Buffer;
            int Count = Size / 2;
            for (int i = 0; i < Count; i++) {
                Dest[i] = SwapByte32(Source[i]);
            }
        }
    }

    return Size;
}

long TVmeController_UniverseII::DmaWrite(TRoomVmeAccessProperty* Property, const void* Buffer, int Size)
{
    throw THardwareException(
        "TVmeController_UniverseII::DmaWrite()", "function not implemented"
    );

    return -1;
}

off_t TVmeController_UniverseII::DmaSeek(TRoomVmeAccessProperty* Property, off_t OffsetAddress)
{
    return Seek(Property, OffsetAddress);
}

bool TVmeController_UniverseII::IsSignalOnInterruptAvailable(void)
{
    return false;
}

void TVmeController_UniverseII::RegisterInterruptNotification(TRoomVmeAccessProperty* Property, int SignalId)
{
    if (Property->DeviceDescriptor < 0) {
        throw THardwareException(
            "TVmeController_UniverseII::RegisterInterruptNotification()",
            "The device must be Map()ped before interrupt registration"
	);
    }

    _VmeBridge->setupIrq(
        Property->DeviceDescriptor,   // image
        Property->InterruptNumber, Property->InterruptVector,  // IRQ, Vector
        0, 0, 0, 0   // "enable register" and "clear register", unused
    );
}

void TVmeController_UniverseII::UnregisterInterruptNotification(TRoomVmeAccessProperty* Property)
{
    _VmeBridge->freeIrq(
        Property->DeviceDescriptor,
        Property->InterruptNumber, Property->InterruptVector
    );
}

void TVmeController_UniverseII::EnableInterruptNotification(TRoomVmeAccessProperty* Property)
{
}

void TVmeController_UniverseII::DisableInterruptNotification(TRoomVmeAccessProperty* Property)
{
}

bool TVmeController_UniverseII::WaitForInterruptNotification(TRoomVmeAccessProperty* Property, unsigned TimeOut)
{
    int result = _VmeBridge->waitIrq(
        Property->InterruptNumber, Property->InterruptVector, TimeOut
    );

    return (result == 0);  //... TEST THIS ...//
}

void TVmeController_UniverseII::SetInterruptNotificationAutodisable(TRoomVmeAccessProperty* Property)
{
    throw THardwareException(
        "TRoomVmeController::SetInterruptNotificationAutodisable()", 
	"function is not available for controller: UniverseII"
    );
}
