/* KinokoBuilder.cc */
/* Created by Enomoto Sanshiro on 25 August 2009. */
/* Updated by Enomoto Sanshiro on 19 May 2010. */
/* Updated by Enomoto Sanshiro on 7 July 2010. */
/* Last updated by Enomoto Sanshiro on 26 November 2010. */


#include <string>
#include <iostream>
#include <vector>
#include <exception>
#include "KinokoDataProcessor.hh"
#include "KinokoBlockDataSection.hh"
#include "KinokoMemoryPool.hh"
#include "KinokoBuilderInlet.hh"
#include "KinokoEventChopper.hh"
#include "KinokoEventSorter.hh"
#include "KinokoSorterProcessor.hh"
#include "KinokoEventEnvelope.hh"
#include "KinokoSorterController.hh"
#include "KinokoBuilderMonitor.hh"
#include "KinokoBuilder.hh"

using namespace std;

//#define DEBUG(x) x
#define DEBUG(x)

static const int InitialBufferSize = 0x10000;
static const long DefaultReportInterval = 300;


TKinokoBuilder::TKinokoBuilder(void)
{
    _MyLogger = new TKinokoBuilderNullLogger();
    _MyMonitor = new TKinokoBuilderMonitor();

    _Logger = _MyLogger;
    _Monitor = _MyMonitor;
    _SorterController = 0;

    _MemoryPool = 0;
    _BufferSize = InitialBufferSize;

    _EnvelopeWriter = 0;

    _StartDataSize = 0;
    _StartPacketCycles = 0;

    _IsInputPreserved = false;

    _ReportInterval = DefaultReportInterval;
    _LastReportTime = 0;

    DEBUG(cout << "debug flag enabled for KinokoBuilder.cc" << endl);
}

TKinokoBuilder::~TKinokoBuilder()
{
    delete _SorterController;
    delete _EnvelopeWriter;

    if (_MemoryPool) {
	_MemoryPool->Finalize();
	delete _MemoryPool;
    }

    delete _MyMonitor;
    delete _MyLogger;
}

void TKinokoBuilder::SetOutputDataSourceName(const std::string& DataSourceName)
{
    _OutputDataSourceName = DataSourceName;
}

void TKinokoBuilder::SetBufferSize(size_t BufferSize)
{
    if (_BufferSize > 0) {
	_BufferSize = BufferSize;
    }
    else {
	_BufferSize = InitialBufferSize;
    }
}

void TKinokoBuilder::SetStartDataSize(size_t StartDataSize)
{
    _StartDataSize = StartDataSize;
}

void TKinokoBuilder::SetStartPacketCycles(int NumberOfPacketCycles)
{
    _StartPacketCycles = NumberOfPacketCycles;
}

void TKinokoBuilder::SetStartChannelList(const std::vector<long>& ChannelList)
{
    _StartChannelList = ChannelList;
}

void TKinokoBuilder::SetBuilderLogger(TKinokoBuilderLogger* Logger)
{
    _Logger = Logger;
}

void TKinokoBuilder::SetBuilderMonitor(TKinokoBuilderMonitor* Monitor)
{
    _Monitor = Monitor;
}

void TKinokoBuilder::SetInputPreservation(bool IsInputPreserved)
{
    _IsInputPreserved = IsInputPreserved;
}

void TKinokoBuilder::SetActivityWindowLength(int WindowLength)
{
    //... for backward compatibility ...//
    TKinokoChannelWatcher::SetDefaultActivityWindowLength(WindowLength);
}

void TKinokoBuilder::SetReportInterval(long Interval_sec)
{
    _ReportInterval = Interval_sec;
}

void TKinokoBuilder::Construct(const std::string& Name, TKinokoInputStream* InputStream, TKinokoOutputStream* OutputStream, int DataSourceId)
{
    if (_OutputDataSourceName.empty()) {
	_OutputDataSourceName = Name;
    }
    
    TKinokoDataProcessor::Construct(
	_OutputDataSourceName, InputStream, OutputStream, DataSourceId
    );
}

void TKinokoBuilder::OnConstruct(void) 
{
    if (_MemoryPool) {
	return;
    }

    _MemoryPool = new TKinokoMemoryPool(_BufferSize);
    _MemoryPool->Initialize();

    _EnvelopeWriter = new TKinokoEventEnvelopeWriter(this);
    _EnvelopeWriter->SetLogger(_Logger);

    TKinokoEventSorter& Sorter = BuilderConstruct();
    _SorterController = new TKinokoSorterController(Sorter, _MemoryPool);
    _Monitor->SetSorter(&Sorter);

    if (! _StartChannelList.empty()) {
	for (unsigned i = 0; i < _StartChannelList.size(); i++) {
	    _SorterController->RegisterStream(_StartChannelList[i]);
	}
    }
    if (_StartDataSize != 0) {
	_SorterController->SetStartDataSize(_StartDataSize);
    }
    if (_StartPacketCycles != 0) {
	_SorterController->SetStartPacketCycles(_StartPacketCycles);
    }
}

void TKinokoBuilder::OnDestruct(void)
{
    if (_MemoryPool == 0) {
	return;
    }

    _Monitor->Report();
    _Monitor->ReportSummary();

    delete _SorterController;
    delete _EnvelopeWriter;

    _MemoryPool->Finalize();
    delete _MemoryPool;

    _SorterController = 0;
    _EnvelopeWriter = 0;
    _MemoryPool = 0;
}

void TKinokoBuilder::BuildDataSource(TKinokoDataSource* DataSource)
{
    if (_EnvelopeWriter == 0) {
	OnConstruct();
    }
    _EnvelopeWriter->BuildDataSource(DataSource);
}

void TKinokoBuilder::OnRunBegin(void)
{
    _SorterController->ProcessRunBegin();
    BuilderBegin();
}

void TKinokoBuilder::OnRunEnd(void)
{
    BuilderEnd();

    DEBUG(cout << "##### SORTER STATUS ON THE RUN END #####" << endl);
    DEBUG(_Monitor->Dump(cout));
    DEBUG(cout << "##### - E - N - D - of - E - N - D #####" << endl);
}

void TKinokoBuilder::OnReceiveCommandPacket(void* Packet, long PacketSize)
{
    long Timestamp = TKinokoDataStreamScanner::CommandTimeOf(Packet);
    if (Timestamp - _LastReportTime > _ReportInterval) {
        _Monitor->Report();
        _LastReportTime = Timestamp;
    }
}

void TKinokoBuilder::OnReceivePacket(void* Packet, long PacketSize)
{
    if (BuilderProcess(Packet, PacketSize)) {
	DEBUG(cout << "##### SORTER STATUS #####" << endl);
	DEBUG(_Monitor->Dump(cout));
	DEBUG(cout << "##### - E - N - D - #####" << endl);

	_SorterController->DoTransaction();

	if (_IsInputPreserved) {
	    SendPacket(Packet, PacketSize);
	}
    }
    else {
	SendPacket(Packet, PacketSize);
    }
}
