/* KinokoBuilderMonitor.cc */
/* Created by Enomoto Sanshiro on 1 September 2009. */
/* Last updated by Enomoto Sanshiro on 7 July 2010. */


#include <string>
#include <iostream>
#include <sstream>
#include "KinokoMemoryPool.hh"
#include "KinokoEventSorter.hh"
#include "KinokoBuilderLogger.hh"
#include "KinokoBuilderMonitor.hh"

using namespace std;


TKinokoBuilderMonitor::TKinokoBuilderMonitor(void)
{
}

TKinokoBuilderMonitor::~TKinokoBuilderMonitor(void)
{
}

void TKinokoBuilderMonitor::SetSorter(const TKinokoEventSorter* Sorter)
{
    _Sorter = Sorter;
}




TKinokoBuilderCompositeMonitor::TKinokoBuilderCompositeMonitor(void)
{
}

TKinokoBuilderCompositeMonitor::~TKinokoBuilderCompositeMonitor(void)
{
}

void TKinokoBuilderCompositeMonitor::AddMonitor(TKinokoBuilderMonitor* Monitor)
{
    _MonitorList.push_back(Monitor);
}

void TKinokoBuilderCompositeMonitor::SetSorter(const TKinokoEventSorter* Sorter)
{
    for (unsigned i = 0; i < _MonitorList.size(); i++) {
        _MonitorList[i]->SetSorter(Sorter);
    }
}

void TKinokoBuilderCompositeMonitor::Report(void)
{
    for (unsigned i = 0; i < _MonitorList.size(); i++) {
        _MonitorList[i]->Report();
    }
}

void TKinokoBuilderCompositeMonitor::ReportSummary(void)
{
    for (unsigned i = 0; i < _MonitorList.size(); i++) {
        _MonitorList[i]->ReportSummary();
    }
}



TKinokoBuilderLoggingMonitor::TKinokoBuilderLoggingMonitor(TKinokoBuilderLogger* Logger)
{
    _Logger = Logger;
}

TKinokoBuilderLoggingMonitor::~TKinokoBuilderLoggingMonitor()
{
}

void TKinokoBuilderLoggingMonitor::Report(void)
{
    if (_Sorter == 0) {
	return;
    }

    {
	ostringstream os;
	os << "built: " << _Sorter->NumberOfBuiltEvents() << ", ";
	os << "pending: " << _Sorter->HoldingEventSet().size() << ", ";
        os << "waiting: " << _Sorter->NumberOfPendingEarlyArrivals() << ", ";
	os << "last built key: " << _Sorter->LastBuiltEventKey() << ", ";
	os << "head:tail key: ";
	if (_Sorter->HoldingEventSet().empty()) {
	    os << "none:none";
	}
	else {
	    os << _Sorter->HoldingEventSet().begin()->first;
	    os << ":" << _Sorter->HoldingEventSet().rbegin()->first;
	}
	_Logger->Write(os.str(), TKinokoBuilderLogger::mlLog);
    }

    {
	ostringstream os;
	const TKinokoMemoryPool& MemoryPool = _Sorter->MemoryPool();
	os << "buffer used/limit: " << MemoryPool.UsedSize();
	os << "/" << MemoryPool.Capacity() << ", ";
	os << "entries: " << MemoryPool.NumberOfEntries();
	_Logger->Write(os.str(), TKinokoBuilderLogger::mlLog);
    }
}

void TKinokoBuilderLoggingMonitor::ReportSummary(void)
{
    {
        ostringstream os;
        os << "# NumberOfEvents: ";
        os << _Sorter->NumberOfBuiltEvents();
        _Logger->WriteInformation(os.str());
    }
    {
        ostringstream os;
        os << "# NumberOfPieces: ";
        os << _Sorter->NumberOfPieces();
        _Logger->WriteInformation(os.str());
    }
    {
        ostringstream os;
        os << "# NumberOfEarlyArrivals: ";
        os << _Sorter->NumberOfEarlyArrivals();
        _Logger->WriteInformation(os.str());
    }
    {
        ostringstream os;
        os << "# NumberOfLateArrivals: ";
        os << _Sorter->NumberOfLateArrivals();
        _Logger->WriteInformation(os.str());
    }
    {
        ostringstream os;
        os << "# NumberOfOutOfOrderArrivals: ";
        os << _Sorter->NumberOfOutOfOrderArrivals();
        _Logger->WriteInformation(os.str());
    }
    {
        ostringstream os;
        os << "# NumberOfCorruptedPieces: ";
        os << _Sorter->NumberOfCorruptedPieces();
        _Logger->WriteInformation(os.str());
    }
    {
        ostringstream os;
        os << "# MaximumBufferUsage: ";
        os << _Sorter->MemoryPool().MaximumUsage();
        _Logger->WriteInformation(os.str());
    }
    if (_Sorter->NumberOfLateArrivals() > 0) {
        ostringstream os;
        os << "Unaligned pieces in output: ";
        os << _Sorter->NumberOfLateArrivals() << " pieces, ";
        os << _Sorter->LateArrivalPieceSize() << " bytes";
        _Logger->WriteWarning(os.str());
    }
    if (_Sorter->NumberOfCorruptedPieces() > 0) {
        ostringstream os;
        os << "Corrupted pieces in output: ";
        os << _Sorter->NumberOfCorruptedPieces() << " pieces, ";
        os << _Sorter->CorruptedPieceSize() << " bytes";
        _Logger->WriteWarning(os.str());
    }
}

void TKinokoBuilderMonitor::Dump(std::ostream& Output)
{
    if (_Sorter == 0) {
	return;
    }

    // event queue status //
    int NumberOfBuiltEvents = _Sorter->NumberOfBuiltEvents();
    Output << "# of built: " << NumberOfBuiltEvents << endl;

    TEventKey LastBuiltEventKey = _Sorter->LastBuiltEventKey();
    if (LastBuiltEventKey >= 0) {
	Output << "last built: " << LastBuiltEventKey << endl;
    }
    else {
	Output << "last built: none" << endl;
    }

    int NumberOfShown = 0;
    Output << "pending: ";
    const map<TEventKey, pair<int, size_t> >& HoldingEventSet = (
	_Sorter->HoldingEventSet()
    );
    map<TEventKey, pair<int, size_t> >::const_iterator EventEntry;
    for (
	EventEntry = HoldingEventSet.begin();
	EventEntry != HoldingEventSet.end();
	EventEntry++
    ){
	Output << EventEntry->first << "(";
	Output << EventEntry->second.first << ",";
	Output << EventEntry->second.second << ") ";

        if (++NumberOfShown > 100) {
            Output << endl << "..... ";
            Output << (HoldingEventSet.size() - 100 - 1) << " more entries";
            Output << " ..... last entry: ";
            Output << (HoldingEventSet.rbegin())->first << "(";
            Output << (HoldingEventSet.rbegin())->second.first << ",";
            Output << (HoldingEventSet.rbegin())->second.second << ") ";
            break;
        }
    }
    if (HoldingEventSet.empty()) {
	Output << "none";
    }
    Output << endl;
    
    // buffer details //

    const TKinokoMemoryPool& MemoryPool = _Sorter->MemoryPool();
    float MemoryOccupancy = MemoryPool.UsedSize() / MemoryPool.Capacity();
    Output << "buffer: ";
    Output << "size=" << MemoryPool.UsedSize();
    Output << "(" << (int) (100.0 * MemoryOccupancy) << "%), ";
    Output << "entries=" << MemoryPool.NumberOfEntries() << endl;

    // channel details //

    const vector<TChannelKey>& ChannelKeyList = (
	_Sorter->ChannelKeyList()
    );
    const vector<TKinokoChannelWatcher>& ChannelWatcherList = (
	_Sorter->ChannelWatcherList()
    );
    const vector<TEventKey>& CompleteEventKeyList = (
	_Sorter->CompleteEventKeyList()
    );
    const vector<std::deque<TEventKey> >& EventKeyQueueList = (
	_Sorter->EventKeyQueueList()
    );
    const vector<std::set<TEventKey> >& EarlyArrivalEventKeySetList = (
	_Sorter->EarlyArrivalEventKeySetList()
    );
    for (unsigned i = 0; i < ChannelKeyList.size(); i++) {
	Output << "channel " << ChannelKeyList[i] << ": ";
	if (ChannelWatcherList[i].IsAlive()) {
	    Output << "ALIVE";
	}
	else {
	    Output << "DEAD";
	}
	Output << "(" << ChannelWatcherList[i].AbsenceCount() << "/";
	Output << (int) (100*ChannelWatcherList[i].Occupancy()) << "%), ";

	if (ChannelWatcherList[i].AbandonedCount() > 0) {
	    Output << "abandoned ";
	    Output << ChannelWatcherList[i].AbandonedCount() << ", ";
	}
	if (ChannelWatcherList[i].SuspensionCount() > 0) {
	    Output << "suspended ";
	    Output << ChannelWatcherList[i].SuspensionCount() << ", ";
	}

	if (CompleteEventKeyList[i] >= 0) {
	    Output << "completed " << CompleteEventKeyList[i] << ", ";
	}
	else {
	    Output << "no complete event, ";
	}

	if (! EventKeyQueueList[i].empty()) {
	    Output << "pending ";
	    Output << *EventKeyQueueList[i].begin() << "-";
	    Output << *EventKeyQueueList[i].rbegin() << ":";
	    Output << EventKeyQueueList[i].size() << ", ";
	}
	else {
	    Output << "no pending pieces, ";
	}

	if (! EarlyArrivalEventKeySetList[i].empty()) {
	    Output << "early arrivals ";
	    Output << *EarlyArrivalEventKeySetList[i].begin() << "-";
	    Output << *EarlyArrivalEventKeySetList[i].rbegin() << ":";
	    Output << EarlyArrivalEventKeySetList[i].size();
	}
	else {
	    Output << "no early arrivals";
	}
	Output << endl;
    }    
}



TKinokoBuilderSummaryReport::TKinokoBuilderSummaryReport(std::ostream& os)
: _Output(os)
{
}

TKinokoBuilderSummaryReport::~TKinokoBuilderSummaryReport(void)
{
}

void TKinokoBuilderSummaryReport::ReportSummary(void)
{
    _Output << "# NumberOfEvents: ";
    _Output << _Sorter->NumberOfBuiltEvents();
    _Output << endl;
    _Output << "# NumberOfPieces: ";
    _Output << _Sorter->NumberOfPieces();
    _Output << endl;
    _Output << "# NumberOfLateArrivals: ";
    _Output << _Sorter->NumberOfLateArrivals();
    _Output << endl;
    _Output << "# LateArrivalDataSize: ";
    _Output << _Sorter->LateArrivalPieceSize();
    _Output << endl;
    _Output << "# NumberOfEarlyArrivals: ";
    _Output << _Sorter->NumberOfEarlyArrivals();
    _Output << endl;
    _Output << "# NumberOfOutOfOrderArrivals: ";
    _Output << _Sorter->NumberOfOutOfOrderArrivals();
    _Output << endl;
    _Output << "# NumberOfCorruptedPieces: ";
    _Output << _Sorter->NumberOfCorruptedPieces();
    _Output << endl;
    _Output << "# CorruptedPieceDataSize: ";
    _Output << _Sorter->CorruptedPieceSize();
    _Output << endl;
    _Output << "# MaximumBufferUsage: ";
    _Output << _Sorter->MemoryPool().MaximumUsage() ;
    _Output << endl;
}
