/* kdf2root.cc */
/* Created by Enomoto Sanshiro on 4 August 2015. */
/* Last updated by Enomoto Sanshiro on 4 August 2015. */


#include <cstdlib>
#include <iostream>
#include <string>
#include <limits>
#include <TFile.h>
#include <TTree.h>
#include "MushFileSystem.hh"
#include "KinokoKdfReader.hh"

using namespace std;



class TKinokoRootFiller: public TKinokoDataAnalyzer {
  public:
    TKinokoRootFiller(const std::string& OutputFileName);
    virtual ~TKinokoRootFiller();
    virtual void ReadDataSource(TKinokoDataSource* DataSource);
    virtual int ProcessTrailerPacket(void* Packet, TKinokoDataSource* DataSource = 0);
    virtual int ProcessCommandPacket(void* Packet, TKinokoDataSource* DataSource = 0);
  protected:
    virtual int ProcessIndexedData(void* DataPacket, TKinokoIndexedDataSection* DataSection);
    virtual int ProcessTaggedData(void* DataPacket, TKinokoTaggedDataSection* DataSection);
  protected:
    std::string _DataSourceName;
    int _EventCount, _EventTime, _StartTime;
    std::vector<std::string> _FieldNameList;
    std::map<long long, int> _FieldKeyTable;
    std::vector<float> _FirstEventDataList;
  private:
    TFile* _File;
    TTree* _Tree;
    UInt_t _TimeStampBuffer;
    Float_t* _DataBuffer;
};



TKinokoRootFiller::TKinokoRootFiller(const std::string& OutputFileName)
{
    _EventCount = 0;
    _EventTime = 0;
    _StartTime = 0;

    _Tree = 0;
    _DataBuffer = 0;

    _File = new TFile(OutputFileName.c_str(), "RECREATE");
}

TKinokoRootFiller::~TKinokoRootFiller()
{
    if (_File && _Tree) {
        _Tree->Write();
    }
    delete _File;
    
    delete[] _DataBuffer;
}

void TKinokoRootFiller::ReadDataSource(TKinokoDataSource* DataSource)
{
    TKinokoDataAnalyzer::ReadDataSource(DataSource);
    _DataSourceName = DataSource->DataSourceName();
}

int TKinokoRootFiller::ProcessIndexedData(void* DataPacket, TKinokoIndexedDataSection* DataSection)
{
    TKinokoIndexedDataSectionScanner* Scanner = DataSection->Scanner();
    int SectionId = DataSection->SectionId();
    int NumberOfElements = Scanner->NumberOfElements(DataPacket);
    int SubAddressWidth = DataSection->SubAddressWidth();
    int Depth = DataSection->ElementDepth();
    if (NumberOfElements == 0) {
	return 0;
    }

    int Address, SubAddress; float Data;
    for (int Index = 0; Index < NumberOfElements; Index++) {
	Scanner->ReadFloatFrom(
	    DataPacket, Index, Address, SubAddress, Data
	);
        long long Key = (
            ((long long) SectionId << 48) |
            ((long long) Address << 16) |
            ((long long) SubAddress)
        );

	if (_EventCount == 0) {        
            //... BUG: FullSectionName may contain special characters
            ostringstream os;
	    os << DataSection->FullSectionName() << "_" << Address;
	    if (SubAddressWidth > 0) {
		os << "_" << SubAddress;
	    }
	    if (Depth > 1) {
		os << "_" << (Index % Depth);
	    }
            string FieldName = os.str();

	    _FieldKeyTable[Key] = _FieldNameList.size();
	    _FieldNameList.push_back(FieldName);
            _FirstEventDataList.push_back(Data);
	}
        else {
            map<long long, int>::iterator iter = _FieldKeyTable.find(Key);
            if (iter != _FieldKeyTable.end()) {
                _DataBuffer[iter->second] = Data;
            }
        }
    }

    return 1;
}

int TKinokoRootFiller::ProcessTaggedData(void* DataPacket, TKinokoTaggedDataSection* DataSection)
{
    TKinokoTaggedDataSectionScanner* Scanner = DataSection->Scanner();
    int SectionId = DataSection->SectionId();
    int NumberOfFields = DataSection->NumberOfFields();
    if (NumberOfFields == 0) {
	return 0;
    }
    
    string FieldName;
    float Data;
    for (int Index = 0; Index < NumberOfFields; Index++) {
	Scanner->ReadFloatFrom(DataPacket, Index, Data);
        long long Key = (
            ((long long) SectionId << 48) |
            ((long long) Index << 16) |
            ((long long) 0)
        );

        if (_EventCount == 0) {
            //... BUG: FullSectionName may contain special characters
            FieldName = (
                DataSection->FullSectionName() + "_" + DataSection->FieldNameOf(Index)
            );
	    _FieldKeyTable[Key] = _FieldNameList.size();
            _FieldNameList.push_back(FieldName);
            _FirstEventDataList.push_back(Data);
        }
        else {
            map<long long, int>::iterator iter = _FieldKeyTable.find(Key);
            if (iter != _FieldKeyTable.end()) {
                _DataBuffer[iter->second] = Data;
            }
        }
    }

    return 1;
}

int TKinokoRootFiller::ProcessTrailerPacket(void* Packet, TKinokoDataSource* DataSource)
{
    if (_FirstEventDataList.empty()) {
        return 0;
    }
    
    if (_EventCount == 0) {
        _DataBuffer = new Float_t[_FieldNameList.size()];
        _Tree = new TTree(_DataSourceName.c_str(), _DataSourceName.c_str());

        _Tree->Branch("Time", (void*) &_TimeStampBuffer, "Time/i");
        for (unsigned i = 0; i < _FieldNameList.size(); i++) {
            string Name = _FieldNameList[i];
	    string Leaf = Name + "/F";
	    _Tree->Branch(Name.c_str(), (void*) &(_DataBuffer[i]), Leaf.c_str());
	}

        for (unsigned i = 0; i < _FieldNameList.size(); i++) {
            _DataBuffer[i] = _FirstEventDataList[i];
        }
    }

    _TimeStampBuffer = (_EventTime - _StartTime);
    _Tree->Fill();

    _EventCount++;
    for (unsigned i = 0; i < _FieldNameList.size(); i++) {
        _DataBuffer[i] = numeric_limits<float>::quiet_NaN();
    }

    return 1;
}

int TKinokoRootFiller::ProcessCommandPacket(void* Packet, TKinokoDataSource* DataSource)
{
    _EventTime = TKinokoDataStreamScanner::CommandTimeOf(Packet);

    if (_StartTime == 0) {
	_StartTime = _EventTime;
    }

    return TKinokoDataAnalyzer::ProcessCommandPacket(Packet, DataSource);
}



int main(int argc, char** argv)
{
    if (argc < 2) {
	cerr << "Usage: " << argv[0];
	cerr << " InputKdfFileName [ OutputRootFileName ]" << endl;
        return -1;
    }

    string InputFileName = argv[1];
    string OutputFileName;
    if (argc > 2) {
	OutputFileName = argv[2];
    }
    else {
        // ROOT does not like '#' in file name
        string RootName = TMushFileAttribute(InputFileName).FileRootName();
        bool IsLastCharDoller = false;
        for (string::iterator i = RootName.begin(); i != RootName.end(); i++) {
            if (IsLastCharDoller) {
                if (isalpha(*i)) {
                    IsLastCharDoller = false;
                    continue;
                }
                else {
                    ;  // remove only '$'
                }
            }
            if (*i == '#') {
                continue;
            }
            if (*i == '$') {
                IsLastCharDoller = true;
                continue;
            }
            OutputFileName += *i;
        }
        OutputFileName += ".root";
    }
    string DataSourceName = "";
    
    TKinokoKdfReader* KdfReader = new TKinokoKdfReader(InputFileName);
    TKinokoRootFiller* Analyzer = new TKinokoRootFiller(OutputFileName);
    
    try {
	KdfReader->RegisterAnalyzer(DataSourceName, Analyzer);
	KdfReader->Start();
    }
    catch (TKinokoException& e) {
	cerr << "ERROR: " << e << endl;
        return EXIT_FAILURE;
    }

    delete Analyzer;
    delete KdfReader;

    return 0;
}
