/* KinokoCanvasColorScale.cc */
/* Created by Enomoto Sanshiro on 14 January 2001. */
/* Last updated by Enomoto Sanshiro on 14 January 2001. */


#include <cmath>
#include <cstdlib>
#include <sstream>
#include "KinokoCanvas.hh"
#include "KinokoCanvasImageArea.hh"
#include "KinokoCanvasColorScale.hh"

using namespace std;


TKinokoCanvasColorScale* TKinokoCanvasColorScaleFactory::Create(TKinokoCanvas* Canvas, const std::string& Name)
{
    TKinokoCanvasColorScale* Scale = 0;
    
    string Type;
    string BackgroundColor = "";
    int NumberOfColors = 256;

    istringstream is(Name);
    is >> Type >> BackgroundColor >> NumberOfColors;
    
    if (Type == "rainbow") {
        Scale = new TKinokoCanvasRainbowColorScale(Canvas, NumberOfColors);
    }
    else if (Type == "parula") {
        Scale = new TKinokoCanvasParulaColorScale(Canvas, NumberOfColors);
    }
    else if (Type == "gray") {
        Scale = new TKinokoCanvasGrayColorScale(Canvas, NumberOfColors);
    }
    else if ((Type == "gradient") || (Type == "gradientpoints")) {
        string Parameter; getline(is, Parameter);
        Scale = new TKinokoCanvasGradientColorScale(
            Canvas, NumberOfColors, Type, Parameter
        );
    }

    if (! Scale) {
        return Scale;
    }
    
    float BackgroundR = -1, BackgroundG = -1, BackgroundB = -1;
    if (BackgroundColor == "white") {
        BackgroundR = 1, BackgroundG = 1, BackgroundB = 1;
    }
    else if (BackgroundColor == "black") {
        BackgroundR = 0, BackgroundG = 0, BackgroundB = 0;
    }
    else if (BackgroundColor == "gray") {
        BackgroundR = 0.5, BackgroundG = 0.5, BackgroundB = 0.5;
    }
    else {
        istringstream is(BackgroundColor);
        int ColorHex;
        if (is >> hex >> ColorHex) {
            BackgroundR = ((ColorHex >> 16) & 0xff) / (float) 0xff;
            BackgroundG = ((ColorHex >> 8) & 0xff) / (float) 0xff;
            BackgroundB = ((ColorHex >> 0) & 0xff) / (float) 0xff;
        }
    }
    if (BackgroundR >= 0) {
        Scale->SetBackgroundColor(
            Canvas->ImageArea()->AllocateColorRgb(BackgroundR, BackgroundG, BackgroundB)
        );                
    }

    Scale->Initialize();
    
    return Scale;
}



TKinokoCanvasColorScale::TKinokoCanvasColorScale(TKinokoCanvas* Canvas, int NumberOfColors)
{
    _Canvas = Canvas;
    _ImageArea = 0;

    _NumberOfColors = NumberOfColors;
    _ColorTable = 0;

    _BackgroundColor = -1;
}

TKinokoCanvasColorScale::~TKinokoCanvasColorScale()
{
    //... TODO: release allocated color???
    
    delete[] _ColorTable;
}

void TKinokoCanvasColorScale::SetBackgroundColor(int ColorIndex)
{
    _BackgroundColor = ColorIndex;
}

void TKinokoCanvasColorScale::Initialize(void)
{
    if (_ColorTable) {
	delete[] _ColorTable;
    }

    _ImageArea = _Canvas->ImageArea();
    _ColorTable = new int[_NumberOfColors];

    CreateColorTable();
}

int TKinokoCanvasColorScale::ColorIndexOf(double Value)
{
    int ScaleIndex;
    if (Value < 0) {
	ScaleIndex = 0;
    }
    else if (Value >= 1) {
	ScaleIndex = _NumberOfColors - 1;
    }
    else {
	ScaleIndex = (int) (Value * _NumberOfColors);
    }

    return _ColorTable[ScaleIndex];
}

int TKinokoCanvasColorScale::BackgroundColorIndex(void)
{
    return (_BackgroundColor >= 0) ? _BackgroundColor : _ColorTable[0];
}



TKinokoCanvasGrayColorScale::TKinokoCanvasGrayColorScale(TKinokoCanvas* Canvas, int NumberOfColors)
: TKinokoCanvasColorScale(Canvas, NumberOfColors)
{
}

TKinokoCanvasGrayColorScale::~TKinokoCanvasGrayColorScale()
{
}

void TKinokoCanvasGrayColorScale::CreateColorTable(void)
{
    for (int i = 0; i < _NumberOfColors; i++) {
	//float v = 1.0 - (float) i / (_NumberOfColors - 1);
	float v = (float) i / (_NumberOfColors - 1);
	float r = v;
	float g = v;
	float b = v;

	int Color = _ImageArea->AllocateColorRgb(r, g, b);
	_ColorTable[i] = Color;
    }
}



TKinokoCanvasGradientColorScale::TKinokoCanvasGradientColorScale(TKinokoCanvas* Canvas, int NumberOfColors, const string& Type, const string& Parameters)
: TKinokoCanvasColorScale(Canvas, NumberOfColors)
{
    bool IsPoints = (Type == "gradientpoints");
    
    istringstream is(Parameters);
    float x;
    int ColorHex;
    while (1) {
        if (IsPoints) {
            if (!(is >> x)) {
                break;
            }
            _PointX.push_back(x);
        }
        if (! (is >> hex >> ColorHex)) {
            break;
        }
        _PointR.push_back(((ColorHex >> 16) & 0xff)/(float) 0xff);
        _PointG.push_back(((ColorHex >> 8) & 0xff)/(float) 0xff);
        _PointB.push_back(((ColorHex >> 0) & 0xff)/(float) 0xff);
    }

    if (! IsPoints) {
        int n = _PointR.size();
        _PointX.push_back(0);
        for (int k = 1; k < n; k++) {
            _PointX.push_back(k/(float)(n-1));
        }
    }
    
    // error corrections //
    if (! _PointX.empty()) {
        _PointX.front() = 0;
        _PointX.back() = 1;
    }
    for (unsigned i = 1; i < _PointX.size(); i++) {
        if (_PointX[i] < _PointX[i-1]) {
            _PointX[i] = _PointX[i-1];
        }
        if (_PointX[i] > 1) {
            _PointX[i] = 1;
        }
    }
    _PointX.push_back(100); // sentry
}

TKinokoCanvasGradientColorScale::~TKinokoCanvasGradientColorScale()
{
}

void TKinokoCanvasGradientColorScale::CreateColorTable(void)
{
    // less than two points --> fallback to gray scale
    if (_PointR.size() < 2) {
        for (int i = 0; i < _NumberOfColors; i++) {
            float v = (float) i / (_NumberOfColors - 1);
            _ColorTable[i] = _ImageArea->AllocateColorRgb(v, v, v);
        }
        return;
    }

    int k = 0;
    for (int i = 0; i < _NumberOfColors; i++) {
	float Value = (float) i / (_NumberOfColors - 1);
        while (_PointX[k+1] < Value) {  // we have a sentry
            k++;
        }
        float Offset = (Value - _PointX[k]) / (_PointX[k+1] - _PointX[k]);
        float r = _PointR[k] + Offset * (_PointR[k+1] - _PointR[k]);
        float g = _PointG[k] + Offset * (_PointG[k+1] - _PointG[k]);
        float b = _PointB[k] + Offset * (_PointB[k+1] - _PointB[k]);
	_ColorTable[i] = _ImageArea->AllocateColorRgb(r, g, b);
    }
}



TKinokoCanvasRainbowColorScale::TKinokoCanvasRainbowColorScale(TKinokoCanvas* Canvas, int NumberOfColors)
: TKinokoCanvasColorScale(Canvas, NumberOfColors)
{
}

TKinokoCanvasRainbowColorScale::~TKinokoCanvasRainbowColorScale()
{
}

void TKinokoCanvasRainbowColorScale::CreateColorTable(void)
{
    static const float HueCutOff = 4.0/6.0;
    static const float Saturation = 1.0;
    static const float Brightness = 0.5;
    
    for (int i = 0; i < _NumberOfColors; i++) {
	float Value = (float) i / (_NumberOfColors - 1);
	float h = HueCutOff * (1.0 - Value);
	float s = Saturation;
	float v = Brightness;

	h = (
	    +1.00 * h
	    +0.40 * sin(2*M_PI * 3*h)/(2*M_PI * 3)
	    +0.05 * sin(2*M_PI * 6*h)/(2*M_PI * 6)
	);

	float r, g, b;
	HSV2RGB(h, s, v, r, g, b);

	int Color = _ImageArea->AllocateColorRgb(r, g, b);
	_ColorTable[i] = Color;
    }
}

static float HSV2RGB_aux(float rm1, float rm2, float h);

void TKinokoCanvasRainbowColorScale::HSV2RGB(float h, float s, float v, float& r, float& g, float& b)
{
    // see HIGZ routine IGHTOR

    // normalization
    h = ((h >= 0) && (h <= 1.0)) ? h : fmod(h + abs((int) h) + 1, 1.0f);
    s = ((s >= 0) && (s <= 1.0)) ? s : fmod(s + abs((int) s) + 1, 1.0f);
    v = ((v >= 0) && (v <= 1.0)) ? v : fmod(v + abs((int) v) + 1, 1.0f);

    // use degree unit for convenience...
    h *= 360;

    float rm1, rm2;
    if (v < 0.5) {
	rm2 = v * (1.0 + s);
    }
    else {
	rm2 = v + s - v * s;
    }
    rm1 = 2.0 * v - rm2;

    if (s == 0) {
	r = g = b = v;
    }
    else {
	r = HSV2RGB_aux(rm1, rm2, h + 120);
	g = HSV2RGB_aux(rm1, rm2, h);
	b = HSV2RGB_aux(rm1, rm2, h - 120);
    }
}

static float HSV2RGB_aux(float rm1, float rm2, float h)
{
    // see HIGZ routine IGHR01
    h = (h > 360) ? (h - 360) : ((h < 0) ? (h + 360) : h);

    if (h < 60) {
	return rm1 + (rm2 - rm1) * h / 60;
    }
    else if (h < 180) {
	return rm2;
    }
    else if (h < 240) {
	return rm1 + (rm2 - rm1) * (240 - h) / 60;
    }
    else {
	return rm1;
    }
}




TKinokoCanvasParulaColorScale::TKinokoCanvasParulaColorScale(TKinokoCanvas* Canvas, int NumberOfColors)
: TKinokoCanvasColorScale(Canvas, NumberOfColors)
{
}

TKinokoCanvasParulaColorScale::~TKinokoCanvasParulaColorScale()
{
}

void TKinokoCanvasParulaColorScale::Value2RGB(float v, float& r, float& g, float& b)
{
    static const float ColorList[64][3] = {
        { 0.2081, 0.1663, 0.5292 },
        { 0.2116238095, 0.1897809524, 0.5776761905 }, 
        { 0.212252381, 0.2137714286, 0.6269714286 },
        { 0.2081, 0.2386, 0.6770857143 }, 
        { 0.1959047619, 0.2644571429, 0.7279 },
        { 0.1707285714, 0.2919380952, 0.779247619 },
        { 0.1252714286, 0.3242428571, 0.8302714286 }, 
        { 0.0591333333, 0.3598333333, 0.8683333333 },
        { 0.0116952381, 0.3875095238, 0.8819571429 },
        { 0.0059571429, 0.4086142857, 0.8828428571 }, 
        { 0.0165142857, 0.4266, 0.8786333333 },
        { 0.032852381, 0.4430428571, 0.8719571429 },
        { 0.0498142857, 0.4585714286, 0.8640571429 }, 
        { 0.0629333333, 0.4736904762, 0.8554380952 },
        { 0.0722666667, 0.4886666667, 0.8467 },
        { 0.0779428571, 0.5039857143, 0.8383714286 }, 
        { 0.079347619, 0.5200238095, 0.8311809524 },
        { 0.0749428571, 0.5375428571, 0.8262714286 },
        { 0.0640571429, 0.5569857143, 0.8239571429 }, 
        { 0.0487714286, 0.5772238095, 0.8228285714 },
        { 0.0343428571, 0.5965809524, 0.819852381 },
        { 0.0265, 0.6137, 0.8135 },
        { 0.0238904762, 0.6286619048, 0.8037619048 },
        { 0.0230904762, 0.6417857143, 0.7912666667 }, 
        { 0.0227714286, 0.6534857143, 0.7767571429 },
        { 0.0266619048, 0.6641952381, 0.7607190476 },
        { 0.0383714286, 0.6742714286, 0.743552381 }, 
        { 0.0589714286, 0.6837571429, 0.7253857143 }, 
        { 0.0843, 0.6928333333, 0.7061666667 },
        { 0.1132952381, 0.7015, 0.6858571429 }, 
        { 0.1452714286, 0.7097571429, 0.6646285714 },
        { 0.1801333333, 0.7176571429, 0.6424333333 },
        { 0.2178285714, 0.7250428571, 0.6192619048 }, 
        { 0.2586428571, 0.7317142857, 0.5954285714 },
        { 0.3021714286, 0.7376047619, 0.5711857143 },
        { 0.3481666667, 0.7424333333, 0.5472666667 }, 
        { 0.3952571429, 0.7459, 0.5244428571 },
        { 0.4420095238, 0.7480809524, 0.5033142857 },
        { 0.4871238095, 0.7490619048, 0.4839761905 }, 
        { 0.5300285714, 0.7491142857, 0.4661142857 },
        { 0.5708571429, 0.7485190476, 0.4493904762 },
        { 0.609852381, 0.7473142857, 0.4336857143 }, 
        { 0.6473, 0.7456, 0.4188 },
        { 0.6834190476, 0.7434761905, 0.4044333333 }, 
        { 0.7184095238, 0.7411333333, 0.3904761905 }, 
        { 0.7524857143, 0.7384, 0.3768142857 },
        { 0.7858428571, 0.7355666667, 0.3632714286 },
        { 0.8185047619, 0.7327333333, 0.3497904762 }, 
        { 0.8506571429, 0.7299, 0.3360285714 },
        { 0.8824333333, 0.7274333333, 0.3217 }, 
        { 0.9139333333, 0.7257857143, 0.3062761905 },
        { 0.9449571429, 0.7261142857, 0.2886428571 },
        { 0.9738952381, 0.7313952381, 0.266647619 }, 
        { 0.9937714286, 0.7454571429, 0.240347619 },
        { 0.9990428571, 0.7653142857, 0.2164142857 },
        { 0.9955333333, 0.7860571429, 0.196652381 }, 
        { 0.988, 0.8066, 0.1793666667 },
        { 0.9788571429, 0.8271428571, 0.1633142857 }, 
        { 0.9697, 0.8481380952, 0.147452381 },
        { 0.9625857143, 0.8705142857, 0.1309 }, 
        { 0.9588714286, 0.8949, 0.1132428571 },
        { 0.9598238095, 0.9218333333, 0.0948380952 },
        { 0.9661, 0.9514428571, 0.0755333333 }, 
        { 0.9763, 0.9831, 0.0538 }
    };
    
    int k = (int) (v * 64);
    r = ColorList[k][0];
    g = ColorList[k][1];
    b = ColorList[k][2];
}
    
void TKinokoCanvasParulaColorScale::CreateColorTable(void)
{
    float r, g, b;
    for (int i = 0; i < _NumberOfColors; i++) {
        Value2RGB((float) i / _NumberOfColors, r, g, b);
	int Color = _ImageArea->AllocateColorRgb(r, g, b);
        _ColorTable[i] = Color;
    }
}
