/* KinokoCanvasSchematicObject.cc */
/* Created by Enomoto Sanshiro on 24 September. */
/* Last updated by Enomoto Sanshiro on 24 September. */


#include <string>
#include <iostream>
#include <cstdio>
#include <cmath>
#include <algorithm>
#include "KinokoCanvasPictureObject.hh"
#include "KinokoCanvasColorScale.hh"
#include "KinokoCanvasSchematicObject.hh"

using namespace std;

template<typename T> static inline T sqr(const T& x) { return x*x; }


TKinokoCanvasSchematicObject::TKinokoCanvasSchematicObject(TKinokoCanvas* Canvas)
: TKinokoCanvasPictureObject(Canvas)
{
}

TKinokoCanvasSchematicObject::~TKinokoCanvasSchematicObject()
{
    map<string, TKinokoCanvasSchematicScale*>::iterator i;
    for (i = _ScaleTable.begin(); i != _ScaleTable.end(); i++) {
	delete i->second;
    }

    map<int, TKinokoCanvasSchematicItem*>::iterator j;
    for (j = _ItemTable.begin(); j != _ItemTable.end(); j++) {
	delete j->second;
    }

    map<string, TKinokoCanvasSchematicItem*>::iterator k;
    for (
	k = _ItemPrototypeTable.begin(); 
	k != _ItemPrototypeTable.end(); 
	k++
    ){
	delete k->second;
    }
}

TKinokoShellObject* TKinokoCanvasSchematicObject::Clone(void)
{
    return new TKinokoCanvasSchematicObject(_Canvas);
}

void TKinokoCanvasSchematicObject::Initialize(void) 
{
    TKinokoCanvasPictureObject::Initialize();

    _ItemPrototypeTable["text"] = new TKinokoCanvasSchematicTextItem();
    _ItemPrototypeTable["rectangle"] = new TKinokoCanvasSchematicRectangleItem();
    _ItemPrototypeTable["circle"] = new TKinokoCanvasSchematicCircleItem();
    _ItemPrototypeTable["pie"] = new TKinokoCanvasSchematicPieItem();
}

int TKinokoCanvasSchematicObject::ProcessCommand(const std::string& Command, std::istream& InputStream)
{
    if (Command == "createscale") {
	return ProcessCreateScaleCommand(InputStream);
    }
    else if (Command == "setscalerange") {
	return ProcessSetScaleRangeCommand(InputStream);
    }
    else if (Command == "setscalelog") {
	return ProcessSetScaleLogCommand(InputStream);
    }
    else if (Command == "setscaletitle") {
	return ProcessSetScaleTitleCommand(InputStream);
    }
    else if (Command == "createitem") {
	return ProcessCreateItemCommand(InputStream);
    }
    else if (Command == "setvalue") {
	return ProcessSetValueCommand(InputStream);
    }
    else if (Command == "drawscale") {
	return ProcessDrawScaleCommand(InputStream);
    }
    else if (Command == "drawtitle") {
	return ProcessDrawTitleCommand(InputStream);
    }

    return TKinokoCanvasPictureObject::ProcessCommand(Command, InputStream);
}

int TKinokoCanvasSchematicObject::ProcessCreateScaleCommand(std::istream& InputStream)
{
    double Min, Max;
    string Name;
    if (! (InputStream >> Name >> Min >> Max)) {
	return 0;
    }

    string ScaleType = "parula", Background = "none";
    InputStream >> ScaleType >> Background;
    
    TKinokoCanvasColorScale* ColorScale = (
        TKinokoCanvasColorScaleFactory::Create(_Canvas, ScaleType + " " + Background)
    );
    if (ColorScale == 0) {
	return 0;
    }
    ColorScale->Initialize();

    _ScaleTable[Name] = new TKinokoCanvasSchematicScale(ColorScale, Min, Max);

    return 1;
}

int TKinokoCanvasSchematicObject::ProcessSetScaleRangeCommand(std::istream& InputStream)
{
    double Min, Max;
    string Name;
    if (! (InputStream >> Name >> Min >> Max)) {
	return 0;
    }

    TKinokoCanvasSchematicScale* Scale = this->GetScale(Name);
    if (! Scale) {
        return 0;
    }
    Scale->SetRange(Min, Max);

    return 1;
}

int TKinokoCanvasSchematicObject::ProcessSetScaleTitleCommand(std::istream& InputStream)
{
    string Name, Title;
    if (! (InputStream >> Name)) {
	return 0;
    }

    TKinokoCanvasSchematicScale* Scale = this->GetScale(Name);
    if (! Scale) {
        return 0;
    }

    if (getline(InputStream >> ws, Title, '\n')) {
        Scale->SetTitle(Title);
    }

    return 1;
}

int TKinokoCanvasSchematicObject::ProcessSetScaleLogCommand(std::istream& InputStream)
{
    string Name;
    if (! (InputStream >> Name)) {
	return 0;
    }

    TKinokoCanvasSchematicScale* Scale = this->GetScale(Name);
    if (! Scale) {
        return 0;
    }
    Scale->SetLog();

    return 1;
}

int TKinokoCanvasSchematicObject::ProcessCreateItemCommand(std::istream& InputStream
)
{
    int ItemId;
    string ItemTypeName;
    TKinokoCanvasSchematicItem* Item = 0;
    if (InputStream >> ItemId >> ItemTypeName) {
	Item = _ItemPrototypeTable[ItemTypeName];
    }

    int OriginalColor = _ImageArea->SetColor(
	_ColorIndexList[Color_Foreground]
    );
    int OriginalBackgroundColor = _ImageArea->SetBackgroundColor(
	_ColorIndexList[Color_Background]
    );

    if (Item) {
	Item = Item->Clone();
	Item->Attach(this);
	Item->Construct(InputStream);
	_ItemTable[ItemId] = Item;
    }

    _ImageArea->SetColor(OriginalColor);
    _ImageArea->SetBackgroundColor(OriginalBackgroundColor);

    return (Item != 0) ? 1 : 0;
}

int TKinokoCanvasSchematicObject::ProcessSetValueCommand(std::istream& InputStream)
{
    int Result = 0;

    int OriginalColor = _ImageArea->SetColor(-1);
    string OriginalAdjustment = _ImageArea->SetTextAdjustment("tl");

    int ItemId;
    TKinokoCanvasSchematicItem* Item;
    while (InputStream >> ItemId) {
	Item = _ItemTable[ItemId];
	if (! Item) {
	    break;
	}
	if (! Item->SetValue(InputStream)) {
	    break;
	}
	Result++;
    }

    _ImageArea->SetColor(OriginalColor);
    _ImageArea->SetTextAdjustment(OriginalAdjustment);

    return Result;
}

int TKinokoCanvasSchematicObject::ProcessDrawScaleCommand(std::istream& InputStream)
{
    float x0, y0, x1, y1;
    string Name;
    if (! (InputStream >> Name >> x0 >> y0 >> x1 >> y1)) {
	return 0;
    }

    TKinokoCanvasSchematicScale* Scale = this->GetScale(Name);
    if (! Scale) {
        return 0;
    }
    Scale->Draw(this, x0, y0, x1, y1);

    return 1;
}

int TKinokoCanvasSchematicObject::ProcessDrawTitleCommand(std::istream& InputStream)
{
    TKinokoCanvasPictureObject::DrawFrame();
    return 1;
}

int TKinokoCanvasSchematicObject::ProcessRangeSelect(int X0, int Y0, int X1, int Y1)
{
    int Result = 0;

    typedef map<int, TKinokoCanvasSchematicItem*>::iterator Iterator;
    for (Iterator i = _ItemTable.begin(); i != _ItemTable.end(); i++) {
	if (i->second->Includes(X0, Y0)) {
	    cerr << "Schematic: item #" << i->first << " clicked" << endl;
	    Result = 1;
	    break;
	}
    }

    return Result;
}

TKinokoCanvasImageArea* TKinokoCanvasSchematicObject::GetImageArea(void)
{
    return _ImageArea;
}

TKinokoCanvasTypesetter* TKinokoCanvasSchematicObject::GetTypesetter(void)
{
    return _Typesetter;
}

TKinokoCanvasMathTextComposer* TKinokoCanvasSchematicObject::GetMathTextComposer(void)
{
    return _MathTextComposerList[Font_Label];
}

TKinokoCanvasSchematicScale* TKinokoCanvasSchematicObject::GetScale(const std::string& Name)
{
    return _ScaleTable[Name];
}



TKinokoCanvasSchematicScale::TKinokoCanvasSchematicScale(TKinokoCanvasColorScale* ColorScale, double Min, double Max)
: TKinokoCanvasColorScale(0, 0)
{
    _ColorScale = ColorScale;
    _DataScale = 0;

    _Min = Min;
    _Max = Max;
    _Width = _Max - _Min;

    _IsLog = false;
}

TKinokoCanvasSchematicScale::~TKinokoCanvasSchematicScale()
{
    delete _ColorScale;
    delete _DataScale;
}

int TKinokoCanvasSchematicScale::ColorIndexOf(double Value)
{
    double r;
    if (! _IsLog) {
        r = (Value - _Min) / _Width;
    }
    else {
        if (Value <= 0) {
            r = 0;
        }
        else {
            r = (log10(Value) - _LogMin) / _LogWidth;
        }
    }

    if (Value == 0) {
        _ColorScale->BackgroundColorIndex();
    }
    
    return _ColorScale->ColorIndexOf(min(max(r, 0.0), 1.0));
}

void TKinokoCanvasSchematicScale::Initialize(void)
{
    _ColorScale->Initialize();
}

void TKinokoCanvasSchematicScale::SetRange(double Min, double Max)
{
    _Min = min(Min, Max);
    _Max = max(Min, Max);
    _Width = _Max - _Min;

    if (_IsLog) {
        SetLog();
    }
}

void TKinokoCanvasSchematicScale::SetLog(bool IsLog)
{
    if (IsLog) {
	if (_Min <= 0) {
            if (_Max < 0) {
                _Max = 1.0;
            }
	    _Min = min<double>(_Max / 100.0, 0.5);
	}

	_LogMin = log10(_Min);
	_LogMax = log10(_Max);
	_LogWidth = _LogMax - _LogMin;
    }

    if (IsLog != _IsLog) {
        delete _DataScale;
        if (IsLog) {
            _DataScale = new TKinokoCanvasLogDataScale(_Min, _Max);
        }
        else {
            _DataScale = new TKinokoCanvasDataScale(_Min, _Max);
        }
    }

    _IsLog = IsLog;
}

void TKinokoCanvasSchematicScale::SetTitle(const std::string& Title)
{
    _Title = Title;
}

void TKinokoCanvasSchematicScale::Draw(TKinokoCanvasSchematicObject* SchematicObject, float x0, float y0, float x1, float y1)
{
    TKinokoCanvasImageArea* ImageArea = SchematicObject->GetImageArea();
    TKinokoCanvasTypesetter* Typesetter = SchematicObject->GetTypesetter();
    TKinokoCanvasMathTextComposer* MathTextComposer = (
        SchematicObject->GetMathTextComposer()
    );
    int cx0 = SchematicObject->CanvasXOf(x0);
    int cy0 = SchematicObject->CanvasYOf(y0);
    int cx1 = SchematicObject->CanvasXOf(x1);
    int cy1 = SchematicObject->CanvasYOf(y1);

    if (! _DataScale) {
        if (_IsLog) {
            _DataScale = new TKinokoCanvasLogDataScale(_Min, _Max);
        }
        else {
            _DataScale = new TKinokoCanvasDataScale(_Min, _Max);
        }
    }
    else {
        _DataScale->Rescale(_Min, _Max);
    }

    int OldColorIndex = ImageArea->SetColor(-1);
    int OldFontIndex = ImageArea->SetFont(-1);

    // Color Box //
    int NumberOfColors = 256;
    float ColorStep = 1.0 / NumberOfColors;
    int ScaleHeight = cy1 - cy0;
    ImageArea->DrawRect(cx0, cy0, cx1, cy1);
    for (float z = 0; z < 1.0; z += ColorStep) {
	ImageArea->SetColor(_ColorScale->ColorIndexOf(z));
	ImageArea->DrawRectFill(
	    cx0, cy1 - (int) (ScaleHeight * (z + ColorStep)),
	    cx1, cy1 - (int) (ScaleHeight * z)
	);
    }

    ImageArea->SetColor(OldColorIndex);

    // Title //
    if (! _Title.empty()) {
        ImageArea->DrawText(cx0-5, cy0-10, _Title, "bl");
    }

    // Labels //
    int NumberOfDivisions = _DataScale->NumberOfDivisions();
    for (int i = 0; i < NumberOfDivisions; i++) {
        float zs = _DataScale->DivisionValueOf(i), y;
        if (! _IsLog) {
            y = cy1 - (ScaleHeight * (zs - _Min) / _Width);
        }
        else {
            y = cy1 - (ScaleHeight * (log10(zs) - _LogMin) / _LogWidth);
        }
        ImageArea->DrawLine(cx1, y, cx1 + 3, y);
	    
        string Label = _DataScale->DivisionLabelOf(i);
        Typesetter->Typeset(
            cx1 + 3, y, 
            MathTextComposer->Compose(Label),
            "cl"
	);
    }	

    // Tick Marks //
    int NumberOfTicks = _DataScale->NumberOfTicks();
    for (int i = 0; i < NumberOfTicks; i++) {
        float zs = _DataScale->TickValueOf(i);
        float y;
        if (! _IsLog) {
            y = cy1 - ScaleHeight * (zs - _Min) / _Width;
        }
        else {
            y = cy1 - ScaleHeight * (log10(zs) - _LogMin) / _LogWidth;
        }
        ImageArea->DrawLine(cx1, y, cx1 + 2, y);
    }

    ImageArea->SetFont(OldFontIndex);
}



TKinokoCanvasSchematicItem::TKinokoCanvasSchematicItem(void)
{
    _SchematicObject = 0;
}

TKinokoCanvasSchematicItem::~TKinokoCanvasSchematicItem()
{
}

void TKinokoCanvasSchematicItem::Attach(TKinokoCanvasSchematicObject* SchematicObject)
{
    _SchematicObject = SchematicObject;

    _ImageArea = _SchematicObject->GetImageArea();
}



TKinokoCanvasSchematicTextItem::TKinokoCanvasSchematicTextItem(void)
{
    _Buffer = 0;
    _BufferSize = 0;

    _Color = -1;
}

TKinokoCanvasSchematicTextItem::~TKinokoCanvasSchematicTextItem()
{
    delete[] _Buffer;
}

TKinokoCanvasSchematicItem* TKinokoCanvasSchematicTextItem::Clone(void)
{
    return new TKinokoCanvasSchematicTextItem();
}

int TKinokoCanvasSchematicTextItem::Construct(std::istream& InputStream)
{
    int Result = 0;

    float X0, Y0;
    if (InputStream >> X0 >> Y0) {
	_X0 = CanvasXOf(X0);
	_Y0 = CanvasYOf(Y0);
	Result = 1;
    }

    getline(InputStream >> ws, _Format);
    if (_Format.empty()) {
	_Format = "%.0f";
    }

    delete[] _Buffer;
    _BufferSize = _Format.size() + 32;
    _Buffer = new char[_BufferSize + 1];

    _Color = _ImageArea->SetColor(-1);

    return Result;
}

int TKinokoCanvasSchematicTextItem::SetValue(std::istream& InputStream)
{
    int Result = 0;

    double Value;
    if (InputStream >> Value) {
	snprintf(_Buffer, _BufferSize, _Format.c_str(), Value);
	_ImageArea->SetColor(_Color);
	_ImageArea->DrawText(_X0, _Y0, _Buffer);
	_X1 = _X0 + _ImageArea->TextWidthOf(_Buffer);
	_Y1 = _Y0 + _ImageArea->TextHeightOf(_Buffer);
	Result = 1;
    }

    return Result;
}

bool TKinokoCanvasSchematicTextItem::Includes(int X, int Y)
{
    return ((X >= _X0) && (Y >= _Y0) && (X < _X1) && (Y < _Y1));
}



TKinokoCanvasSchematicRectangleItem::TKinokoCanvasSchematicRectangleItem(void)
{
}

TKinokoCanvasSchematicRectangleItem::~TKinokoCanvasSchematicRectangleItem()
{
}

TKinokoCanvasSchematicItem* TKinokoCanvasSchematicRectangleItem::Clone(void)
{
    return new TKinokoCanvasSchematicRectangleItem();
}

int TKinokoCanvasSchematicRectangleItem::Construct(std::istream& InputStream)
{
    int Result = 0;

    float X0, Y0, X1, Y1;
    string ScaleName;
    if (InputStream >> X0 >> Y0 >> X1 >> Y1 >> ScaleName) {
	_X0 = CanvasXOf(X0);
	_Y0 = CanvasYOf(Y0);
	_X1 = CanvasXOf(X1);
	_Y1 = CanvasYOf(Y1);

	_Scale = _SchematicObject->GetScale(ScaleName);
	Result = (_Scale) ? 1 : 0;
    }

    return Result;
}

int TKinokoCanvasSchematicRectangleItem::SetValue(std::istream& InputStream)
{
    int Result = 0;

    double Value;
    if (InputStream >> Value) {
	_ImageArea->SetColor(_Scale->ColorIndexOf(Value));
	_ImageArea->DrawRectFill(_X0, _Y0, _X1, _Y1);
	Result = 1;
    }

    return Result;
}

bool TKinokoCanvasSchematicRectangleItem::Includes(int X, int Y)
{
    return ((X >= _X0) && (Y >= _Y0) && (X < _X1) && (Y < _Y1));
}



TKinokoCanvasSchematicCircleItem::TKinokoCanvasSchematicCircleItem(void)
{
}

TKinokoCanvasSchematicCircleItem::~TKinokoCanvasSchematicCircleItem()
{
}

TKinokoCanvasSchematicItem* TKinokoCanvasSchematicCircleItem::Clone(void)
{
    return new TKinokoCanvasSchematicCircleItem();
}

int TKinokoCanvasSchematicCircleItem::Construct(std::istream& InputStream)
{
    int Result = 0;

    float X0, Y0, Radius;
    string ScaleName;
    if (InputStream >> X0 >> Y0 >> Radius >> ScaleName) {
	_X0 = CanvasXOf(X0);
	_Y0 = CanvasYOf(Y0);
	_Rx = fabs(CanvasXOf(X0 + Radius) - CanvasXOf(X0));
	_Ry = fabs(CanvasYOf(Y0 + Radius) - CanvasYOf(Y0));

	_Scale = _SchematicObject->GetScale(ScaleName);
	Result = (_Scale) ? 1 : 0;
    }

    return Result;
}

int TKinokoCanvasSchematicCircleItem::SetValue(std::istream& InputStream)
{
    int Result = 0;

    double Value;
    if (InputStream >> Value) {
	_ImageArea->SetColor(_Scale->ColorIndexOf(Value));
	_ImageArea->DrawEllipseFill(_X0, _Y0, _Rx, _Ry);
	Result = 1;
    }

    return Result;
}

bool TKinokoCanvasSchematicCircleItem::Includes(int X, int Y)
{
    return ((fabs(X - _X0) < _Rx) && (fabs(Y - _Y0) < _Ry));
}



TKinokoCanvasSchematicPieItem::TKinokoCanvasSchematicPieItem(void)
{
}

TKinokoCanvasSchematicPieItem::~TKinokoCanvasSchematicPieItem()
{
}

TKinokoCanvasSchematicItem* TKinokoCanvasSchematicPieItem::Clone(void)
{
    return new TKinokoCanvasSchematicPieItem();
}

int TKinokoCanvasSchematicPieItem::Construct(std::istream& InputStream)
{
    int Result = 0;

    float X0, Y0, Radius, Rx, Ry;
    string ScaleName;
    if (InputStream >> X0 >> Y0 >> Radius >> _Theta0 >> _Theta1 >> ScaleName) {
	_X0 = CanvasXOf(X0);
	_Y0 = CanvasYOf(Y0);
        Rx = fabs(CanvasXOf(X0 + Radius) - CanvasXOf(X0));
	Ry = fabs(CanvasYOf(Y0 + Radius) - CanvasYOf(Y0));
        _R = min(Rx, Ry);

	_Scale = _SchematicObject->GetScale(ScaleName);
        _BorderColor = _ImageArea->AllocateColor("lightgray");
	Result = (_Scale) ? 1 : 0;
    }

    return Result;
}

int TKinokoCanvasSchematicPieItem::SetValue(std::istream& InputStream)
{
    int Result = 0;

    double Value;
    if (InputStream >> Value) {
	_ImageArea->SetColor(_Scale->ColorIndexOf(Value));
	_ImageArea->DrawArcFill(_X0, _Y0, _R, _R, _Theta0, _Theta1);

        _ImageArea->SetColor(_BorderColor);
	_ImageArea->DrawArc(_X0, _Y0, _R, _R, _Theta0, _Theta1);
	_ImageArea->DrawLine(
            _X0, _Y0, _X0 + _R * cos(_Theta0), _Y0 - _R * sin(_Theta0)
        );
	_ImageArea->DrawLine(
            _X0, _Y0, _X0 + _R * cos(_Theta1), _Y0 - _R * sin(_Theta1)
        );

	Result = 1;
    }

    return Result;
}

bool TKinokoCanvasSchematicPieItem::Includes(int X, int Y)
{
    if (sqr(X-_X0) + sqr(Y-_Y0) > sqr(_R)) {
        return false;
    }

    double Theta = atan2(Y-_Y0, X-_X0);
    if (Theta < 0) {
        Theta += 2 * M_PI;
    }

    return (Theta >= _Theta0) && (Theta <= _Theta1);
}


#if 0
.create schematic s 0 0 100 100;
s createscale scale0 0 10 parula black;
s createitem 1 text 0.2 0.2 hello: %10.2f;
s createitem 2 rectangle 0.5 0.5 0.7 0.7 scale0;
s setvalue 1 3.275;
s setvalue 2 2.789;
#endif
